/* disasm.c */
#include "z80em.h"

#ifdef Z80_DEBUG

unsigned char read_mem(unsigned char *, int);

#define GETWORD(i) ((read_mem(reg, i+1)<<8)+read_mem(reg, i))

char *str_ali[8] = { "ADD", "ADC", "SUB", "SBC", "AND", "XOR", "OR", "CP" };
char *str_r[8] = { "B", "C", "D", "E", "H", "L", "(HL)", "A" };
char *str_cc[8] = { "NZ", "Z", "NC", "C", "PO", "PE", "P", "M" };
char *str_ss[4] = { "BC", "DE", "HL", "SP" };
char *str_qq[4] = { "BC", "DE", "HL", "AF" };
char *str_ixiy[2] = { "IX", "IY" };

char *cb00[8] = { "RLC", "RRC", "RL", "RR", "SLA", "SRA", "SLL", "SRL" };
char *cbxx[3] = { "BIT", "RES", "SET" };

const char htoa[16] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

char *put_s(char *buf, char *s) {
    while (*s) { *buf++ = *s++; }
    *buf = 0;
    return buf;
}
char *put_c(char *buf, char c) {
    *buf++ = c;
    *buf = 0;
    return buf;
}
char *put_d(char *buf, int x) {
    int first = 0;
    char c;
    for (c = '0'; x > 9999; x -= 10000) {
	c++;
    }
    if (!first && c != '0') { *buf++ = c; first = 1; }
    for (c = '0'; x > 999; x -= 1000) {
	c++;
    }
    if (!first && c != '0') { *buf++ = c; first = 1; }
    for (c = '0'; x > 99; x -= 100) {
	c++;
    }
    if (!first && c != '0') { *buf++ = c; first = 1; }
    for (c = '0'; x > 9; x -= 10) {
	c++;
    }
    if (!first && c != '0') { *buf++ = c; first = 1; }
    *buf++ = '0' + x;
    *buf = 0;
    return buf;
}
char *put_4x(char *buf, int x) {
    *buf++ = htoa[(x >> 12) & 0x0f];
    *buf++ = htoa[(x >>  8) & 0x0f];
    *buf++ = htoa[(x >>  4) & 0x0f];
    *buf++ = htoa[x & 0x0f];
    *buf = 0;
    return buf;
}
char *put_2x(char *buf, int x) {
    *buf++ = htoa[(x >>  4) & 0x0f];
    *buf++ = htoa[x & 0x0f];
    *buf = 0;
    return buf;
}

int sext(unsigned char x) {
	if (x < 0x80) return (int)x;
	return (int)x - 256;
}

char *sprintf_ixiy(char *p, int isiy, int d) {
	char *r;
	if (!d)
	    return put_s(p, isiy ? "(IY)" : "(IX)");
	r = put_s(p, isiy ? "(IY" : "(IX");
	r = put_c(r, d > 0 ? '+' : '-');
	r = put_d(r, d > 0 ? d : -d);
	r = put_c(r, ')');
	return r;
}

int ixiy(unsigned char *reg, int *indexp, int isiy, char *p) {
	unsigned char opc;
	int index;
	int aaa, bbb, d;

	index = *indexp;
	opc = read_mem(reg, index);
	switch (opc) {
	    case 0xED:
undefined:
		put_s(p, "????");
		return 0;

	    case 0xCB:
		index++;
		d   = sext(read_mem(reg, index++));
		opc = read_mem(reg, index++);
		aaa = (opc >> 3) & 0x07;
		bbb = opc & 0x07;
		switch (opc >> 6) {
		    case 0:
			p = put_s(p, cb00[aaa]);
			p = put_c(p, ' ');
			p = sprintf_ixiy(p, isiy, d);
			if (bbb != 6) {
			    p = put_c(p, ';');
			    p = put_s(p, str_r[bbb]);
			}
			break;
		    default:
			p = put_s(p, cbxx[(opc >> 6)-1]);
			p = put_c(p, ' ');
			p = put_d(p, aaa);
			p = put_c(p, ',');
			sprintf_ixiy(p, isiy, d);
			break;
		}
		break;

	    case 0xDD:
	    case 0xFD:
		goto undefined;

	    default:
		aaa = (opc >> 3) & 0x07;
		bbb = opc & 0x07;
		switch (opc >> 6) {
		    case 0:
			switch (bbb) {
			    case 0:
				goto undefined;

			    case 1:
				if (aaa & 1) {
				    p = put_s(p, "ADD ");
				    if (aaa == 5) {
					p = put_s(p, str_ixiy[isiy]);
					p = put_c(p, ',');
					p = put_s(p, str_ixiy[isiy]);
				    } else {
					p = put_s(p, str_ixiy[isiy]);
					p = put_c(p, ',');
					p = put_s(p, str_ss[aaa >> 1]);
				    }
				    index++;
				} else {
				    if (aaa != 4) goto undefined;
				    p = put_s(p, "LD ");
				    p = put_s(p, str_ixiy[isiy]);
				    p = put_c(p, ',');
				    p = put_4x(p, GETWORD(index+1));
				    index += 3;
				}
				break;
			    case 2:
				if (aaa == 4) {
				    p = put_s(p, "LD (");
				    p = put_4x(p, GETWORD(index+1));
				    p = put_s(p, "),");
				    p = put_s(p, str_ixiy[isiy]);
				    index += 3;
				} else if(aaa == 5) {
				    p = put_s(p, "LD ");
				    p = put_s(p, str_ixiy[isiy]);
				    p = put_s(p, ",(");
				    p = put_4x(p, GETWORD(index+1));
				    p = put_c(p, ')');
				    index += 3;
				} else goto undefined;
				break;
			    case 3:
				if (aaa == 4 || aaa == 5) {
				    p = put_s(p, (aaa & 1) ? "DEC " : "INC ");
				    p = put_s(p, str_ixiy[isiy]);
				    index++;
				} else goto undefined;
				break;
			    case 4:
				if (aaa == 4 || aaa == 5) {
				    p = put_s(p, "INC ");
				    p = put_s(p, str_ixiy[isiy]);
				    p = put_s(p, str_r[aaa]);
				} else if (aaa == 6) {
				    d = sext(read_mem(reg, ++index));
				    p = put_s(p, "INC ");
				    p = sprintf_ixiy(p, isiy, d);
				} else goto undefined;
				index++;
				break;
			    case 5:
				if (aaa == 4 || aaa == 5) {
				    p = put_s(p, "DEC ");
				    p = put_s(p, str_ixiy[isiy]);
				    p = put_s(p, str_r[aaa]);
				} else if (aaa == 6) {
				    d = sext(read_mem(reg, ++index));
				    p = put_s(p, "DEC ");
				    p = sprintf_ixiy(p, isiy, d);
				} else goto undefined;
				index++;
				break;
			    case 6:
				if (aaa == 4 || aaa == 5) {
				    p = put_s(p, "LD ");
				    p = put_s(p, str_ixiy[isiy]);
				    p = put_s(p, str_r[aaa]);
				    p = put_c(p, ',');
				    p = put_2x(p, read_mem(reg, ++index));
				} else if (aaa == 6) {
				    p = put_s(p, "LD ");
				    p = sprintf_ixiy(p, isiy, sext(read_mem(reg, ++index)));
				    p = put_2x(p, read_mem(reg, ++index));
				} else goto undefined;
				index++;
				break;
			    case 7:
				goto undefined;
			}
			break;
		case 1:	/* LD r,r */
		    if (opc == 0x76) goto undefined;
		    p = put_s(p, "LD ");
		    if (aaa == 6)
			p = sprintf_ixiy(p, isiy, sext(read_mem(reg, ++index)));
		    else {
			if (aaa == 4 || aaa == 5) p = put_s(p, str_ixiy[isiy]);
			p = put_s(p, str_r[aaa]);
		    }
		    p = put_c(p, ',');
		    if (bbb == 6)
			p = sprintf_ixiy(p, isiy, sext(read_mem(reg, ++index)));
		    else {
			if (bbb == 4 || bbb == 5) p = put_s(p, str_ixiy[isiy]);
			p = put_s(p, str_r[bbb]);
		    }
		    index++;
		    break;
		case 2:	/* 8bit alithmetic */
		    p = put_s(p, str_ali[aaa]);
		    p = put_c(p, ' ');
		    if (aaa<4 && aaa!=2) p = put_s(p, "A,");
		    if (aaa == 6)
			p = sprintf_ixiy(p, isiy, sext(read_mem(reg, ++index)));
		    else {
			if (aaa == 4 || aaa == 5) p = put_s(p, str_ixiy[isiy]);
			p = put_s(p, str_r[aaa]);
		    }
		    index++;
		    break;
		case 3:
		    switch (bbb) {
			case 0: /* RET cc */
			case 2: /* JP cc */
			case 4: /* CALL cc */
			case 6: /* 8bit alithmetic with immidiate */
			case 7: /* RST xxh */
			    goto undefined;
			case 1: /* POP qq, etc. */
			    switch (aaa) {
				case 4:
				    p = put_s(p, "POP ");
				    p = put_s(p, str_ixiy[isiy]);
				    break;
				case 5:
				    p = put_s(p, "JP (");
				    p = put_s(p, str_ixiy[isiy]);
				    p = put_c(p, ')');
				    break;
				case 7:
				    p = put_s(p, "LD SP, ");
				    p = put_s(p, str_ixiy[isiy]);
				    break;
				default:
				    goto undefined;
			    }
			    index++;
			    break;
			case 3: /* JP, etc. */
			    if (aaa == 4) {
				p = put_s(p, "EX (SP),");
				p = put_s(p, str_ixiy[isiy]);
			    } else goto undefined;
			    index++;
			    break;
			case 5: /* PUSH qq, CALL nn */
			    if (aaa == 4) {
				p = put_s(p, "PUSH ");
				p = put_s(p, str_ixiy[isiy]);
			    } else goto undefined;
			    index++;
			    break;
		    }
		    break;
	    }
	    break;
	}
	*indexp = index;
	return 1;
}

void disasm(unsigned char *reg, int index, char *p) {
	unsigned char opc;

	int aaa, bbb;

	opc = read_mem(reg, index);

	switch (opc) {
	    case 0xED:
		index++;
		opc = read_mem(reg, index);
		aaa = (opc >> 3) & 0x07;
		bbb = opc & 0x07;
		switch (opc >> 6) {
		    case 0:
		    case 3:
undefined:
			p = put_s(p, "????");
			break;
		    case 1:
			switch (bbb) {
			    case 0:
				p = put_s(p, "IN ");
				p = put_s(p, str_r[aaa]);
				p = put_s(p, ",(C)");
				index++;
				break;
			    case 1:
				p = put_s(p, "OUT (C),");
				p = put_s(p, str_r[aaa]);
				index++;
				break;
			    case 2:
				p = put_s(p, (aaa & 1) ? "ADC HL," : "SBC HL,");
				p = put_s(p, str_ss[aaa >> 1]);
				index++;
				break;
			    case 3:
				if (aaa & 1) {
				    p = put_s(p, "LD ");
				    p = put_s(p, str_ss[aaa >> 1]);
				    p = put_s(p, ",(");
				    p = put_4x(p, GETWORD(index+1));
				    p = put_s(p, ")");
				} else {
				    p = put_s(p, "LD (");
				    p = put_4x(p, GETWORD(index+1));
				    p = put_s(p, "),");
				    p = put_s(p, str_ss[aaa >> 1]);
				}
				index += 3;
				break;
			    case 4:
				if (aaa == 0) {
				    p = put_s(p, "NEG");
				    index++;
				} else goto undefined;
				break;
			    case 5:
				if (aaa < 2) {
				    p = put_s(p, "RET");
				    p = put_c(p, (aaa == 0) ? 'N' : 'I');
				    index++;
				} else goto undefined;
				break;
			    case 6:
				if (aaa == 0 || aaa == 2 || aaa == 3) {
				    p = put_s(p, "IM ");
				    p = put_c(p, (aaa == 0) ? '0' : (aaa == 2) ? '1' : '2');
				    index++;
				} else goto undefined;
				break;
			    case 7:
				if (aaa < 6) {
				    char *ed01xxx111[6] =
					{ "LD I,A", "LD R,A", "LD A,I", "LD A,R", "RRD", "RLD"};
				    p = put_s(p, ed01xxx111[aaa]);
				    index++;
				} else goto undefined;
				break;
			}
			break;
		    case 2:
			if (aaa >= 4 && bbb < 4) {
			    static char *ed10xxxxxx[16] = {
				"LDI", "LDD", "LDIR", "LDDR",
				"CPI", "CPD", "CPIR", "CPDR",
				"INI", "IND", "INIR", "INDR",
				"OUTI", "OUTD", "OTIR", "OTDR"
			    };
			    p = put_s(p, ed10xxxxxx[(aaa & 0x03) | (bbb << 2)]);
			    index++;
			} else goto undefined;
			break;
		}
		break;
	    case 0xCB:
		index++;
		opc = read_mem(reg, index);
		aaa = (opc >> 3) & 0x07;
		bbb = opc & 0x07;
		switch (opc >> 6) {
		    case 0:
			p = put_s(p, cb00[aaa]);
			p = put_c(p, ' ');
			p = put_s(p, str_r[bbb]);
			index++;
			break;
		    default:
			p = put_s(p, cbxx[(opc >> 6)-1]);
			p = put_c(p, ' ');
			p = put_d(p, aaa);
			p = put_c(p, ',');
			p = put_s(p, str_r[bbb]);
			index++;
			break;
		}
		break;
	    case 0xDD:
		index++;
		ixiy(reg, &index, 0, p);
		break;
	    case 0xFD:
		index++;
		ixiy(reg, &index, 1, p);
		break;
	    default:
		aaa = (opc >> 3) & 0x07;
		bbb = opc & 0x07;
		switch (opc >> 6) {
		    case 0:
			switch (bbb) {
			    case 0:
				if (aaa >= 4) { /* JR cc,e */
				    p = put_s(p, "JR ");
				    p = put_s(p, str_cc[aaa & 0x03]);
				    p = put_c(p, ',');
				    p = put_4x(p, index + 2 + sext(read_mem(reg, index+1)));
				    index += 2;
				} else if (aaa >= 2) { /* DJNZ, JR */
				    p = put_s(p, (aaa == 2) ? "DJNZ " : "JR ");
				    p = put_4x(p, index + 2 + sext(read_mem(reg, index+1)));
				    index += 2;
				} else if (aaa == 1) {
				    p = put_s(p, "EX AF,AF'");
				    index++;
				} else {
				    p = put_s(p, "NOP");
				    index++;
				}
				break;
			    case 1:
				if (aaa & 1) {
				    p = put_s(p, "ADD HL,");
				    p = put_s(p, str_ss[aaa >> 1]);
				    index++;
				} else {
				    p = put_s(p, "LD ");
				    p = put_s(p, str_ss[aaa >> 1]);
				    p = put_c(p, ',');
				    p = put_4x(p, GETWORD(index+1));
				    index += 3;
				}
				break;
			    case 2:
				if (aaa < 4) {
				    char *s000xx010[4] =
					{ "LD (BC),A", "LD A,(BC)", "LD (DE),A", "LD A,(DE)" };
				    p = put_s(p, s000xx010[aaa & 0x03]);
				    index++;
				} else {
				    char *s001xx010_1[4] =
					{ "LD (", "LD HL,(", "LD (", "LD A,(" };
				    char *s001xx010_2[4] =
					{ "),HL", ")", "),A", ")" };
				    p = put_s(p, s001xx010_1[aaa & 0x03]);
				    p = put_4x(p, GETWORD(index+1));
				    p = put_s(p, s001xx010_2[aaa & 0x03]);
				    index += 3;
				}
				break;
			    case 3:
				p = put_s(p, (aaa & 1) ? "DEC " : "INC ");
				p = put_s(p, str_ss[aaa >> 1]);
				index++;
				break;
			    case 4:
				p = put_s(p, "INC ");
				p = put_s(p, str_r[aaa]);
				index++;
				break;
			    case 5:
				p = put_s(p, "DEC ");
				p = put_s(p, str_r[aaa]);
				index++;
				break;
			    case 6:
				p = put_s(p, "LD ");
				p = put_s(p, str_r[aaa]);
				p = put_c(p, ',');
				p = put_2x(p, read_mem(reg, index+1));
				index += 2;
				break;
			    case 7: {
				char *s00xxx111[8] =
					{ "RLCA", "RRCA", "RLA", "RRA", "DAA", "CPL", "SCF", "CCF" };
				p = put_s(p, s00xxx111[aaa]);
				index++;
				}
				break;
			}
			break;
		    case 1:	/* LD r,r */
			if (opc == 0x76) p = put_s(p, "HALT");
			else {
			    p = put_s(p, "LD ");
			    p = put_s(p, str_r[aaa]);
			    p = put_c(p, ',');
			    p = put_s(p, str_r[bbb]);
			}
			index++;
			break;
		    case 2:	/* 8bit alithmetic */
			p = put_s(p, str_ali[aaa]);
			p = put_c(p, ' ');
			if (aaa<4 && aaa!=2) p = put_s(p, "A,");
			p = put_s(p, str_r[bbb]);
			index++;
			break;
		    case 3:
			switch (bbb) {
			    case 0: /* RET cc */
				p = put_s(p, "RET ");
				p = put_s(p, str_cc[aaa]);
				index++;
				break;
			    case 1: /* POP qq, etc. */
				if (aaa & 1) {
				    char *s11xx1001[4] =
					{ "RET", "EXX", "JP (HL)", "LD SP,HL" };
				    p = put_s(p, s11xx1001[aaa >> 1]);
				} else {
				    p = put_s(p, "POP ");
				    p = put_s(p, str_qq[aaa >> 1]);
				}
				index++;
				break;
			    case 2: /* JP cc */
				p = put_s(p, "JP ");
				p = put_s(p, str_cc[aaa]);
				p = put_c(p, ',');
				p = put_4x(p, GETWORD(index+1));
				index += 3;
				break;
			    case 3: /* JP, etc. */
				switch (aaa) {
				    case 0: /* JP nn */
					p = put_s(p, "JP ");
					p = put_4x(p, GETWORD(index+1));
					index += 3;
					break;
				    case 1: /* CB prefix */
					break;
				    case 2: /* OUT (n),A */
					p = put_s(p, "OUT (");
					p = put_2x(p, read_mem(reg, index+1));
					p = put_s(p, "),A");
					index += 2;
					break;
				    case 3: /* IN A,(n) */
					p = put_s(p, "IN A,(");
					p = put_2x(p, read_mem(reg, index+1));
					p = put_s(p, ")");
					index += 2;
					break;
				    case 4: /* EX (SP),HL */
					p = put_s(p, "EX (SP),HL");
					index++;
					break;
				    case 5: /* EX DE,HL */
					p = put_s(p, "EX DE,HL");
					index++;
					break;
				    case 6: /* DI */
					p = put_s(p, "DI");
					index++;
					break;
				    case 7: /* EI */
					p = put_s(p, "EI");
					index++;
					break;
				}
				break;
			    case 4: /* CALL cc */
				p = put_s(p, "CALL ");
				p = put_s(p, str_cc[aaa]);
				p = put_c(p, ',');
				p = put_4x(p, GETWORD(index+1));
				index += 3;
				break;
			    case 5: /* PUSH qq, CALL nn */
				if (aaa & 1) {
				    p = put_s(p, "CALL ");
				    p = put_4x(p, GETWORD(index+1));
				    index += 3;
				} else { /* CALL nn (DD, ED, FD prefix are handled above)*/
				    p = put_s(p, "PUSH ");
				    p = put_s(p, str_qq[aaa >> 1]);
				    index++;
				}
				break;
			    case 6:	/* 8bit alithmetic with immidiate */
				p = put_s(p, str_ali[aaa]);
				p = put_c(p, ' ');
				if (aaa<4 && aaa!=2) p = put_s(p, "A,");
				p = put_2x(p, read_mem(reg, index+1));
				index += 2;
				break;
			    case 7:	/* RST xxh */
				p = put_s(p, "RST ");
				p = put_2x(p, aaa<<3);
				index++;
				break;
			}
			break;
		}
		break;
	}
}
#undef GETWORD

#endif
