#include "\devkitadv\crtls\asm-macros.h"
#include "z80em.h"

	.section .iwram,"awx",%progbits

#define FIRSTRASTER 5

@====================================================================
FUNCTION(interrupt_setup)

	mov	r1,     #0x04000000	@ I/O base

1:	ldrh	r0, [r1, #6]		@ read VCount
	cmp	r0, #160
	bhs	1b			@ wait VDISP period
1:	ldrh	r0, [r1, #6]
	cmp	r0, #160
	blo	1b			@ wait VBLNK period

	orr	r2, r1, #0x00000200
	mov	r0, #0
	strh	r0, [r2, #0x08]		@ disable interrupt
	ldrh	r0, [r2]
	orr	r0, r0, #0x0004
	strh	r0, [r2]			@ enable VCount interrupt
	mov	r0, #((FIRSTRASTER<<8)+0x20)	@ enable VCount interrupt
	strh	r0, [r1, #4]
1:	add	r0, pc, #(dmasource-1b-8)
1:	str	r0, [pc, #(shiftcount-1b-8)]

	mov	r1,     #0x03000000
	orr	r1, r1, #0x00007f00
1:	add	r0, pc, #(interrupt_handler-1b-8)
	str	r0, [r1, #0x00fc]		@ entry interrupt handler

	mov	r0, #1
	strh	r0, [r2, #0x08]		@ enable interrupt

	bx	lr

END(interrupt_setup)

FUNCTION(interrupt_handler)
	mov	r1,     #0x04000000
	orr	r2, r1, #0x00000200
	mov	r0, #0
	strh	r0, [r2, #0x08]		@ disable interrupt
	ldrh	r0, [r2, #0x02]		@ check interrupt state
	tst	r0, #0x0004		@ VCount?
	bxeq	lr
	strh	r0, [r2, #0x02]		@ clear interrupt state

	ldrh	r0, [r1, #4]		@ DISPSTAT
	cmp	r0, #(160 << 8)		@ interrupt at VBLANK period?

	addls	r0, r0, #(6 << 8)	@ Next VCount
	movhi	r0, #((FIRSTRASTER << 8)+0x20)	@ Reset VCount
	strh	r0, [r1, #4]
1:	ldrls	r0, [pc, #(shiftcount-1b-8)]
	addls	r0, r0, #16
1:	addhi	r0, pc, #(dmasource-1b-8)
1:	str	r0, [pc, #(shiftcount-1b-8)]
	str	r0, [r1, #0xd4]		@ DMA3SAD
	add	r0, r1, #0x10		@ BG0HOFS
	str	r0, [r1, #0xd8]		@ DMA3DAD
	mov	r0,     #0x00000004	@ 4 word
	orrls	r0, r0, #0xa4000000	@ 1010 0100 0000 0000 (DMA start, no IRQ, start at HBLNK, 32bit)
	orrhi	r0, r0, #0x84000000	@ If in VBLNK period, no trigger by HBLNK! So start at once.
	str	r0, [r1, #0xdc]		@ start DMA

	bhi	update_key_state	@ update key state once per frame

	mov	r0, #1
	strh	r0, [r2, #0x08]		@ enable interrupt
	bx	lr

@
@	update key state
@
update_key_state:
	mov	r3,	#WorkAreaH
	orr	r3, r3, #WorkAreaL
	mvn	r0, #0
	str	r0, [r3, #KeyMatrix  ]
	str	r0, [r3, #KeyMatrix+4]
	str	r0, [r3, #KeyMatrix+8]
	ldr	r1, [r1, #0x130]	@ Key input
	tst	r1, #0x0308
	beq	softreset
	and	r0, r1, #0xa0		@ D-L-----
	and	r2, r1, #0x50		@ -U-R----
	orr	r2, r2, r0, lsr #2	@ -UDRL---
	mov	r0, #0xc3		@ 11----11
	orr	r0, r0, r2, lsr #1	@ 11UDRL11
	strb	r0, [r3, #KeyMatrix+7]	@ Arrow keys
	ldr	r0, [r3, #KeyCfgAB]
	ldrb	r2, [r3, r0, lsr #24]	@ row for B button
	tst	r1, #0x0002		@ check B button state
	biceq	r2, r2, r0, lsr #16	@ bit mask for B button
	strb	r2, [r3, r0, lsr #24]
	mov	r0, r0, ror #16		@ row for A button
	ldrb	r2, [r3, r0, lsr #24]
	tst	r1, #0x0001		@ check A button state
	biceq	r2, r2, r0, lsr #16	@ bit mask for A button
	strb	r2, [r3, r0, lsr #24]
	ldr	r0, [r3, #KeyCfgLR]
	ldrb	r2, [r3, r0, lsr #24]	@ row for R button
	tst	r1, #0x0100		@ check R button state
	biceq	r2, r2, r0, lsr #16	@ bit mask for R button
	strb	r2, [r3, r0, lsr #24]
	mov	r0, r0, ror #16		@ row for L button
	ldrb	r2, [r3, r0, lsr #24]
	tst	r1, #0x0200		@ check L button state
	biceq	r2, r2, r0, lsr #16	@ bit mask for L button
	strb	r2, [r3, r0, lsr #24]
	ldr	r0, [r3, #KeyCfgSTART]
	mov	r0, r0, ror #16		@ row for START button
	ldrb	r2, [r3, r0, lsr #24]
	tst	r1, #0x0008		@ check START button state
	biceq	r2, r2, r0, lsr #16	@ bit mask for START button
	strb	r2, [r3, r0, lsr #24]

	ldr	r0, [r3, #KeyPressOnce]
	tst	r0, r0
	bne	sw_kbd

xit:
	mov	r2,     #0x04000000	@ restore r2
	orr	r2, r2, #0x00000200

	mov	r0, #1
	strh	r0, [r2, #0x08]		@ enable interrupt

	bx	lr

shiftcount:
	.word	0

dmasource:
	.hword	0,  0,  4,  0,  0,  0,  4,  0	@  0
	.hword	0,  2,  4,  2,  0,  2,  4,  2	@  1
	.hword	0,  4,  4,  4,  0,  4,  4,  4	@  2
	.hword	0,  6,  4,  6,  0,  6,  4,  6	@  3
	.hword	0,  8,  4,  8,  0,  8,  4,  8	@  4
	.hword	0, 10,  4, 10,  0, 10,  4, 10	@  5
	.hword	0, 12,  4, 12,  0, 12,  4, 12	@  6
	.hword	0, 14,  4, 14,  0, 14,  4, 14	@  7
	.hword	0, 16,  4, 16,  0, 16,  4, 16	@  8
	.hword	0, 18,  4, 18,  0, 18,  4, 18	@  9
	.hword	0, 20,  4, 20,  0, 20,  4, 20	@ 10
	.hword	0, 22,  4, 22,  0, 22,  4, 22	@ 11
	.hword	0, 24,  4, 24,  0, 24,  4, 24	@ 12
	.hword	0, 26,  4, 26,  0, 26,  4, 26	@ 13
	.hword	0, 28,  4, 28,  0, 28,  4, 28	@ 14
	.hword	0, 30,  4, 30,  0, 30,  4, 30	@ 15
	.hword	0, 32,  4, 32,  0, 32,  4, 32	@ 16
	.hword	0, 34,  4, 34,  0, 34,  4, 34	@ 17
	.hword	0, 36,  4, 36,  0, 36,  4, 36	@ 18
	.hword	0, 38,  4, 38,  0, 38,  4, 38	@ 19
	.hword	0, 40,  4, 40,  0, 40,  4, 40	@ 20
	.hword	0, 42,  4, 42,  0, 42,  4, 42	@ 21
	.hword	0, 44,  4, 44,  0, 44,  4, 44	@ 22
	.hword	0, 46,  4, 46,  0, 46,  4, 46	@ 23
	.hword	0, 48,  4, 48,  0, 48,  4, 48	@ 24
	.hword	0, 50,  4, 50,  0, 50,  4, 50	@ 25
	.hword	0, 52,  4, 52,  0, 52,  4, 52	@ 26

sw_kbd:
	and	r1, r0, #0x0000000f	@ save key-matrix-row
	and	r2, r0, #0x0000ff00	@ save key-matrix-mask
	tst	r0,	#0x00ff0000	@ wait before keydown
	subne	r0, r0, #0x00010000	@ before key down: Plus
	orreq	r0, r0, #0x80000000	@ key pressed: Minus
	subeqs	r0, r0, #0x01000000
	movvs	r0, #0			@ key released: Overflow, Plus
	str	r0, [r3, #KeyPressOnce]
	add	r3, r3, r1
	ldrb	r0, [r3, #KeyMatrix]
	bicmi	r0, r0, r2, lsr #8
@@@	orrpl	r0, r0, r2, lsr #8
	strb	r0, [r3, #KeyMatrix]
	b	xit

softreset:
	mov	r1, #0x04000000
1:	ldr	r0, [r1, #0x130]
	and	r0, r0, #0x0308
	cmp	r0, #0x0308
	bne	1b
	mov	r1, #0x03000000
	orr	r1, r1, #0x7f00
	strb	r1, [r1, #0xfa]
	swi	0

END(interrupt_handler)

	.end
