#include "gba.h"
#include "z80em.h"

void WaitForVsync(void);
__attribute__ ((section(".iwram"), long_call)) u8 *z80emu(u32);
__attribute__ ((section(".iwram"), long_call)) void *z80reset(void);

#ifdef Z80_DEBUG
void disasm(u8 *, int, char *);
#endif

void vraminit(void);
void soundinit(void);

u8 *dram = (u8 *)0x02000000;
u8 *irom = (u8 *)0x02010000;
u8 *vram = (u8 *)0x02011000;

void __crt0_move_memory(u32, u32, u32);

extern u8 __load_start_appended;

#define IPLROM_START	(&__load_start_appended)
#define IPLROM_END	(&__load_start_appended + 0x1000)
#define TAPEIMG_START	IPLROM_END
#define TAPEINDEX_START	IPLROM_END

u8 **tapeptr = (u8 **)(WorkAreaH + WorkAreaL + TapePtr);
u8 **tapeendptr = (u8 **)(WorkAreaH + WorkAreaL + TapeEndPtr);

/* ASCII to MZ-700 Display code table */
const u8 asc2disp[96] = {
      /* SP    !     "     #     $     %     &     '     (     )     *     +     ,     -     .     / */
	0x00, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6b, 0x6a, 0x2f, 0x2a, 0x2e, 0x2d,
      /* 0     1     2     3     4     5     6     7     8     9     :     ;     <     =     >     ? */
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x4f, 0x2c, 0x51, 0x2b, 0x57, 0x49,
      /* @     A     B     C     D     E     F     G     H     I     J     K     L     M     N     O */
	0x55, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
      /* P     Q     R     S     T     U     V     W     X     Y     Z     [     \     ]     ^     _ */
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x52, 0x59, 0x54, 0x50, 0x3c,
      /* `     a     b     c     d     e     f     g     h     i     j     k     l     m     n     o */
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
      /* p     q     r     s     t     u     v     w     x     y     z     {     |     }     ~       */
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x00, 0x79, 0x00, 0xe3, 0
};

const char hextbl[16] = {
	'0', '1', '2', '3', '4', '5', '6', '7', 
	'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
};

#ifdef Z80_DEBUG
const u8 instlentbl[256] = {
    1, 3, 1, 1, 1, 1, 2, 1,
    1, 1, 1, 1, 1, 1, 2, 1,
    2, 3, 1, 1, 1, 1, 2, 1,
    2, 1, 1, 1, 1, 1, 2, 1,
    2, 3, 3, 1, 1, 1, 2, 1,
    2, 1, 3, 1, 1, 1, 2, 1,
    2, 3, 3, 1, 1, 1, 2, 1,
    2, 1, 3, 1, 1, 1, 2, 1,

    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,

    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1,

    1, 1, 3, 3, 3, 1, 2, 1,
    1, 1, 3, 2, 3, 3, 2, 1,
    1, 1, 3, 2, 3, 1, 2, 1,
    1, 1, 3, 2, 3, 0, 2, 1,
    1, 1, 3, 1, 3, 1, 2, 1,
    1, 1, 3, 1, 3, 0, 2, 1,
    1, 1, 3, 1, 3, 1, 2, 1,
    1, 1, 3, 1, 3, 0, 2, 1
};
#endif

void putattr(u32 x, u32 y, u32 a) {
	u16 db;
	volatile u16 *ab;
	db = ((a & 0x07) << 12);
	if (x & 1) {
	    db |= 0x0200;
	}
	db |= 0x43;
	ab = (((x&3)==0)||((x&3)==3)) ? (u16 *)0x06009000 : (u16 *)0x06009800;
	ab += y*32 + (x>>2)*3 + (((x&3)<3) ? (x&3) : 2);
	*ab = db;
}
void put1chr(u32 x, u32 y, u32 c, u32 a) {
	u16 df, db;
	volatile u16 *af, *ab;
	df = ((a & 0x70) <<  8);
	db = ((a & 0x07) << 12);
	if (x & 1) {
	    df |= 0x0200;
	    db |= 0x0200;
	}
	df |= c;
	db |= 0x43;
	if (a & 0x80) df |= 0x100;
	af = (((x&3)==0)||((x&3)==3)) ? (u16 *)0x06008000 : (u16 *)0x06008800;
	ab = (((x&3)==0)||((x&3)==3)) ? (u16 *)0x06009000 : (u16 *)0x06009800;
	af += y*32 + (x>>2)*3 + (((x&3)<3) ? (x&3) : 2);
	ab += y*32 + (x>>2)*3 + (((x&3)<3) ? (x&3) : 2);
	*af = df;
	*ab = db;
}
void testprint(u32 xx, u32 yy, u8 *str, u8 attr) {
    u8 c;
    while (!!(c = *str++)) {
	u32 j;
	if      (c >= 'A' && c <= 'Z') j = c - 'A' + 1;
	else if (c >= '0' && c <= '9') j = c - '0' + 0x20;
	else if (c >= 'a' && c <= 'z') j = c - 'a' + 0x101;
	else if (c >= '!' && c <= ')') j = c - '!' + 0x61;
	else if (c == '+') j = 0x6a;
	else if (c == '-') j = 0x2a;
	else if (c == ',') j = 0x2f;
	else j = 0;
	put1chr(xx, yy, j, attr);
	xx++;
	if (xx >= 40) { xx = 0; yy++; }
    }
}

#ifdef Z80_DEBUG
u8 read_mem(u8 *regs, u32 i) {
	if (i >= 0x0000 && i < 0x1000) {
	    if (regs[saver9+1] & 0x01) {
		return irom[i];
	    } else {
		return dram[i];
	    }
	} else if (i >= 0xd000) {
	    if (regs[saver9+1] & 0x02) {
		return vram[i];
	    } else {
		return dram[i];
	    }
	} else {
	    return dram[i];
	}
}
int instlen(u8 opc, u8 opc2) {
	if        (opc == 0xed) {
	    return ((opc2 & 0xc7) == 0x43) ? 4 : 2;
	} else if (opc == 0xdd || opc == 0xfd) {
	    if (opc2 == 0xcb) return 4;
	    switch ((opc2 & 0xc0)>>6) {
		case 0:
		    if (((opc2 & 0x38) == 0x30) &&
			((opc2 & 0x07) >= 0x04) &&
			((opc2 & 0x07) <= 0x06))
			return (int)instlentbl[opc2]+2;
		    break;
		case 1:
		    if (((opc2 & 0x38) == 0x30) ||
			((opc2 & 0x07) == 0x06))
			return (int)instlentbl[opc2]+2;
		    break;
		case 2:
		    if ((opc2 & 0x07) == 0x06)
			return (int)instlentbl[opc2]+2;
		    break;
	    }
	    return (int)instlentbl[opc2]+1;
	}
	return (int)instlentbl[opc];
}
void testprinthex2(u32 x, u32 y, u32 v) {
	char	buf[4];
	buf[0] = hextbl[(v & 0xf0) >> 4];
	buf[1] = hextbl[v & 0x0f];
	buf[2] = 0;
	testprint(x, y, buf, 0x70);
}
u32 testprintreg(u32 bkpt, u32 *nextip) {
	u8 buf[32];
	u8 opc, opc2;
	u8 *regs;
	u32 i, zpc;
	regs = z80emu(bkpt);
	testprinthex2(0,  25, regs[reg_A]);	// A
	testprinthex2(3,  25, regs[reg_B]);	// B
	testprinthex2(5,  25, regs[reg_C]);	// C
	testprinthex2(8,  25, regs[reg_D]);	// D
	testprinthex2(10, 25, regs[reg_E]);	// E
	testprinthex2(13, 25, regs[reg_H]);	// H
	testprinthex2(15, 25, regs[reg_L]);	// L
	testprinthex2(18, 25, regs[reg_PC+1]);	// PC-h
	testprinthex2(20, 25, regs[reg_PC]);	// PC-l
	testprinthex2(13, 24, regs[reg_IXH]);	// IXH
	testprinthex2(15, 24, regs[reg_IXL]);	// IXL
	testprinthex2(18, 24, regs[reg_IYH]);	// IYH
	testprinthex2(20, 24, regs[reg_IYL]);	// IYL
	i = regs[saver9/*reg_F*/];
	buf[0] = (i & 0x80) ? 'S' : 's';
	buf[1] = (i & 0x40) ? 'Z' : 'z';
	buf[2] = (i & 0x10) ? 'H' : 'h';
	buf[3] = (i & 0x04) ? 'V' : 'v';
	buf[4] = (i & 0x02) ? 'N' : 'n';
	buf[5] = (i & 0x01) ? 'C' : 'c';
	buf[6] = 0;
	testprint(23, 25, buf, 0x70);
	zpc = (u32)regs[reg_PC]|((u32)regs[reg_PC+1]<<8);
	opc  = read_mem(regs, zpc  );
	opc2 = read_mem(regs, zpc+1);
#if 0
	for (i=0; i<instlen(opc,opc2); i++) {
		testprinthex2(30+i*3, 25, read_mem(regs, zpc+i));
	}
	for (; i<4; i++) {
		testprint(30+i*3, 25, "--", 0x00);
	}
#else
	disasm(regs, zpc, buf);
	testprint(23, 24, "               ", 0x70);
	testprint(23, 24, buf, 0x70);
#endif
	*nextip = zpc + instlen(opc, opc2);
	return zpc;	/* next pc */
}

void tapeload_test(void) {
	u32 i;
	u32 addr_start, addr_size;
	u8 *tapeimg = (u8 *)TAPEIMG_START;
	addr_start = ((u32)tapeimg[0x15]<<8)|(u32)tapeimg[0x14];
	addr_size  = ((u32)tapeimg[0x13]<<8)|(u32)tapeimg[0x12];
	for (i=0; i<addr_size; i++)
	    dram[addr_start+i] = tapeimg[128+i];
	*tapeptr = &tapeimg[128+i];
}
#endif /*Z80_DEBUG*/

u32 chkstr(u8 *str1, u8 *str2) {
	u8 c1, c2;
	for(;;) {
	    c1 = *str1++;
	    c2 = *str2++;
	    if (c2 == 0) return 1;
	    if (c1 != c2) return 0;
	}
}

void patch_iplrom(void) {
	/* patch hot start entry to jump LOAD command address directly */
	if (chkstr(&irom[0x6f3], "1Z-009")) {
	    irom[0x00ad] = 0xc3;
	    irom[0x00ae] = 0x08;
	    irom[0x00af] = 0x01;
	} else if (chkstr(&irom[0x142],"MZ\x90\x37\x30\x30")) {
	    irom[0x0082] = 0xc3;
	    irom[0x0083] = 0xd9;
	    irom[0x0084] = 0x00;
	}
	irom[0x04d8] = 0xed;	/* RDINF entry */
	irom[0x04d9] = 0xf0;
	irom[0x04da] = 0x01;
	irom[0x04f8] = 0xed;	/* RDDATA entry */
	irom[0x04f9] = 0xf0;
	irom[0x04fa] = 0x02;
}

void config_buttons(void) {
	u32 *kconfig = (u32 *)(WorkAreaH + WorkAreaL + KeyCfgAB);
	*kconfig++ = 0x10061006;    /* B|A */
//	*kconfig++ = 0x01084008;    /* B=SHIFT|A=CTRL */
	*kconfig++ = 0x00000000;    /* R|L */
	*kconfig++ = 0x00002002;    /* START */
}

#ifdef SPEED_METER
/* speed meter for debug */
void init_speedmeter(void) {
	volatile u16 *oam = 0x07000000;
	int i;
	for (i=0; i<40; i++) {
	    put1chr(i, 25, 0x73, (i<22) ? 0x04 : 0x02);
	}
	*((volatile u16*)0x04000000) |= 0x1000;
	*((volatile u16*)0x05000202) = 0x0000;
	*((volatile u16*)0x05000204) = 0x7fff;
	*((volatile u32*)0x06010000) = 0x00001000;
	*((volatile u32*)0x06010004) = 0x00012100;
	*((volatile u32*)0x06010008) = 0x00122210;
	*((volatile u32*)0x0601000c) = 0x01222221;
	*((volatile u32*)0x06010010) = 0x01222221;
	*((volatile u32*)0x06010014) = 0x01222221;
	*((volatile u32*)0x06010018) = 0x01222221;
	*((volatile u32*)0x0601001c) = 0x01111111;
	*oam++ = 0x0000+152;
	*oam++ = 0x0000;
	*oam++ = 0x0000;
	oam++;
	for (i=0; i<127; i++) {
	    *oam++ = 0x0200;
	    *oam++ = 0;
	    *oam++ = 0;
	    oam++;
	}
}
#endif /*SPEED_METER*/

/* select tape image */
void draw_selection_bar(u32 y, u32 color) {
	int i;
	for (i=10; i<30; i++) {
	    putattr(i, y, color);
	}
}
void select_tape(void) {
	u32 i, j;
	u32 num;
	u32 *ofs;
	u16 keya, keyb = 0xffff;
	u8 c;
	u8 *p;
	u8 *keycfgdat = (u8 *)(WorkAreaH + WorkAreaL + KeyCfgAB);
	for (i=0; i<25; i++)
	    for (j=0; j<40; j++)
		put1chr(j, i, 0, 0x70);
	testprint(13, 0, "MAKE READY CMT", 0x70);

	/* count the number of tape image */
	num = 0;
	for (p = (u8 *)TAPEINDEX_START; *p; p += 32) {
	    for (j=0; j<16; j++) {
		c = p[j];
		if (c >= ' ' && c <= '~') put1chr(12+j, num*2+2, asc2disp[c-' '], 0x70);
	    }
	    num++;
	}
	if (num > 12) num = 12;

	/* select tape */
	i = 0;
	draw_selection_bar(2, 2);
	while(1) {
	    WaitForVsync();
	    keya = *KEYS;
	    if ((keyb & KEY_UP) && !(keya & KEY_UP)) {
		draw_selection_bar(i*2+2, 0);
		if (i-- == 0) i = num-1;
		draw_selection_bar(i*2+2, 2);
	    } else if ((keyb & KEY_DOWN) && !(keya & KEY_DOWN)) {
		draw_selection_bar(i*2+2, 0);
		i++;
		if (i>=num) i = 0;
		draw_selection_bar(i*2+2, 2);
	    } else if ( ((keyb & KEY_A) && !(keya & KEY_A)) ||
			((keyb & KEY_B) && !(keya & KEY_B)) ) {
		ofs = (u32 *)TAPEINDEX_START + i*8 + 4;	/* 32=4*8 */
		*tapeptr = (u8 *)TAPEIMG_START + (*ofs);
		ofs += 8;
		*tapeendptr = (u8 *)TAPEIMG_START + (*ofs);
		p = (u8 *)TAPEINDEX_START + i*32 + 20;
		for (j=0; j<10; j++) *keycfgdat++ = *p++;
		return;
	    }
	    keyb = keya;
	}
}

// C֐
__attribute__ ((naked)) void AgbMain(void)
{
#ifdef Z80_DEBUG
	int i;
	int krepcnt = 60;
	u16 k = 0xffff;
	u16 lastkey = 0xffff;
	u32 bkpt1, bkpt2;
	volatile u16 *oam = 0x07000002;
#endif

	vraminit();
	soundinit();

#ifdef SPEED_METER
	init_speedmeter();
#endif /*SPEED_METER*/

	/* load IPL ROM */
	__crt0_move_memory(0x02010000, (u32)IPLROM_START, (u32)IPLROM_END);
	patch_iplrom();

	z80reset();

	config_buttons();

	select_tape();

#ifndef Z80_DEBUG
	z80emu(0);  /* do not return */
#else
	bkpt1 = testprintreg(0x0000004a, &bkpt2);
//	bkpt1 = testprintreg(0x00000db5, &bkpt2);
	while(1) {
	    WaitForVsync();
	    k = *KEYS;
	    *oam = 240;
	    if ((lastkey & KEY_A) && !(k & KEY_A)) {		/* pressed */
		bkpt1 = testprintreg(0x80000000, &bkpt2);	/* traced exec */
		krepcnt = 60;
	    } else if (!(lastkey & KEY_A) && !(k & KEY_A)) {	/* keep pressed */
		if (!(--krepcnt)) {
		    bkpt1 = testprintreg(0x80000000, &bkpt2);	/* traced exec */
		    krepcnt = 2;
		}
	    }
	    if ((lastkey & KEY_B) && !(k & KEY_B)) {		/* pressed */
		bkpt1 = testprintreg(bkpt2, &bkpt2);		/* traced exec */
		krepcnt = 60;
	    }
	    lastkey = k;
	}
#endif
}

void WaitForVsync(void)
{
	while (*(volatile u16*)0x4000006 >= 160) {};
	while (*(volatile u16*)0x4000006 < 160) {};

}

