@--------------------------------------------------------------------
@	Memory Map
@		02000000 - 0200FFFF	DRAM
@		02010000 - 02010FFF	IPL ROM
@		02011000 - 02011FFF	VRAM
@--------------------------------------------------------------------
#include "\devkitadv\crtls\asm-macros.h"
#include "z80em.h"

	.section .iwram,"awx",%progbits

@--------------------------------------------------------------------
@	macro definition
@--------------------------------------------------------------------
	.macro COUNT arg1
	add	r11, r11, #(\arg1 << 16)
	.endm

	.macro	READ_M1 arg1
	add	r9, r9, #f_Rreg_INC
	ldrb	\arg1, [r10], #1
@@@	tst	r10, #0x00010000
@@@	addne	r11, r11, #(1 << 16)	@ ROM wait
	.endm

	.macro	READ_n	arg1
	ldrb	\arg1, [r10], #1
@@@	tst	r10, #0x00010000
@@@	addne	r11, r11, #(1 << 16)	@ ROM wait
	.endm

	.macro	READ_nn arg1
	ldrb	\arg1, [r10], #1
	ldrb	r2, [r10], #1
	orr	\arg1, \arg1, r2, lsl #8
@@@	tst	r10, #0x00010000
@@@	addne	r11, r11, #(2 << 16)	@ ROM wait
	.endm

	.macro	READ_pc
	ldrb	r2,  [r10], #1
	ldrb	r10, [r10]
	orr	r10, r2, r10, lsl #8
	.endm

	.macro	READ_e
	ldrsb	r0, [r10], #1
	add	r10, r0, r10
@@@	tst	r10, #0x00010000
@@@	addne	r11, r11, #(1 << 16)
	.endm

	.macro	READ_d arg1
	ldrsb	\arg1, [r10], #1
@@@	tst	r10, #0x00010000
@@@	addne	r11, r11, #(1 << 16)
	.endm

	.macro	VRAM_WAIT
	mov	r1, r11, lsl #16	@ clear out HBLNK
	cmp	r1, #(62 << 16)		@ if VBLNK period, no-wait
	addhi	r11, r11, #(25 << 16)	@ average cycle time
	.endm

	.macro	MEM_WRITE_CYCLE
	and	r1, r0, #0x0000f000		@ 8
	orr	r1, r1, #0x00010000
	mov	lr, pc
	ldr	pc, [r8, r1, lsr #12-2]
	.endm

	.macro	MEM_READ_CYCLE
	and	r1, r0, #0xf000			@ 7
	mov	lr, pc
	ldr	pc, [r8, r1, lsr #12-2]
	.endm

@====================================================================
FUNCTION(z80reset)
	stmdb	sp!, {r4-r12, lr}

create_parity_table:
	mov	r3,	#ParityFlagTblH
#if ParityFlagTblL
	orr	r3, r3, #ParityFlagTblL
#endif
	mov	r2, #256
1:	mov	r4, #0
	rsbs	r0, r2, #256
	orreq	r4, r4, #FLAG_Z
	tst	r0, #0x0080
	orrne	r4, r4, #FLAG_S
	and	r1, r0, #0xaa
	and	r0, r0, #0x55
	add	r0, r0, r1, lsr #1
	and	r1, r0, #0xcc
	and	r0, r0, #0x33
	add	r0, r0, r1, lsr #2
	and	r1, r0, #0xf0
	and	r0, r0, #0x0f
	add	r0, r0, r1, lsr #4
	tst	r0, #1
	orreq	r4, r4, #FLAG_PV
	strb	r4, [r3], #1
	subs	r2, r2, #1
	bne	1b

create_dec_flag_table:
	mov	r3,	#DecFlagTblH
#if DecFlagTblL
	orr	r3, r3, #DecFlagTblL
#endif
	mov	r2, #256
1:	mov	r4, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	rsb	r0, r2, #256
	tst	r0, #0x0f
	bicne	r4, r4, #FLAG_H
	cmp	r0, #0x80
	bicne	r4, r4, #FLAG_PV
	subs	r0, r0, #1
	bicne	r4, r4, #FLAG_Z
	biccs	r4, r4, #FLAG_Ci
	tst	r0, #0x80
	biceq	r4, r4, #FLAG_S
	strb	r4, [r3], #1
	subs	r2, r2, #1
	bne	1b

create_inc_flag_table:
	mov	r3,	#IncFlagTblH
#if IncFlagTblL
	orr	r3, r3, #IncFlagTblL
#endif
	mov	r2, #256
1:	mov	r4, #0
	rsb	r0, r2, #256
	cmp	r0, #0xff
	orreq	r4, r4, #(FLAG_Z|FLAG_Ci)
	cmp	r0, #0x7f
	orreq	r4, r4, #FLAG_PV
	add	r0, r0, #1
	tst	r0, #0x80
	orrne	r4, r4, #FLAG_S
	tst	r0, #0x0f
	orreq	r4, r4, #FLAG_H
	strb	r4, [r3], #1
	subs	r2, r2, #1
	bne	1b

initialize:
	mov	r8,	#WorkAreaH
	orr	r8, r8, #WorkAreaL

	mov	r0, #0			@ clear entire workarea
	mov	r1, #64-1		@ 256/4
1:	str	r0, [r8, r1, lsl #2]
	subs	r1, r1, #1
	bpl	1b

	mov	r0, #(f_IPLBANK | f_VRAMBANK)	@ default bank settings
	orr	r0, r0, #f_SOUND_STOP		@ no sound
	str	r0, [r8, #saver9]
	mov	r0, #0x0100
	orr	r0, r0, #0x0006			@ 0x0106 = 262
	str	r0, [r8, #saver11]		@ (H count << 16 | V count)

create_mem_access_table:
1:	ldr	r0, [pc, #(init_memr-1b-8)]
1:	ldr	r1, [pc, #(init_memw-1b-8)]
	mov	r2, #16
	mov	r3, r8
1:	str	r1, [r3, #64]
	str	r0, [r3], #4
	subs	r2, r2, #1
	bne	1b
	mov	r0, #0xe4
	bl	io_write_cycle

	bl	init_tempo_timer
	bl	stop_sound

	mvn	r0, #0
	str	r0, [r8, #KeyMatrix  ]
	str	r0, [r8, #KeyMatrix+4]
	str	r0, [r8, #KeyMatrix+8]

	ldmia	sp!, {r4-r12, lr}
	bx	lr

init_memr:	.word	mr_x
init_memw:	.word	mw_x

END(Z80reset)

@====================================================================
@	r6	[31:16] BC reg	[15:0] DE reg
@	r7	[31:24] A  reg	[23:0] 0
@	r12	[31:16] HL reg	[15:0] 0
@	r8	workarea base address	03000200
@	r9	[7:0] Z80 Flag / [24:8] System flags / [31:25] R reg
@	r10	Instruction fetch pointer (modified Z80 PC)
@	r11	[31:16] hcount	[15:0] vcount
@	r13	[31:16] SP reg	[15:0] 8253 Ch1
@
FUNCTION(z80emu)
	stmdb	sp!, {r4-r12, lr}

	mov	r8,	#WorkAreaH
	orr	r8, r8, #WorkAreaL

	str	sp, [r8, #save_sp]

	ldr	r6,  [r8, #reg_DE]
	ldr	r9,  [r8, #saver9]
	ldr	r11, [r8, #saver11]
	ldr	r13, [r8, #saver13]
#ifdef Z80_DEBUG
	mov	r10, r0
	tst	r10, #0x80000000
	bleq	pc2ptr
	str	r10, [r8, #bkpoint]
#endif
	ldrh	r10, [r8, #reg_PC]
	bl	pc2ptr
	ldrb	r7, [r8, #reg_A]
	mov	r7, r7, lsl #24
	ldrh	r12, [r8, #reg_HL]
	mov	r12, r12, lsl #16

	b	inst_start

@--------------------------------------------------------------------
@	exit
@--------------------------------------------------------------------
exit_emulation:
	bl	ptr2pc
	strh	r5,  [r8, #reg_PC]
	mov	r12, r12, lsr #16
	orr	r12, r12, r7
	str	r12, [r8, #reg_HL]
	str	r6,  [r8, #reg_DE]
	str	r9,  [r8, #saver9]
	str	r11, [r8, #saver11]
	str	r13, [r8, #saver13]
	ldr	sp, [r8, #save_sp]
	mov r0, r8

	ldmia	sp!, {r4-r12, lr}
	bx	lr

	.pool

@--------------------------------------------------------------------
@	Z80 Emulation
@--------------------------------------------------------------------
inst_end:
	cmp	r11, #(228 << 16)
	blhs	count_clock

inst_end_ei:

check_breakpoint:
#ifdef Z80_DEBUG
	ldr	r0, [r8, #bkpoint]
	tst	r0, #0x80000000		@ trace mode
	bne	exit_emulation
	cmp	r0, r10			@ break point
	beq	exit_emulation
#endif

inst_start:
	READ_M1	r2
	ldr	pc, [pc, r2, lsl #2]	@ load jump address to PC
	mov	r0, r0			@ NOP
@--------------------------------------------------------------------
@	Instruction table
@--------------------------------------------------------------------
 .word	inst_end,  ld_bc_nn,  ld_bc_a,   inc_bc,    inc_b,     dec_b,     ld_b_n,    rlca_
 .word	ex_af_af,  add_hl_bc, ld_a_bc,   dec_bc,    inc_c,     dec_c,     ld_c_n,    rrca_
 .word	djnz_e,    ld_de_nn,  ld_de_a,   inc_de,    inc_d,     dec_d,     ld_d_n,    rla_
 .word	jr_e,      add_hl_de, ld_a_de,   dec_de,    inc_e,     dec_e,     ld_e_n,    rra_
 .word	jr_nz_e,   ld_hl_nn,  ld_nn_hl_, inc_hl,    inc_h,     dec_h,     ld_h_n,    daa_
 .word	jr_z_e,    add_hl_hl, ld_hl_nn_, dec_hl,    inc_l,     dec_l,     ld_l_n,    cpl_
 .word	jr_nc_e,   ld_sp_nn,  ld_nn_a,   inc_sp,    inc_m,     dec_m,     ld_m_n,    scf_
 .word	jr_c_e,    add_hl_sp, ld_a_nn,   dec_sp,    inc_a,     dec_a,     ld_a_n,    ccf_

 .word	inst_end,  ld_b_c,    ld_b_d,    ld_b_e,    ld_b_h,    ld_b_l,    ld_b_m,    ld_b_a
 .word	ld_c_b,    inst_end,  ld_c_d,    ld_c_e,    ld_c_h,    ld_c_l,    ld_c_m,    ld_c_a
 .word	ld_d_b,    ld_d_c,    inst_end,  ld_d_e,    ld_d_h,    ld_d_l,    ld_d_m,    ld_d_a
 .word	ld_e_b,    ld_e_c,    ld_e_d,    inst_end,  ld_e_h,    ld_e_l,    ld_e_m,    ld_e_a
 .word	ld_h_b,    ld_h_c,    ld_h_d,    ld_h_e,    inst_end,  ld_h_l,    ld_h_m,    ld_h_a
 .word	ld_l_b,    ld_l_c,    ld_l_d,    ld_l_e,    ld_l_h,    inst_end,  ld_l_m,    ld_l_a
 .word	ld_m_b,    ld_m_c,    ld_m_d,    ld_m_e,    ld_m_h,    ld_m_l,    halt_cpu,  ld_m_a
 .word	ld_a_b,    ld_a_c,    ld_a_d,    ld_a_e,    ld_a_h,    ld_a_l,    ld_a_m,    inst_end

 .word	add_a_b,   add_a_c,   add_a_d,   add_a_e,   add_a_h,   add_a_l,   add_a_m,   add_a_a
 .word	adc_a_b,   adc_a_c,   adc_a_d,   adc_a_e,   adc_a_h,   adc_a_l,   adc_a_m,   adc_a_a
 .word	sub_b,     sub_c,     sub_d,     sub_e,     sub_h,     sub_l,     sub_m,     sub_a
 .word	sbc_a_b,   sbc_a_c,   sbc_a_d,   sbc_a_e,   sbc_a_h,   sbc_a_l,   sbc_a_m,   sbc_a_a
 .word	and_b,     and_c,     and_d,     and_e,     and_h,     and_l,     and_m,     and_a
 .word	xor_b,     xor_c,     xor_d,     xor_e,     xor_h,     xor_l,     xor_m,     xor_a
 .word	or_b,      or_c,      or_d,      or_e,      or_h,      or_l,      or_m,      or_a
 .word	cp_b,      cp_c,      cp_d,      cp_e,      cp_h,      cp_l,      cp_m,      cp_a

 .word	ret_nz,    pop_bc,    jp_nz_nn,  jp_nn,     call_nz_nn,push_bc,   add_a_n,   rst_00
 .word	ret_z,     ret_,      jp_z_nn,   cb_pfx,    call_z_nn, call_nn,   adc_a_n,   rst_08
 .word	ret_nc,    pop_de,    jp_nc_nn,  out_n_a,   call_nc_nn,push_de,   sub_n,     rst_10
 .word	ret_c,     exx_,      jp_c_nn,   in_a_n,    call_c_nn, dd_pfx,    sbc_a_n,   rst_18
 .word	ret_po,    pop_hl,    jp_po_nn,  ex_sp_hl,  call_po_nn,push_hl,   and_n,     rst_20
 .word	ret_pe,    jp_hl,     jp_pe_nn,  ex_de_hl,  call_pe_nn,ed_pfx,    xor_n,     rst_28
 .word	ret_p,     pop_af,    jp_p_nn,   di_,       call_p_nn, push_af,   or_n,      rst_30
 .word	ret_m,     ld_sp_hl,  jp_m_nn,   ei_,       call_m_nn, fd_pfx,    cp_n,      rst_38

@--------------------------------------------------------------------
@	count CPU clock cycle
@--------------------------------------------------------------------
count_clock:
	sub	r11, r11, #(228 << 16)	@ Reload HCount

	tst	r9, #f_PIT1_RELOAD	@ 8253 Ch.1 Counter set?
	bne	reload_8253_ch1_w
	mov	r13, r13, ror #16
	sub	r13, r13, #0x00010000	@ Count 8253 Ch.1
	movs	r0, r13, lsr #16	@ 0?
	mov	r13, r13, ror #16
	beq	reload_8253_ch1
count_raster:
	sub	r11, r11, #1
	movs	r0, r11, lsl #16
	bxne	lr
	orr	r11, r11, #0x0100	@ Reload VCount
	orr	r11, r11, #0x0006	@ 262 = 256 + 6
#if 1
	mov	r1,     #0x04000000	@ I/O base
1:	ldrh	r0, [r1, #6]		@ read VCount
	cmp	r0, #114		@ if VCount is in lower half of VTotal
	bhs	1b			@ wait next frame
#endif
#ifdef SPEED_METER
	mov	r2, #0x04000000		@ Speed meter begin
	ldr	r0, [r2, #0x0100]	@ read timer #0
	ldrh	r1, [r8, #lasttime]
	strh	r0, [r8, #lasttime]
	sub	r0, r0, r1
	mov	r0, r0, lsl #16		@ clear out upper 16bits
	mov	r0, r0, lsr #17		@ restore lower 16bits and divide by 4
	cmp	r0, #240		@ Speed meter
	movhi	r0, #240
	sub	r0, r0, #4
	mov	r1, #0x07000000
	strh	r0, [r1, #2]		@ Speed meter end
#endif
#ifdef Z80_DEBUG
	mov	r1, #0x04000000
	orr	r1, r1, #0x0130
	ldrh	r0, [r1]
	tst	r0, #0x0004		@ select
	subeq	r1, r1, #(0x0130-0x68)
	moveq	r0, #0x0080
	streqh	r0, [r1]		@ Stop sound
	beq	exit_emulation
#endif
	bx	lr

reload_8253_ch1:
	ldrh	r0, [r8, #PIT_PRESET1]		@ Reload Ch1
	orr	r13, r13, r0
	ldrb	r1, [r8, #PIT_STT2]		@ Count down Ch.2
	tst	r1, #PIT_RELOAD			@ reload?
	bicne	r1, r1, #PIT_RELOAD
	strneh	r1, [r8, #PIT_STT2]
	ldrneh	r0, [r8, #PIT_PRESET2]
	ldreqh	r0, [r8, #PIT_CNT2]
	subeqs	r0, r0, #1
	strh	r0, [r8, #PIT_CNT2]
	bne	count_raster
	tst	r1, #PIT_INTDONE
	bne	count_raster
	orr	r1, r1, #PIT_INTDONE
	strb	r1, [r8, #PIT_STT2]
	and	r0, r9, #(f_INT_EN|f_INTMSK)
	cmp	r0,     #(f_INT_EN|f_INTMSK)	@ interrupt enabled?
1:	addeq	lr, pc, #(ack_interrupt-1b-8)	@ if enabled, trigger interrupt
	orrne	r9, r9, #f_INT_REQ		@ if disabled, just set REQ flag
	b	count_raster

reload_8253_ch1_w:
	bic	r9, r9, #f_PIT1_RELOAD
	mov	r13, r13, lsr #16		@ save upper 16bit(SP)
	ldrh	r0, [r8, #PIT_PRESET1]		@ Reload Ch1
	orr	r13, r0, r13, lsl #16
	b	count_raster

@--------------------------------------------------------------------
@	interrupt acknowledge
@--------------------------------------------------------------------
i_e_0:	cmp	r11, #(228 << 16)
i_e_i:	.word	0xea000000 + ((ack_interrupt - inst_end - 8) >> 2)	@ b ack_interrupt

ack_interrupt:
1:	ldr	r0, [pc, #-(1b - i_e_0 + 8)]	@ remove the patch
1:	str	r0, [pc, #-(1b - inst_end + 8)]
	bic	r9, r9, #f_INT_EN
	COUNT	1
	@@ fallthru @@

rst_38:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x38
	b	pc2ptr_end

check_interrupt:
	tst	r9, #f_INT_EN			@ interrupt is enabled in CPU?
	tstne	r9, #f_INTMSK			@ and interrupt is enabled by 8255?
	tstne	r9, #f_INT_REQ			@ and interrupt is pending?
1:	ldrne	r0, [pc, #-(1b - i_e_i + 8)]	@ then patch inst_end to ack_interrupt
1:	strne	r0, [pc, #-(1b - inst_end + 8)]
	bicne	r9, r9, #f_INT_REQ
	bx	lr

@--------------------------------------------------------------------
@	Instructions: Group 00 xxx xxx
@--------------------------------------------------------------------
ex_af_af:
	mov	r0, r9, lsl #8		@ xxF0
	orr	r0, r0, r7, lsr #24	@ xxFA
	ldrh	r1, [r8, #reg_FA_]	@ A' F'
	strh	r0, [r8, #reg_FA_]
	mov	r7, r1, lsl #24		@ A' -> A
	bic	r9, r9, #0x00ff
	orr	r9, r9, r1, lsr #8	@ F' -> F
	COUNT	4
	b	inst_end

djnz_e:					@ (B!=0: 5,3,5 / B==0: 5,3)
	COUNT	8
	sub	r6, r6, #0x01000000	@ reg B
	tst	r6,	#0xff000000
	READ_d	r0
	addne	r10, r10, r0
	addne	r11, r11, #(5 << 16)	@ COUNT	5
	b	inst_end

jr_e:					@ (4,3,5)
	COUNT	12
	READ_e
	b	inst_end

jr_nz_e:				@ (4,3,5) if jump / (4,3) if non-jump
	COUNT	7
	tst	r9, #FLAG_Z
	addne	r10, r10, #1		@ if non-jump, inc PC
	bne	inst_end
	READ_e
	COUNT	5
	b	inst_end

jr_z_e:					@ (4,3,5) if jump / (4,3) if non-jump
	COUNT	7
	tst	r9, #FLAG_Z
	addeq	r10, r10, #1		@ if non-jump, inc PC
	beq	inst_end
	READ_e
	COUNT	5
	b	inst_end

jr_nc_e:				@ (4,3,5) if jump / (4,3) if non-jump
	COUNT	7
	tst	r9, #FLAG_Cy
	addne	r10, r10, #1		@ if non-jump, inc PC
	bne	inst_end
	READ_e
	COUNT	5
	b	inst_end

jr_c_e:					@ (4,3,5) if jump / (4,3) if non-jump
	COUNT	7
	tst	r9, #FLAG_Cy
	addeq	r10, r10, #1		@ if non-jump, inc PC
	beq	inst_end
	READ_e
	COUNT	5
	b	inst_end

@--------------------------------------------------------------------
@	LD ss,nn	00 xx0 001
@--------------------------------------------------------------------
ld_bc_nn:
	READ_nn	r0
	bic	r6, r6, #0xff000000
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsl #16
	COUNT	10
	b	inst_end
ld_de_nn:
	READ_nn	r0
	mov	r6, r6, lsr #16
	orr	r6, r0, r6, lsl #16
	COUNT	10
	b	inst_end
ld_hl_nn:
	READ_nn	r12
	mov	r12, r12, lsl #16
	COUNT	10
	b	inst_end
ld_sp_nn:
	READ_nn	r0
	bic	r13, r13, #0xff000000
	bic	r13, r13, #0x00ff0000
	orr	r13, r13, r0, lsl #16
	COUNT	10
	b	inst_end
ld_ix_nn:
	READ_nn	r0
	strh	r0, [r8, #reg_IX]
	COUNT	10
	b	inst_end
ld_iy_nn:
	READ_nn	r0
	strh	r0, [r8, #reg_IY]
	COUNT	10
	b	inst_end

@--------------------------------------------------------------------
@	ADD HL,ss	00 xx1 001
@--------------------------------------------------------------------
	.macro	ADD16_SUB arg2
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)	@ OUT: r12 = r12 + r2
	eor	r0, r12, \arg2
	adds	r12, r12, \arg2
	orrcs	r9, r9, #FLAG_C
	eor	r0, r0, r12
	tst	r0, #0x10000000		@ half-carry
	orrne	r9, r9, #FLAG_H
	.endm

add_hl_bc:				@ (4,4,3)
	mov	r2, r6, lsr #16
	mov	r2, r2, lsl #16
	ADD16_SUB r2
	COUNT	11
	b	inst_end
add_hl_hl:
	ADD16_SUB r12
	COUNT	11
	b	inst_end
add_hl_sp:
	mov	r2, r13, lsr #16
	mov	r2, r2, lsl #16
	b	add_16bit
add_hl_de:
	mov	r2, r6, lsl #16
add_16bit:
	ADD16_SUB r2
	COUNT	11
	b	inst_end

add_ix_bc:				@ (4,4,4,3)
	mov	r2, r6, lsr #16
	mov	r2, r2, lsl #16
	b	add_16bit_ix
add_ix_ix:
	ldrh	r1, [r8, #reg_IX]
	mov	r1, r1, lsl #16
	mov	r2, r1
	b	1f
add_ix_sp:
	mov	r2, r13, lsr #16
	mov	r2, r2, lsl #16
	b	add_16bit_ix
add_ix_de:
	mov	r2, r6, lsl #16
add_16bit_ix:				@ IN : r1, r2 (byte)
	ldrh	r1, [r8, #reg_IX]	@ OUT: r0 = r1 + r2
	mov	r1, r1, lsl #16
1:	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	adds	r0, r1, r2
	orrcs	r9, r9, #FLAG_C
	eor	r1, r1, r2
	eor	r1, r1, r0
	tst	r1, #0x10000000		@ half-carry
	orrne	r9, r9, #FLAG_H
	mov	r0, r0, lsr #16
	strh	r0, [r8, #reg_IX]
	COUNT	15
	b	inst_end

add_iy_bc:				@ (4,4,4,3)
	mov	r2, r6, lsr #16
	mov	r2, r2, lsl #16
	b	add_16bit_iy
add_iy_iy:
	ldrh	r2, [r8, #reg_IY]
	mov	r1, r1, lsl #16
	mov	r2, r1
	b	1f
add_iy_sp:
	mov	r2, r13, lsr #16
	mov	r2, r2, lsl #16
	b	add_16bit_iy
add_iy_de:
	mov	r2, r6, lsl #16
add_16bit_iy:				@ IN : r1, r2 (byte)
	ldrh	r1, [r8, #reg_IY]	@ OUT: r0 = r1 + r2
	mov	r1, r1, lsl #16
1:	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	adds	r0, r1, r2
	orrcs	r9, r9, #FLAG_C
	eor	r1, r1, r2
	eor	r1, r1, r0
	tst	r1, #0x10000000		@ half-carry
	orrne	r9, r9, #FLAG_H
	mov	r0, r0, lsr #16
	strh	r0, [r8, #reg_IY]
	COUNT	15
	b	inst_end

@--------------------------------------------------------------------
@	LD (BC),A etc.	00 0xx 010
@--------------------------------------------------------------------
ld_bc_a:
	mov	r0, r6, lsr #16
	mov	r2, r7, lsr #24
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_a_bc:
	mov	r0, r6, lsr #16
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	7
	b	inst_end
ld_de_a:
	bic	r0, r6, #0xff000000
	bic	r0, r0, #0x00ff0000
	mov	r2, r7, lsr #24
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_a_de:
	bic	r0, r6, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	7
	b	inst_end

@--------------------------------------------------------------------
@	LD (nn),HL etc.	00 1xx 010
@--------------------------------------------------------------------
ld_nn_hl_:
	READ_nn	r0
	add	r5, r0, #1
	mov	r2, r12, lsr #16
	and	r2, r2, #0xff
	MEM_WRITE_CYCLE
	bic	r0, r5, #0x00ff0000
	mov	r2, r12, lsr #24
	MEM_WRITE_CYCLE
	COUNT	16
	b	inst_end
ld_hl_nn_:
	READ_nn	r0
	add	r5, r0, #1
	MEM_READ_CYCLE
	mov	r12, r2, lsl #16
	bic	r0, r5, #0x00ff0000
	MEM_READ_CYCLE
	orr	r12, r12, r2, lsl #24
	COUNT	16
	b	inst_end
ld_nn_a:
	READ_nn	r0
	mov	r2, r7, lsr #24
	MEM_WRITE_CYCLE
	COUNT	13
	b	inst_end
ld_a_nn:
	READ_nn	r0
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	13
	b	inst_end

ld_nn_ix_:
	READ_nn	r0
	add	r5, r0, #1
	ldrh	r4, [r8, #reg_IX]
	and	r2, r4, #0xff
	MEM_WRITE_CYCLE
	bic	r0, r5, #0x00ff0000
	mov	r2, r4, lsr #8
	MEM_WRITE_CYCLE
	COUNT	20
	b	inst_end
ld_ix_nn_:
	READ_nn	r0
	add	r5, r0, #1
	MEM_READ_CYCLE
	mov	r4, r2
	bic	r0, r5, #0x00ff0000
	MEM_READ_CYCLE
	orr	r4, r4, r2, lsl #8
	strh	r4, [r8, #reg_IX]
	COUNT	20
	b	inst_end

ld_nn_iy_:
	READ_nn	r0
	add	r5, r0, #1
	ldrh	r4, [r8, #reg_IY]
	mov	r2, r4
	MEM_WRITE_CYCLE
	bic	r0, r5, #0x00ff0000
	mov	r2, r4, lsr #8
	MEM_WRITE_CYCLE
	COUNT	20
	b	inst_end
ld_iy_nn_:
	READ_nn	r0
	add	r5, r0, #1
	MEM_READ_CYCLE
	mov	r4, r2
	bic	r0, r5, #0x00ff0000
	MEM_READ_CYCLE
	orr	r4, r4, r2, lsl #8
	strh	r4, [r8, #reg_IY]
	COUNT	20
	b	inst_end

@--------------------------------------------------------------------
@	INC ss/DEC ss	00 xxx 011
@--------------------------------------------------------------------
inc_bc:
	add	r6, r6, #0x00010000
	COUNT	6
	b	inst_end
inc_de:
	mov	r6, r6, ror #16
	add	r6, r6, #0x00010000
	mov	r6, r6, ror #16
	COUNT	6
	b	inst_end
inc_hl:
	add	r12, r12, #0x00010000
	COUNT	6
	b	inst_end
inc_sp:
	add	r13, r13, #0x00010000
	COUNT	6
	b	inst_end
inc_ix:
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, #1
	strh	r0, [r8, #reg_IX]
	COUNT	10
	b	inst_end
inc_iy:
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, #1
	strh	r0, [r8, #reg_IY]
	COUNT	10
	b	inst_end

dec_bc:
	sub	r6, r6, #0x00010000
	COUNT	6
	b	inst_end
dec_de:
	mov	r6, r6, ror #16
	sub	r6, r6, #0x00010000
	mov	r6, r6, ror #16
	COUNT	6
	b	inst_end
dec_hl:
	sub	r12, r12, #0x00010000
	COUNT	6
	b	inst_end
dec_sp:
	sub	r13, r13, #0x00010000
	COUNT	6
	b	inst_end
dec_ix:
	ldrh	r0, [r8, #reg_IX]
	sub	r0, r0, #1
	strh	r0, [r8, #reg_IX]
	COUNT	10
	b	inst_end
dec_iy:
	ldrh	r0, [r8, #reg_IY]
	sub	r0, r0, #1
	strh	r0, [r8, #reg_IY]
	COUNT	10
	b	inst_end

@--------------------------------------------------------------------
@	DEC r		00 xxx 101
@--------------------------------------------------------------------
dec_a:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r7, lsr #24]
	orr	r9, r9, r0
	sub	r7, r7, #0x01000000
	COUNT	4
	b	inst_end
dec_b:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	sub	r6, r6, #0x01000000
	COUNT	4
	b	inst_end
dec_c:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r6, r6, ror #24		@ CCDDEEBB
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	sub	r6, r6, #0x01000000
	mov	r6, r6, ror #8		@ BBCCDDEE
	COUNT	4
	b	inst_end
dec_d:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r6, r6, ror #16		@ DDEEBBCC
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	sub	r6, r6, #0x01000000
	mov	r6, r6, ror #16		@ BBCCDDEE
	COUNT	4
	b	inst_end
dec_e:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r6, r6, ror #8		@ EEBBCCDD
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	sub	r6, r6, #0x01000000
	mov	r6, r6, ror #24		@ BBCCDDEE
	COUNT	4
	b	inst_end
dec_h:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r12, lsr #24]
	orr	r9, r9, r0
	sub	r12, r12, #0x01000000
	COUNT	4
	b	inst_end
dec_l:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r12, r12, ror #24	@ LL0000HH
	sub	r0, r8, #(WorkAreaL-DecFlagTblL)
	ldrb	r0, [r0, r12, lsr #24]
	orr	r9, r9, r0
	sub	r12, r12, #0x01000000
	mov	r12, r12, ror #8	@ HHLL0000
	COUNT	4
	b	inst_end

	.macro	DEC_SUB		@ IN: r2 ... val
	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	orr	r0, r2, #DecFlagTblH
	ldrb	r0, [r0, #DecFlagTblL]
	orr	r9, r9, r0
	subs	r2, r2, #1
	.endm
dec_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	DEC_SUB
	movmi	r2, #0xff
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	11
	b	inst_end
dec_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	mov	r3, r0
	MEM_READ_CYCLE
	bl	dec_sub
	movmi	r2, #0xff
	mov	r0, r3
	MEM_WRITE_CYCLE
	COUNT	23
	b	inst_end
dec_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	mov	r3, r0
	MEM_READ_CYCLE
	bl	dec_sub
	movmi	r2, #0xff
	mov	r0, r3
	MEM_WRITE_CYCLE
	COUNT	23
	b	inst_end
dec_ixh:ldrb	r2, [r8, #reg_IXH]
	bl	dec_sub
	strb	r2, [r8, #reg_IXH]
	COUNT	8
	b	inst_end
dec_ixl:ldrb	r2, [r8, #reg_IXL]
	bl	dec_sub
	strb	r2, [r8, #reg_IXL]
	COUNT	8
	b	inst_end
dec_iyh:ldrb	r2, [r8, #reg_IYH]
	bl	dec_sub
	strb	r2, [r8, #reg_IYH]
	COUNT	8
	b	inst_end
dec_iyl:ldrb	r2, [r8, #reg_IYL]
	bl	dec_sub
	strb	r2, [r8, #reg_IYL]
	COUNT	8
	b	inst_end

dec_sub:				@ IN: r2 ... val
	DEC_SUB
	bx	lr

@--------------------------------------------------------------------
@	INC r		00 xxx 100
@--------------------------------------------------------------------
inc_a:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r7, lsr #24]
	orr	r9, r9, r0
	add	r7, r7, #0x01000000
	COUNT	4
	b	inst_end
inc_b:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	add	r6, r6, #0x01000000
	COUNT	4
	b	inst_end
inc_c:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r6, r6, ror #24		@ CCDDEEBB
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	add	r6, r6, #0x01000000
	mov	r6, r6, ror #8		@ BBCCDDEE
	COUNT	4
	b	inst_end
inc_d:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r6, r6, ror #16		@ DDEEBBCC
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	add	r6, r6, #0x01000000
	mov	r6, r6, ror #16		@ BBCCDDEE
	COUNT	4
	b	inst_end
inc_e:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r6, r6, ror #8		@ EEBBCCDD
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r6, lsr #24]
	orr	r9, r9, r0
	add	r6, r6, #0x01000000
	mov	r6, r6, ror #24		@ BBCCDDEE
	COUNT	4
	b	inst_end
inc_h:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r12, lsr #24]
	orr	r9, r9, r0
	add	r12, r12, #0x01000000
	COUNT	4
	b	inst_end
inc_l:	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	mov	r12, r12, ror #24	@ LL0000HH
	add	r0, r8, #(IncFlagTblL-WorkAreaL)
	ldrb	r0, [r0, r12, lsr #24]
	orr	r9, r9, r0
	add	r12, r12, #0x01000000
	mov	r12, r12, ror #8	@ HHLL0000
	COUNT	4
	b	inst_end

	.macro	INC_SUB
	bic	r9, r9, #(FLAG_Ci|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	orr	r0, r2, #IncFlagTblH
	ldrb	r0, [r0, #IncFlagTblL]
	orr	r9, r9, r0
	add	r2, r2, #1
	.endm
inc_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	INC_SUB
	and	r2, r2, #0xff
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	11
	b	inst_end
inc_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	mov	r3, r0
	MEM_READ_CYCLE
	bl	inc_sub
	and	r2, r2, #0xff
	mov	r0, r3
	MEM_WRITE_CYCLE
	COUNT	23
	b	inst_end
inc_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	mov	r3, r0
	MEM_READ_CYCLE
	bl	inc_sub
	and	r2, r2, #0xff
	mov	r0, r3
	MEM_WRITE_CYCLE
	COUNT	23
	b	inst_end
inc_ixh:ldrb	r2, [r8, #reg_IXH]
	bl	inc_sub
	strb	r2, [r8, #reg_IXH]
	COUNT	8
	b	inst_end
inc_ixl:ldrb	r2, [r8, #reg_IXL]
	bl	inc_sub
	strb	r2, [r8, #reg_IXL]
	COUNT	8
	b	inst_end
inc_iyh:ldrb	r2, [r8, #reg_IYH]
	bl	inc_sub
	strb	r2, [r8, #reg_IYH]
	COUNT	8
	b	inst_end
inc_iyl:ldrb	r2, [r8, #reg_IYL]
	bl	inc_sub
	strb	r2, [r8, #reg_IYL]
	COUNT	8
	b	inst_end

inc_sub:				@ IN: r2 ... val
	INC_SUB
	bx	lr

@--------------------------------------------------------------------
@	LD r,n		00 xxx 110
@--------------------------------------------------------------------
ld_a_n:
	READ_n	r7
	mov	r7, r7, lsl #24
	COUNT	7
	b	inst_end
ld_b_n:
	READ_n	r2
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	7
	b	inst_end
ld_c_n:
	READ_n	r2
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	7
	b	inst_end
ld_d_n:
	READ_n	r2
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	7
	b	inst_end
ld_e_n:
	READ_n	r2
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	7
	b	inst_end
ld_h_n:
	READ_n	r2
	and	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #24
	COUNT	7
	b	inst_end
ld_l_n:
	READ_n	r2
	and	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #16
	COUNT	7
	b	inst_end
ld_m_n:
	READ_n	r2
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	10
	b	inst_end
ld_x_n:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	READ_n	r2
	MEM_WRITE_CYCLE
	COUNT	19
	b	inst_end
ld_y_n:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	READ_n	r2
	MEM_WRITE_CYCLE
	COUNT	19
	b	inst_end
ld_ixh_n:
	READ_n	r2
	strb	r2, [r8, #reg_IXH]
	COUNT	11
	b	inst_end
ld_ixl_n:
	READ_n	r2
	strb	r2, [r8, #reg_IXL]
	COUNT	11
	b	inst_end
ld_iyh_n:
	READ_n	r2
	strb	r2, [r8, #reg_IYH]
	COUNT	11
	b	inst_end
ld_iyl_n:
	READ_n	r2
	strb	r2, [r8, #reg_IYL]
	COUNT	11
	b	inst_end

@--------------------------------------------------------------------
@	RLCA etc.	00 xxx 111
@--------------------------------------------------------------------
rlca_:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsl #1
	orrcs	r7, r7, #0x01000000
	orrcs	r9, r9, #FLAG_C
	COUNT	4
	b	inst_end
rrca_:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsr #25
	orrcs	r7, r7, #0x80
	orrcs	r9, r9, #FLAG_C
	mov	r7, r7, lsl #24
	COUNT	4
	b	inst_end
rla_:
	tst	r9, #FLAG_Cy
	orrne	r7, r7, #0x00800000
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsl #1
	orrcs	r9, r9, #FLAG_C
	COUNT	4
	b	inst_end
rra_:
	and	r1, r9, #FLAG_Cy
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsr #25
	orr	r7, r7, r1, lsl #7
	orrcs	r9, r9, #FLAG_C
	mov	r7, r7, lsl #24
	COUNT	4
	b	inst_end

cpl_:
	orr	r9, r9, #(FLAG_N|FLAG_H)
	eor	r7, r7, #0xff000000
	COUNT	4
	b	inst_end

ccf_:	tst	r9, #FLAG_Cy
	orrne	r9, r9, #FLAG_H
	biceq	r9, r9, #FLAG_H
	eor	r9, r9, #FLAG_C
	COUNT	4
	b	inst_end

scf_:	orr	r9, r9, #FLAG_C
	COUNT	4
	b	inst_end

daa_:
	mov	r0, #0
	tst	r9, #FLAG_N
	orrne	r0, r0, #4
	tst	r9, #FLAG_Ci
	orrne	r0, r0, #2
	tst	r9, #FLAG_H
	orrne	r0, r0, #1
	ldrb	r0, [pc, r0]
	add	pc, pc, r0, lsl #2	@ == bl (r1+(r2<<2))
daa_tbl:
	.byte	((daa_0-daa_tbl)>>2)-1
	.byte	((daa_1-daa_tbl)>>2)-1
	.byte	((daa_2-daa_tbl)>>2)-1
	.byte	((daa_3-daa_tbl)>>2)-1
	.byte	((daa_4-daa_tbl)>>2)-1
	.byte	((daa_5-daa_tbl)>>2)-1
	.byte	((daa_6-daa_tbl)>>2)-1
	.byte	((daa_7-daa_tbl)>>2)-1

daa_0:	and	r0, r7, #0x0f000000
	cmp	r0, #0x0a000000
	bcs	1f
	and	r0, r7, #0xf0000000
	cmp	r0, #0xa0000000
	movcc	r2, #0x00000000
	movcs	r2, #0x60000000
	movcc	r1, #0
	movcs	r1, #FLAG_C
	b	daa_sub
1:	and	r0, r7, #0xf0000000
	cmp	r0, #0x90000000
	movcc	r2, #0x06000000
	movcs	r2, #0x66000000
	movcc	r1, #0
	movcs	r1, #FLAG_C
	b	daa_sub
daa_1:	and	r0, r7, #0xf0000000
	cmp	r0, #0xa0000000
	movcc	r2, #0x06000000
	movcs	r2, #0x66000000
	movcc	r1, #0
	movcs	r1, #FLAG_C
	b	daa_sub
daa_2:	and	r0, r7, #0x0f000000
	cmp	r0, #0x0a000000
	movcc	r2, #0x60000000
	movcs	r2, #0x66000000
	mov	r1, #FLAG_C
	b	daa_sub
daa_3:	mov	r2, #0x66000000
	mov	r1, #FLAG_C
	b	daa_sub
daa_4:	mov	r2, #0x00000000
	mov	r1, #0
	b	daa_sub
daa_5:	mov	r2, #0xfa000000
	mov	r1, #0
	b	daa_sub
daa_6:	mov	r2, #0xa0000000
	mov	r1, #FLAG_C
	b	daa_sub
daa_7:	mov	r2, #0x9a000000
	mov	r1, #FLAG_C
daa_sub:				@ IN : r7, r2 (byte)
	bic	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_H)
	orr	r9, r9, r1		@ Cy flag
	eor	r0, r7, r2
	adds	r7, r7, r2
	orrmi	r9, r9, #FLAG_S
	eor	r1, r0, r7
	tst	r1, #0x10000000
	orrne	r9, r9, #FLAG_H
flag_parity:
	sub	r0, r8, #(WorkAreaL-ParityFlagTblL)	@ parity flag table
	ldrb	r0, [r0, r7, lsr #24]
	and	r0, r0, #(FLAG_Z|FLAG_PV)
	orr	r9, r9, r0
	COUNT	4
	b	inst_end

@--------------------------------------------------------------------
@	LD r,r'		01 xxx xxx
@--------------------------------------------------------------------
ld_b_c:	bic	r6, r6, #0xff000000
	and	r0, r6, #0x00ff0000
	orr	r6, r6, r0, lsl #8
	COUNT	4
	b	inst_end
ld_b_d:	bic	r6, r6, #0xff000000
	and	r0, r6, #0x0000ff00
	orr	r6, r6, r0, lsl #16
	COUNT	4
	b	inst_end
ld_b_e:	bic	r6, r6, #0xff000000
	orr	r6, r6, r6, lsl #24
	COUNT	4
	b	inst_end
ld_b_h:	bic	r6, r6,  #0xff000000
	and	r0, r12, #0xff000000
	orr	r6, r6, r0
	COUNT	4
	b	inst_end
ld_b_l:	bic	r6, r6,  #0xff000000
	and	r0, r12, #0x00ff0000
	orr	r6, r6, r0, lsl #8
	COUNT	4
	b	inst_end
ld_b_a:	bic	r6, r6, #0xff000000
	orr	r6, r6, r7
	COUNT	4
	b	inst_end
ld_b_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	7
	b	inst_end
ld_b_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	19
	b	inst_end
ld_b_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	19
	b	inst_end
ld_b_ixh:
	ldrb	r0, [r8, #reg_IXH]
	bic	r6, r6, #0xff000000
	orr	r6, r6, r0, lsl #24
	COUNT	8
	b	inst_end
ld_b_ixl:
	ldrb	r0, [r8, #reg_IXL]
	bic	r6, r6, #0xff000000
	orr	r6, r6, r0, lsl #24
	COUNT	8
	b	inst_end
ld_b_iyh:
	ldrb	r0, [r8, #reg_IYH]
	bic	r6, r6, #0xff000000
	orr	r6, r6, r0, lsl #24
	COUNT	8
	b	inst_end
ld_b_iyl:
	ldrb	r0, [r8, #reg_IYL]
	bic	r6, r6, #0xff000000
	orr	r6, r6, r0, lsl #24
	COUNT	8
	b	inst_end

ld_c_b:	bic	r6, r6, #0x00ff0000
	and	r0, r6, #0xff000000
	orr	r6, r6, r0, lsr #8
	COUNT	4
	b	inst_end
ld_c_d:	bic	r6, r6, #0x00ff0000
	and	r0, r6, #0x0000ff00
	orr	r6, r6, r0, lsl #8
	COUNT	4
	b	inst_end
ld_c_e:	bic	r6, r6, #0x00ff0000
	and	r0, r6, #0x000000ff
	orr	r6, r6, r0, lsl #16
	COUNT	4
	b	inst_end
ld_c_h:	bic	r6, r6,  #0x00ff0000
	and	r0, r12, #0xff000000
	orr	r6, r6, r0, lsr #8
	COUNT	4
	b	inst_end
ld_c_l:	bic	r6, r6,  #0x00ff0000
	and	r0, r12, #0x00ff0000
	orr	r6, r6, r0
	COUNT	4
	b	inst_end
ld_c_a:	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r7, lsr #8
	COUNT	4
	b	inst_end
ld_c_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	7
	b	inst_end
ld_c_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	19
	b	inst_end
ld_c_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	19
	b	inst_end
ld_c_ixh:
	ldrb	r0, [r8, #reg_IXH]
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsl #16
	COUNT	8
	b	inst_end
ld_c_ixl:
	ldrb	r0, [r8, #reg_IXL]
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsl #16
	COUNT	8
	b	inst_end
ld_c_iyh:
	ldrb	r0, [r8, #reg_IYH]
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsl #16
	COUNT	8
	b	inst_end
ld_c_iyl:
	ldrb	r0, [r8, #reg_IYL]
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsl #16
	COUNT	8
	b	inst_end

ld_d_b:	bic	r6, r6, #0x0000ff00
	and	r0, r6, #0xff000000
	orr	r6, r6, r0, lsr #16
	COUNT	4
	b	inst_end
ld_d_c:	bic	r6, r6, #0x0000ff00
	and	r0, r6, #0x00ff0000
	orr	r6, r6, r0, lsr #8
	COUNT	4
	b	inst_end
ld_d_e:	bic	r6, r6, #0x0000ff00
	and	r0, r6, #0x000000ff
	orr	r6, r6, r0, lsl #8
	COUNT	4
	b	inst_end
ld_d_h:	bic	r6, r6,  #0x0000ff00
	and	r0, r12, #0xff000000
	orr	r6, r6, r0, lsr #16
	COUNT	4
	b	inst_end
ld_d_l:	bic	r6, r6,  #0x0000ff00
	and	r0, r12, #0x00ff0000
	orr	r6, r6, r0, lsr #8
	COUNT	4
	b	inst_end
ld_d_a:	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r7, lsr #16
	COUNT	4
	b	inst_end
ld_d_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	7
	b	inst_end
ld_d_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	19
	b	inst_end
ld_d_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	19
	b	inst_end
ld_d_ixh:
	ldrb	r0, [r8, #reg_IXH]
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsl #8
	COUNT	8
	b	inst_end
ld_d_ixl:
	ldrb	r0, [r8, #reg_IXL]
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsl #8
	COUNT	8
	b	inst_end
ld_d_iyh:
	ldrb	r0, [r8, #reg_IYH]
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsl #8
	COUNT	8
	b	inst_end
ld_d_iyl:
	ldrb	r0, [r8, #reg_IYL]
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsl #8
	COUNT	8
	b	inst_end

ld_e_b:	bic	r6, r6, #0x000000ff
	orr	r6, r6, r6, lsr #24
	COUNT	4
	b	inst_end
ld_e_c:	bic	r6, r6, #0x000000ff
	and	r0, r6, #0x00ff0000
	orr	r6, r6, r0, lsr #16
	COUNT	4
	b	inst_end
ld_e_d:	bic	r6, r6, #0x000000ff
	and	r0, r6, #0x0000ff00
	orr	r6, r6, r0, lsr #8
	COUNT	4
	b	inst_end
ld_e_h:	bic	r6, r6, #0x000000ff
	orr	r6, r6, r12, lsr #24
	COUNT	4
	b	inst_end
ld_e_l:	bic	r6, r6,  #0x000000ff
	and	r0, r12, #0x00ff0000
	orr	r6, r6, r0, lsr #16
	COUNT	4
	b	inst_end
ld_e_a:	bic	r6, r6, #0x000000ff
	orr	r6, r6, r7, lsr #24
	COUNT	4
	b	inst_end
ld_e_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	7
	b	inst_end
ld_e_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	19
	b	inst_end
ld_e_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	19
	b	inst_end
ld_e_ixh:
	ldrb	r0, [r8, #reg_IXH]
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0
	COUNT	8
	b	inst_end
ld_e_ixl:
	ldrb	r0, [r8, #reg_IXL]
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0
	COUNT	8
	b	inst_end
ld_e_iyh:
	ldrb	r0, [r8, #reg_IYH]
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0
	COUNT	8
	b	inst_end
ld_e_iyl:
	ldrb	r0, [r8, #reg_IYL]
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0
	COUNT	8
	b	inst_end

ld_h_b:	bic	r12, r12, #0xff000000
	and	r0,  r6,  #0xff000000
	orr	r12, r12, r0
	COUNT	4
	b	inst_end
ld_h_c:	bic	r12, r12, #0xff000000
	and	r0,  r6,  #0x00ff0000
	orr	r12, r12, r0, lsl #8
	COUNT	4
	b	inst_end
ld_h_d:	bic	r12, r12, #0xff000000
	and	r0,  r6,  #0x0000ff00
	orr	r12, r12, r0, lsl #16
	COUNT	4
	b	inst_end
ld_h_e:	bic	r12, r12, #0xff000000
	orr	r12, r12, r6, lsl #24
	COUNT	4
	b	inst_end
ld_h_l:	bic	r12, r12, #0xff000000
	orr	r12, r12, r12, lsl #8
	COUNT	4
	b	inst_end
ld_h_a:	bic	r12, r12, #0xff000000
	orr	r12, r12, r7
	COUNT	4
	b	inst_end
ld_h_m:	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	7
	b	inst_end
ld_h_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	19
	b	inst_end
ld_h_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	19
	b	inst_end

ld_l_b:	bic	r12, r12, #0x00ff0000
	and	r0,  r6,  #0xff000000
	orr	r12, r12, r0, lsr #8
	COUNT	4
	b	inst_end
ld_l_c:	bic	r12, r12, #0x00ff0000
	and	r0,  r6,  #0x00ff0000
	orr	r12, r12, r0
	COUNT	4
	b	inst_end
ld_l_d:	bic	r12, r12, #0x00ff0000
	and	r0,  r6,  #0x0000ff00
	orr	r12, r12, r0, lsl #8
	COUNT	4
	b	inst_end
ld_l_e:	bic	r12, r12, #0x00ff0000
	and	r0,  r6,  #0x000000ff
	orr	r12, r12, r0, lsl #16
	COUNT	4
	b	inst_end
ld_l_h:	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r12, lsr #8
	COUNT	4
	b	inst_end
ld_l_a:	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r7, lsr #8
	COUNT	4
	b	inst_end
ld_l_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	7
	b	inst_end
ld_l_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	19
	b	inst_end
ld_l_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	19
	b	inst_end

ld_a_b:	and	r7, r6, #0xff000000
	COUNT	4
	b	inst_end
ld_a_c:	and	r7, r6, #0x00ff0000
	mov	r7, r7, lsl #8
	COUNT	4
	b	inst_end
ld_a_d:	and	r7, r6, #0x0000ff00
	mov	r7, r7, lsl #16
	COUNT	4
	b	inst_end
ld_a_e:	mov	r7, r6, lsl #24
	COUNT	4
	b	inst_end
ld_a_h:	and	r7, r12, #0xff000000
	COUNT	4
	b	inst_end
ld_a_l:	mov	r7, r12, lsl #8
	COUNT	4
	b	inst_end
ld_a_m:	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	7
	b	inst_end
ld_a_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	19
	b	inst_end
ld_a_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	19
	b	inst_end
ld_a_ixh:
	ldrb	r7, [r8, #reg_IXH]
	mov	r7, r7, lsl #24
	COUNT	8
	b	inst_end
ld_a_ixl:
	ldrb	r7, [r8, #reg_IXL]
	mov	r7, r7, lsl #24
	COUNT	8
	b	inst_end
ld_a_iyh:
	ldrb	r7, [r8, #reg_IYH]
	mov	r7, r7, lsl #24
	COUNT	8
	b	inst_end
ld_a_iyl:
	ldrb	r7, [r8, #reg_IYL]
	mov	r7, r7, lsl #24
	COUNT	8
	b	inst_end

ld_m_a:	mov	r2, r7,  lsr #24
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_m_b:	mov	r2, r6,  lsr #24
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_m_c:	and	r2, r6, #0x00ff0000
	mov	r2, r2,  lsr #16
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_m_d:	and	r2, r6, #0x0000ff00
	mov	r2, r2,  lsr #8
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_m_e:	and	r2, r6, #0x000000ff
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_m_h:	mov	r2, r12, lsr #24
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end
ld_m_l:	and	r2, r12, #0x00ff0000
	mov	r2, r2,  lsr #16
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	7
	b	inst_end

ld_x_b:	mov	r3, r6, lsr #24
	b	1f
ld_x_c:	mov	r3, r6, lsr #16
	b	1f
ld_x_d:	mov	r3, r6, lsr #8
	b	1f
ld_x_e:	mov	r3, r6
	b	1f
ld_x_h:	mov	r3, r12, lsr #24
	b	1f
ld_x_l:	mov	r3, r12, lsr #16
	b	1f
ld_x_a:	mov	r3, r7, lsr #24
1:	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	and	r2, r3, #0xff
	MEM_WRITE_CYCLE
	COUNT	19
	b	inst_end

ld_y_b:	mov	r3, r6, lsr #24
	b	1f
ld_y_c:	mov	r3, r6, lsr #16
	b	1f
ld_y_d:	mov	r3, r6, lsr #8
	b	1f
ld_y_e:	mov	r3, r6
	b	1f
ld_y_h:	mov	r3, r12, lsr #24
	b	1f
ld_y_l:	mov	r3, r12, lsr #16
	b	1f
ld_y_a:	mov	r3, r7, lsr #24
1:	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	and	r2, r3, #0xff
	MEM_WRITE_CYCLE
	COUNT	19
	b	inst_end

ld_ixh_b:	mov	r0, r6, lsr #24
		strb	r0, [r8, #reg_IXH]
		COUNT	8
		b	inst_end
ld_ixh_c:	mov	r0, r6, lsr #16
		strb	r0, [r8, #reg_IXH]
		COUNT	8
		b	inst_end
ld_ixh_d:	mov	r0, r6, lsr #8
		strb	r0, [r8, #reg_IXH]
		COUNT	8
		b	inst_end
ld_ixh_e:	strb	r6, [r8, #reg_IXH]
		COUNT	8
		b	inst_end
ld_ixh_ixl:	ldrb	r0, [r8, #reg_IXL]
		strb	r0, [r8, #reg_IXH]
		COUNT	8
		b	inst_end
ld_ixh_a:	mov	r0, r7, lsr #24
		strb	r0, [r8, #reg_IXH]
		COUNT	8
		b	inst_end

ld_ixl_b:	mov	r0, r6, lsr #24
		strb	r0, [r8, #reg_IXL]
		COUNT	8
		b	inst_end
ld_ixl_c:	mov	r0, r6, lsr #16
		strb	r0, [r8, #reg_IXL]
		COUNT	8
		b	inst_end
ld_ixl_d:	mov	r0, r6, lsr #8
		strb	r0, [r8, #reg_IXL]
		COUNT	8
		b	inst_end
ld_ixl_e:	strb	r6, [r8, #reg_IXL]
		COUNT	8
		b	inst_end
ld_ixl_ixh:	ldrb	r0, [r8, #reg_IXH]
		strb	r0, [r8, #reg_IXL]
		COUNT	8
		b	inst_end
ld_ixl_a:	mov	r0, r7, lsr #24
		strb	r0, [r8, #reg_IXL]
		COUNT	8
		b	inst_end

ld_iyh_b:	mov	r0, r6, lsr #24
		strb	r0, [r8, #reg_IYH]
		COUNT	8
		b	inst_end
ld_iyh_c:	mov	r0, r6, lsr #16
		strb	r0, [r8, #reg_IYH]
		COUNT	8
		b	inst_end
ld_iyh_d:	mov	r0, r6, lsr #8
		strb	r0, [r8, #reg_IYH]
		COUNT	8
		b	inst_end
ld_iyh_e:	strb	r6, [r8, #reg_IYH]
		COUNT	8
		b	inst_end
ld_iyh_iyl:	ldrb	r0, [r8, #reg_IYL]
		strb	r0, [r8, #reg_IYH]
		COUNT	8
		b	inst_end
ld_iyh_a:	mov	r0, r7, lsr #24
		strb	r0, [r8, #reg_IYH]
		COUNT	8
		b	inst_end

ld_iyl_b:	mov	r0, r6, lsr #24
		strb	r0, [r8, #reg_IYL]
		COUNT	8
		b	inst_end
ld_iyl_c:	mov	r0, r6, lsr #16
		strb	r0, [r8, #reg_IYL]
		COUNT	8
		b	inst_end
ld_iyl_d:	mov	r0, r6, lsr #8
		strb	r0, [r8, #reg_IYL]
		COUNT	8
		b	inst_end
ld_iyl_e:	strb	r6, [r8, #reg_IYL]
		COUNT	8
		b	inst_end
ld_iyl_iyh:	ldrb	r0, [r8, #reg_IYH]
		strb	r0, [r8, #reg_IYL]
		COUNT	8
		b	inst_end
ld_iyl_a:	mov	r0, r7, lsr #24
		strb	r0, [r8, #reg_IYL]
		COUNT	8
		b	inst_end

halt_cpu:
	tst	r9,     #f_INT_EN	@ INT_EN is 1?
	tstne	r9,	#f_INTMSK	@ and INTMSK is 1?
	tstne	r9,	#f_INT_REQ	@ and an interrupt is pending?
	subeq	r10, r10, #1		@ if no interrupt, repeat HALT
	COUNT	4
	b	inst_end

@--------------------------------------------------------------------
@	ADD A,r		10 000 xxx
@--------------------------------------------------------------------
	.macro	ADD8_SUB arg2
	bic	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r1, r7, \arg2
	adds	r7, r7, \arg2
	orreq	r9, r9, #FLAG_Z
	orrcs	r9, r9, #FLAG_C
	orrmi	r9, r9, #FLAG_S
	orrvs	r9, r9, #FLAG_PV
	eor	r1, r7, r1
	tst	r1, #0x10000000		@ Half Cy = r1[4] ^ (r1+r2)[4] ^ r2[4]
	orrne	r9, r9, #FLAG_H
	.endm

add_a_a:
	COUNT	4
	ADD8_SUB r7
	b	inst_end
add_a_b:
	and	r0, r6, #0xff000000
	COUNT	4
	ADD8_SUB r0
	b	inst_end
add_a_c:
	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsl #8
	COUNT	4
	ADD8_SUB r0
	b	inst_end
add_a_d:
	and	r0, r6, #0x0000ff00
	mov	r0, r0, lsl #16
	COUNT	4
	ADD8_SUB r0
	b	inst_end
add_a_e:
	mov	r0, r6, lsl #24
	COUNT	4
	ADD8_SUB r0
	b	inst_end
add_a_h:
	and	r0, r12, #0xff000000
	COUNT	4
	ADD8_SUB r0
	b	inst_end
add_a_l:
	mov	r0, r12, lsl #8
	COUNT	4
	ADD8_SUB r0
	b	inst_end
add_a_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	COUNT	7
	mov	r0, r2, lsl #24
	ADD8_SUB r0
	b	inst_end
add_a_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	add_8bit
add_a_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	add_8bit
add_a_ixh:
	ldrb	r2, [r8, #reg_IXH]
	COUNT	8
	b	add_8bit
add_a_ixl:
	ldrb	r2, [r8, #reg_IXL]
	COUNT	8
	b	add_8bit
add_a_iyh:
	ldrb	r2, [r8, #reg_IYH]
	COUNT	8
	b	add_8bit
add_a_iyl:
	ldrb	r2, [r8, #reg_IYL]
	COUNT	8
	b	add_8bit

add_a_n:
	READ_n	r2
	COUNT	7
	@@ fallthru @@
add_8bit:				@ IN : r7, r2 (byte)
	mov	r0, r2, lsl #24
	ADD8_SUB r0
	b	inst_end

@--------------------------------------------------------------------
@	ADC A,r		10 001 xxx
@--------------------------------------------------------------------
adc_a_a:
	mov	r2, r7, lsr #24
	COUNT	4
	b	adc_8bit
adc_a_b:
	mov	r2, r6, lsr #24
	COUNT	4
	b	adc_8bit
adc_a_c:
	mov	r2, r6, lsr #16
	COUNT	4
	b	adc_8bit
adc_a_d:
	mov	r2, r6, lsr #8
	COUNT	4
	b	adc_8bit
adc_a_e:
	mov	r2, r6
	COUNT	4
	b	adc_8bit
adc_a_h:
	mov	r2, r12, lsr #24
	COUNT	4
	b	adc_8bit
adc_a_l:
	mov	r2, r12, lsr #16
	COUNT	4
	b	adc_8bit
adc_a_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	COUNT	7
	b	adc_8bit
adc_a_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	adc_8bit
adc_a_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	adc_8bit
adc_a_ixh:
	ldrb	r2, [r8, #reg_IXH]
	COUNT	8
	b	adc_8bit
adc_a_ixl:
	ldrb	r2, [r8, #reg_IXL]
	COUNT	8
	b	adc_8bit
adc_a_iyh:
	ldrb	r2, [r8, #reg_IYH]
	COUNT	8
	b	adc_8bit
adc_a_iyl:
	ldrb	r2, [r8, #reg_IYL]
	COUNT	8
	b	adc_8bit

adc_a_n:
	READ_n	r2
	COUNT	7
	@@ fallthru @@

adc_8bit:				@ IN : r7, r2 (byte)
	mov	r2, r2, lsl #24
	tst	r9, #FLAG_Cy
	orrne	r7, r7, #0x00800000	@ carry-in
	orrne	r2, r2, #0x00800000	@ carry-in
	bic	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r1, r7, r2
	adds	r7, r7, r2
	orreq	r9, r9, #FLAG_Z
	orrcs	r9, r9, #FLAG_C
	orrmi	r9, r9, #FLAG_S
	orrvs	r9, r9, #FLAG_PV
	eor	r1, r7, r1
	tst	r1, #0x10000000		@ Half Cy = r1[4] ^ (r1+r2)[4] ^ r2[4]
	orrne	r9, r9, #FLAG_H
	b	inst_end

@--------------------------------------------------------------------
@	SUB r		10 010 xxx
@--------------------------------------------------------------------
sub_a:
	mov	r2, r7, lsr #24
	COUNT	4
	b	sub_8bit
sub_b:
	mov	r2, r6, lsr #24
	COUNT	4
	b	sub_8bit
sub_c:
	mov	r2, r6, lsr #16
	COUNT	4
	b	sub_8bit
sub_d:
	mov	r2, r6, lsr #8
	COUNT	4
	b	sub_8bit
sub_e:
	mov	r2, r6
	COUNT	4
	b	sub_8bit
sub_h:
	mov	r2, r12, lsr #24
	COUNT	4
	b	sub_8bit
sub_l:
	mov	r2, r12, lsr #16
	COUNT	4
	b	sub_8bit
sub_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	COUNT	7
	b	sub_8bit
sub_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	sub_8bit
sub_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	sub_8bit
sub_ixh:
	ldrb	r2, [r8, #reg_IXH]
	COUNT	8
	b	sub_8bit
sub_ixl:
	ldrb	r2, [r8, #reg_IXL]
	COUNT	8
	b	sub_8bit
sub_iyh:
	ldrb	r2, [r8, #reg_IYH]
	COUNT	8
	b	sub_8bit
sub_iyl:
	ldrb	r2, [r8, #reg_IYL]
	COUNT	8
	b	sub_8bit

sub_n:
	READ_n	r2
	COUNT	7
	@@ fallthru @@

sub_8bit:				@ IN : r7, r2 (byte)
	mov	r2, r2, lsl #24		@ OUT: r7 = r7 - r2
	orr	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r1, r7, r2
	subs	r7, r7, r2
	biccs	r9, r9, #FLAG_C		@ (negated)
	bicpl	r9, r9, #FLAG_S
	bicvc	r9, r9, #FLAG_PV
	bicne	r9, r9, #FLAG_Z
	eor	r1, r7, r1		@ Half Cy = r1[4] ^ (r1+r2)[4] ^ r2[4]
	tst	r1, #0x10000000
	biceq	r9, r9, #FLAG_H		@ (negated)
	b	inst_end

@--------------------------------------------------------------------
@	SBC A,r		10 011 xxx
@--------------------------------------------------------------------
sbc_a_a:
	mov	r2, r7, lsr #24
	COUNT	4
	b	sbc_8bit
sbc_a_b:
	mov	r2, r6, lsr #24
	COUNT	4
	b	sbc_8bit
sbc_a_c:
	mov	r2, r6, lsr #16
	COUNT	4
	b	sbc_8bit
sbc_a_d:
	mov	r2, r6, lsr #8
	COUNT	4
	b	sbc_8bit
sbc_a_e:
	mov	r2, r6
	COUNT	4
	b	sbc_8bit
sbc_a_h:
	mov	r2, r12, lsr #24
	COUNT	4
	b	sbc_8bit
sbc_a_l:
	mov	r2, r12, lsr #16
	COUNT	4
	b	sbc_8bit
sbc_a_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	COUNT	7
	b	sbc_8bit
sbc_a_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	sbc_8bit
sbc_a_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	b	sbc_8bit
sbc_a_ixh:
	ldrb	r2, [r8, #reg_IXH]
	COUNT	8
	b	sbc_8bit
sbc_a_ixl:
	ldrb	r2, [r8, #reg_IXL]
	COUNT	8
	b	sbc_8bit
sbc_a_iyh:
	ldrb	r2, [r8, #reg_IYH]
	COUNT	8
	b	sbc_8bit
sbc_a_iyl:
	ldrb	r2, [r8, #reg_IYL]
	COUNT	8
	b	sbc_8bit

sbc_a_n:
	READ_n	r2
	COUNT	7
	b	sbc_8bit
	@@ fallthru @@

sbc_8bit:				@ IN : r1, r2 (byte)
	mov	r2, r2, lsl #24		@ OUT: r0 = r1 - r2 - Cy
	tst	r9, #FLAG_Cy
	orrne	r2, r2, #0x00800000	@ carry-in
	orr	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r1, r7, r2
	subs	r7, r7, r2
	biccs	r9, r9, #FLAG_C		@ (negated)
	bicpl	r9, r9, #FLAG_S
	bicvc	r9, r9, #FLAG_PV
	ands	r7, r7, #0xff000000
	bicne	r9, r9, #FLAG_Z
	eor	r1, r7, r1		@ Half Cy = r1[4] ^ (r1+r2)[4] ^ r2[4]
	tst	r1, #0x10000000
	biceq	r9, r9, #FLAG_H		@ (negated)
	b	inst_end

@--------------------------------------------------------------------
@	CP r		10 111 xxx
@--------------------------------------------------------------------
	.macro	CP_SUB arg2
	orr	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r1, r7, \arg2
	subs	r0, r7, \arg2
	biccs	r9, r9, #FLAG_C		@ (negated)
	bicpl	r9, r9, #FLAG_S
	bicvc	r9, r9, #FLAG_PV
	bicne	r9, r9, #FLAG_Z
	eor	r1, r0, r1		@ Half Cy = r1[4] ^ (r1+r2)[4] ^ r2[4]
	tst	r1, #0x10000000
	biceq	r9, r9, #FLAG_H		@ (negated)
	.endm

cp_a:	CP_SUB	r7
	COUNT	4
	b	inst_end
cp_b:	and	r0, r6, #0xff000000
	CP_SUB	r0
	COUNT	4
	b	inst_end
cp_c:	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsl #8
	CP_SUB	r0
	COUNT	4
	b	inst_end
cp_d:	and	r0, r6, #0x0000ff00
	mov	r0, r0, lsl #16
	CP_SUB	r0
	COUNT	4
	b	inst_end
cp_e:	mov	r0, r6, lsl #24
	CP_SUB	r0
	COUNT	4
	b	inst_end
cp_h:	and	r0, r12, #0xff000000
	CP_SUB	r0
	COUNT	4
	b	inst_end
cp_l:	mov	r0, r12, lsl #8
	CP_SUB	r0
	COUNT	4
	b	inst_end
cp_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	mov	r0, r2, lsl #24
	CP_SUB	r0
	COUNT	7
	b	inst_end
cp_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	bl	cp_8bit
	b	inst_end
cp_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	COUNT	19
	bl	cp_8bit
	b	inst_end
cp_ixh:
	ldrb	r2, [r8, #reg_IXH]
	bl	cp_8bit
	COUNT	8
	b	inst_end
cp_ixl:
	ldrb	r2, [r8, #reg_IXL]
	bl	cp_8bit
	COUNT	8
	b	inst_end
cp_iyh:
	ldrb	r2, [r8, #reg_IYH]
	bl	cp_8bit
	COUNT	8
	b	inst_end
cp_iyl:
	ldrb	r2, [r8, #reg_IYL]
	bl	cp_8bit
	COUNT	8
	b	inst_end

cp_n:
	READ_n	r2
	mov	r0, r2, lsl #24
	CP_SUB	r0
	COUNT	7
	b	inst_end

cp_8bit:				@ IN : r7, r2 (byte)
	mov	r0, r2, lsl #24
	CP_SUB	r0
	bx	lr

@--------------------------------------------------------------------
@	AND r		10 100 xxx
@--------------------------------------------------------------------
and_b:
	and	r7, r7, r6
	COUNT	4
	b	flag_bool_and
and_c:
	and	r7, r7, r6, lsl #8
	COUNT	4
	b	flag_bool_and
and_d:
	and	r7, r7, r6, lsl #16
	COUNT	4
	b	flag_bool_and
and_e:
	and	r7, r7, r6, lsl #24
	COUNT	4
	b	flag_bool_and
and_h:
	and	r7, r7, r12
	COUNT	4
	b	flag_bool_and
and_l:
	and	r7, r7, r12, lsl #8
	COUNT	4
	b	flag_bool_and
and_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	and	r7, r7, r2, lsl #24
	COUNT	7
	b	flag_bool_and
and_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	and	r7, r7, r2, lsl #24
	COUNT	19
	b	flag_bool_and
and_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	and	r7, r7, r2, lsl #24
	COUNT	19
	b	flag_bool_and
and_ixh:
	ldrb	r2, [r8, #reg_IXH]
	and	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool_and
and_ixl:
	ldrb	r2, [r8, #reg_IXL]
	and	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool_and
and_iyh:
	ldrb	r2, [r8, #reg_IYH]
	and	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool_and
and_iyl:
	ldrb	r2, [r8, #reg_IYL]
	and	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool_and

and_n:
	READ_n	r2
	and	r7, r7, r2, lsl #24
	COUNT	7
	b	flag_bool_and

and_a:
	COUNT	4
	@@ fallthru @@
flag_bool_and:
	bic	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N)
	orr	r9, r9, #FLAG_H
	sub	r0, r8, #(WorkAreaL-ParityFlagTblL)		@ parity flag table
	ldrb	r0, [r0, r7, lsr #24]
	orr	r9, r9, r0
	b	inst_end

@--------------------------------------------------------------------
@	XOR r		10 101 xxx
@--------------------------------------------------------------------
xor_a:
	mov	r7, #0
	bic	r9, r9, #(FLAG_C|FLAG_S|FLAG_N|FLAG_H)
	orr	r9, r9, #(FLAG_Z|FLAG_PV)
	COUNT	4
	b	inst_end
xor_b:
	and	r0, r6, #0xff000000
	eor	r7, r7, r0
	COUNT	4
	b	flag_bool
xor_c:
	and	r0, r6, #0x00ff0000
	eor	r7, r7, r0, lsl #8
	COUNT	4
	b	flag_bool
xor_d:
	and	r0, r6, #0x0000ff00
	eor	r7, r7, r0, lsl #16
	COUNT	4
	b	flag_bool
xor_e:
	eor	r7, r7, r6, lsl #24
	COUNT	4
	b	flag_bool
xor_h:
	and	r0, r12, #0xff000000
	eor	r7, r7, r0
	COUNT	4
	b	flag_bool
xor_l:
	eor	r7, r7, r12, lsl #8
	COUNT	4
	b	flag_bool
xor_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	eor	r7, r7, r2, lsl #24
	COUNT	7
	b	flag_bool
xor_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	eor	r7, r7, r2, lsl #24
	COUNT	19
	b	flag_bool
xor_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	eor	r7, r7, r2, lsl #24
	COUNT	19
	b	flag_bool
xor_ixh:
	ldrb	r2, [r8, #reg_IXH]
	eor	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
xor_ixl:
	ldrb	r2, [r8, #reg_IXL]
	eor	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
xor_iyh:
	ldrb	r2, [r8, #reg_IYH]
	eor	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
xor_iyl:
	ldrb	r2, [r8, #reg_IYL]
	eor	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool

xor_n:
	READ_n	r2
	eor	r7, r7, r2, lsl #24
	COUNT	7
	b	flag_bool

@--------------------------------------------------------------------
@	OR r		10 110 xxx
@--------------------------------------------------------------------
or_b:
	and	r0, r6, #0xff000000
	orr	r7, r7, r0
	COUNT	4
	b	flag_bool
or_c:
	and	r0, r6, #0x00ff0000
	orr	r7, r7, r0, lsl #8
	COUNT	4
	b	flag_bool
or_d:
	and	r0, r6, #0x0000ff00
	orr	r7, r7, r0, lsl #16
	COUNT	4
	b	flag_bool
or_e:
	orr	r7, r7, r6, lsl #24
	COUNT	4
	b	flag_bool
or_h:
	and	r0, r12, #0xff000000
	orr	r7, r7, r0
	COUNT	4
	b	flag_bool
or_l:
	orr	r7, r7, r12, lsl #8
	COUNT	4
	b	flag_bool
or_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	orr	r7, r7, r2, lsl #24
	COUNT	7
	b	flag_bool
or_x:
	READ_d	r2
	ldrh	r0, [r8, #reg_IX]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	orr	r7, r7, r2, lsl #24
	COUNT	19
	b	flag_bool
or_y:
	READ_d	r2
	ldrh	r0, [r8, #reg_IY]
	add	r0, r0, r2
	bic	r0, r0, #0xff000000
	bic	r0, r0, #0x00ff0000
	MEM_READ_CYCLE
	orr	r7, r7, r2, lsl #24
	COUNT	19
	b	flag_bool
or_ixh:
	ldrb	r2, [r8, #reg_IXH]
	orr	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
or_ixl:
	ldrb	r2, [r8, #reg_IXL]
	orr	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
or_iyh:
	ldrb	r2, [r8, #reg_IYH]
	orr	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
or_iyl:
	ldrb	r2, [r8, #reg_IYL]
	orr	r7, r7, r2, lsl #24
	COUNT	8
	b	flag_bool
or_n:
	READ_n	r2
	orr	r7, r7, r2, lsl #24
	COUNT	7
	b	flag_bool
or_a:
	COUNT	4
	@@ fallthru @@
flag_bool:
	bic	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	sub	r0, r8, #(WorkAreaL-ParityFlagTblL)		@ parity flag table
	ldrb	r0, [r0, r7, lsr #24]
	orr	r9, r9, r0
	b	inst_end

flag_io:					@ IN: r2	DES: r0
	bic	r9, r9, #(FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	orr	r0, r2, #ParityFlagTblH		@ parity flag table
	ldrb	r0, [r0, #ParityFlagTblL]
	orr	r9, r9, r0
	bx	lr

flag_io_end:					@ IN: r2	DES: r0
	bic	r9, r9, #(FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	orr	r0, r2, #ParityFlagTblH		@ parity flag table
	ldrb	r0, [r0, #ParityFlagTblL]
	orr	r9, r9, r0
	b	inst_end

@--------------------------------------------------------------------
@	RET cc		11 xxx 000
@--------------------------------------------------------------------
ret_nz:
	COUNT	5
	tst	r9, #FLAG_Z
	bne	inst_end
	COUNT	6
	b	pop_pc_end
ret_z:
	COUNT	5
	tst	r9, #FLAG_Z
	beq	inst_end
	COUNT	6
	b	pop_pc_end
ret_nc:
	COUNT	5
	tst	r9, #FLAG_Cy
	bne	inst_end
	COUNT	6
	b	pop_pc_end
ret_c:
	COUNT	5
	tst	r9, #FLAG_Cy
	beq	inst_end
	COUNT	6
	b	pop_pc_end
ret_po:
	COUNT	5
	tst	r9, #FLAG_PV
	bne	inst_end
	COUNT	6
	b	pop_pc_end
ret_pe:
	COUNT	5
	tst	r9, #FLAG_PV
	beq	inst_end
	COUNT	6
	b	pop_pc_end
ret_p:
	COUNT	5
	tst	r9, #FLAG_S
	bne	inst_end
	COUNT	6
	b	pop_pc_end
ret_m:
	COUNT	5
	tst	r9, #FLAG_S
	beq	inst_end
	COUNT	6
	b	pop_pc_end

@--------------------------------------------------------------------
@	POP qq  etc.	11 xxx 001
@--------------------------------------------------------------------
pop_bc:
	bl	pop_nn
	mov	r6, r6, lsl #16		@ DE00
	orr	r6, r6, r0		@ DEBC
	mov	r6, r6, ror #16		@ BCDE
	COUNT	10
	b	inst_end
pop_de:
	bl	pop_nn
	mov	r6, r6, lsr #16		@ 00BC
	orr	r6, r0, r6, lsl #16	@ BCDE
	COUNT	10
	b	inst_end
pop_hl:
	bl	pop_nn
	mov	r12, r0, lsl #16	@ HL00
	COUNT	10
	b	inst_end
pop_af:
	mov	r0, r13, lsr #16
	add	r13, r13, #0x00010000
	MEM_READ_CYCLE
	bic	r9, r9, #0xff
	orr	r9, r9, r2
	mov	r0, r13, lsr #16
	add	r13, r13, #0x00010000
	MEM_READ_CYCLE
	mov	r7, r2, lsl #24
	COUNT	10
	b	inst_end
pop_ix:
	bl	pop_nn
	strh	r0, [r8, #reg_IX]
	COUNT	14
	b	inst_end
pop_iy:
	bl	pop_nn
	strh	r0, [r8, #reg_IY]
	COUNT	14
	b	inst_end

ret_:
	COUNT	10
	b	pop_pc_end

exx_:
	add	r0, r8, #reg_DE_
	swp	r6, r6, [r0]		@ BCDE
	mov	r0, r12, lsr #16	@ 00HL
	ldrh	r12, [r8, #reg_HL_]	@ HL'
	strh	r0, [r8, #reg_HL_]
	mov	r12, r12, lsl #16
	COUNT	4
	b	inst_end

jp_hl:
	mov	r10, r12, lsr #16
	COUNT	4
	b	pc2ptr_end

jp_ix:
	ldrh	r10, [r8, #reg_IX]
	COUNT	8
	b	pc2ptr_end
jp_iy:
	ldrh	r10, [r8, #reg_IY]
	COUNT	8
	b	pc2ptr_end

ld_sp_hl:
	mov	r13, r13, lsl #16
	orr	r13, r12, r13, lsr #16
	COUNT	6
	b	inst_end
ld_sp_ix:
	ldrh	r0, [r8, #reg_IX]
	mov	r13, r13, lsl #16
	mov	r13, r13, lsr #16
	orr	r13, r13, r0, lsl #16
	COUNT	10
	b	inst_end
ld_sp_iy:
	ldrh	r0, [r8, #reg_IY]
	mov	r13, r13, lsl #16
	mov	r13, r13, lsr #16
	orr	r13, r13, r0, lsl #16
	COUNT	10
	b	inst_end

@--------------------------------------------------------------------
@	JP cc,nn	11 xxx 010
@--------------------------------------------------------------------
jp_nz_nn:
	COUNT	10
	tst	r9, #FLAG_Z
	beq	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_z_nn:
	COUNT	10
	tst	r9, #FLAG_Z
	bne	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_nc_nn:
	COUNT	10
	tst	r9, #FLAG_Cy
	beq	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_c_nn:
	COUNT	10
	tst	r9, #FLAG_Cy
	bne	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_po_nn:
	COUNT	10
	tst	r9, #FLAG_PV
	beq	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_pe_nn:
	COUNT	10
	tst	r9, #FLAG_PV
	bne	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_p_nn:
	COUNT	10
	tst	r9, #FLAG_S
	beq	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end
jp_m_nn:
	COUNT	10
	tst	r9, #FLAG_S
	bne	pc2ptr_jp_end
	add	r10, r10, #2
	b	inst_end

@--------------------------------------------------------------------
@	JP nn  etc.	11 xxx 011
@--------------------------------------------------------------------
jp_nn:
	COUNT	10
	b	pc2ptr_jp_end

out_n_a:
	READ_n	r0
	bl	io_write_cycle
	COUNT	11
	b	inst_end

in_a_n:
	mov	r7, #0xff000000		@ No input port!
	COUNT	11
	b	inst_end

ex_sp_hl:				@ (4,3,4,3,5)
	bl	pop_nn
	mov	r5, r12, lsr #16
	mov	r12, r0, lsl #16
	bl	push_nn
	COUNT	19
	b	inst_end
ex_sp_ix:
	bl	pop_nn
	ldrh	r5, [r8, #reg_IX]
	strh	r0, [r8, #reg_IX]
	bl	push_nn
	COUNT	23
	b	inst_end
ex_sp_iy:
	bl	pop_nn
	ldrh	r5, [r8, #reg_IY]
	strh	r0, [r8, #reg_IY]
	bl	push_nn
	COUNT	23
	b	inst_end

ex_de_hl:
	mov	r0, r6			@ copy xxDE
	mov	r6, r6, lsr #16		@ 00BC
	mov	r6, r6, lsl #16		@ BC00
	orr	r6, r6, r12, lsr #16	@ BCHL
	mov	r12, r0, lsl #16	@ DE00
	COUNT	4
	b	inst_end

di_:
	bic	r9, r9, #f_INT_EN
	COUNT	4
	b	inst_end

ei_:
	orr	r9, r9, #f_INT_EN
	COUNT	4
	bl	check_interrupt
	b	inst_end_ei		@ Don't handle interrupt request

@--------------------------------------------------------------------
@	CALL cc,nn	11 xxx 100
@--------------------------------------------------------------------
call_nz_nn:
	COUNT	10
	tst	r9, #FLAG_Z
	addne	r10, r10, #2
	bne	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_z_nn:
	COUNT	10
	tst	r9, #FLAG_Z
	addeq	r10, r10, #2
	beq	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_nc_nn:
	COUNT	10
	tst	r9, #FLAG_Cy
	addne	r10, r10, #2
	bne	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_c_nn:
	COUNT	10
	tst	r9, #FLAG_Cy
	addeq	r10, r10, #2
	beq	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_po_nn:
	COUNT	10
	tst	r9, #FLAG_PV
	addne	r10, r10, #2
	bne	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_pe_nn:
	COUNT	10
	tst	r9, #FLAG_PV
	addeq	r10, r10, #2
	beq	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_p_nn:
	COUNT	10
	tst	r9, #FLAG_S
	addne	r10, r10, #2
	bne	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end
call_m_nn:
	COUNT	10
	tst	r9, #FLAG_S
	addeq	r10, r10, #2
	beq	inst_end
	COUNT	7
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end

@--------------------------------------------------------------------
@	PUSH ss  etc.	11 xxx 101
@--------------------------------------------------------------------
push_bc:
	mov	r5, r6, lsr #16
	bl	push_nn
	COUNT	11
	b	inst_end
push_de:
	mov	r5, r6, lsl #16
	mov	r5, r5, lsr #16
	bl	push_nn
	COUNT	11
	b	inst_end
push_hl:
	mov	r5, r12, lsr #16
	bl	push_nn
	COUNT	11
	b	inst_end
push_af:
	and	r5, r9, #0xff
	orr	r5, r5, r7, lsr #16
	bl	push_nn
	COUNT	11
	b	inst_end
push_ix:
	ldrh	r5, [r8, #reg_IX]
	bl	push_nn
	COUNT	15
	b	inst_end
push_iy:
	ldrh	r5, [r8, #reg_IY]
	bl	push_nn
	COUNT	15
	b	inst_end

call_nn:
	COUNT	17
	READ_nn	r1
	bl	ptr2pc
	mov	r10, r1
	bl	push_nn
	b	pc2ptr_end

@--------------------------------------------------------------------
@	RST n		11 xxx 111
@--------------------------------------------------------------------
rst_00:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x00
	b	pc2ptr_end
rst_08:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x08
	b	pc2ptr_end
rst_10:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x10
	b	pc2ptr_end
rst_18:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x18
	b	pc2ptr_end
rst_20:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x20
	b	pc2ptr_end
rst_28:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x28
	b	pc2ptr_end
rst_30:
	bl	ptr2pc
	bl	push_nn
	COUNT	11
	mov	r10, #0x30
	b	pc2ptr_end

@@ rst_38 is moved into interrupt acknowledge process @@

@--------------------------------------------------------------------
@	DD prefix
@--------------------------------------------------------------------
dd_pfx:
	READ_M1	r2
	add	r0, r2, r2		@ calc jump table offset
	ldrsh	r0, [pc, r0]
	add	pc, pc, r0

#define T(x) (x - dd_tbl - 4)
dd_tbl:
  .hword  T(inst_end),  T(ld_bc_nn),  T(ld_bc_a),   T(inc_bc),    T(inc_b),     T(dec_b),     T(ld_b_n),    T(rlca_)
  .hword  T(ex_af_af),  T(add_ix_bc), T(ld_a_bc),   T(dec_bc),    T(inc_c),     T(dec_c),     T(ld_c_n),    T(rrca_)
  .hword  T(djnz_e),    T(ld_de_nn),  T(ld_de_a),   T(inc_de),    T(inc_d),     T(dec_d),     T(ld_d_n),    T(rla_)
  .hword  T(jr_e),      T(add_ix_de), T(ld_a_de),   T(dec_de),    T(inc_e),     T(dec_e),     T(ld_e_n),    T(rra_)
  .hword  T(jr_nz_e),   T(ld_ix_nn),  T(ld_nn_ix_), T(inc_ix),    T(inc_ixh),   T(dec_ixh),   T(ld_ixh_n),  T(daa_)
  .hword  T(jr_z_e),    T(add_ix_ix), T(ld_ix_nn_), T(dec_ix),    T(inc_ixl),   T(dec_ixl),   T(ld_ixl_n),  T(cpl_)
  .hword  T(jr_nc_e),   T(ld_sp_nn),  T(ld_nn_a),   T(inc_sp),    T(inc_x),     T(dec_x),     T(ld_x_n),    T(scf_)
  .hword  T(jr_c_e),    T(add_ix_sp), T(ld_a_nn),   T(dec_sp),    T(inc_a),     T(dec_a),     T(ld_a_n),    T(ccf_)

  .hword  T(inst_end),  T(ld_b_c),    T(ld_b_d),    T(ld_b_e),    T(ld_b_ixh),  T(ld_b_ixl),  T(ld_b_x),    T(ld_b_a)
  .hword  T(ld_c_b),    T(inst_end),  T(ld_c_d),    T(ld_c_e),    T(ld_c_ixh),  T(ld_c_ixl),  T(ld_c_x),    T(ld_c_a)
  .hword  T(ld_d_b),    T(ld_d_c),    T(inst_end),  T(ld_d_e),    T(ld_d_ixh),  T(ld_d_ixl),  T(ld_d_x),    T(ld_d_a)
  .hword  T(ld_e_b),    T(ld_e_c),    T(ld_e_d),    T(inst_end),  T(ld_e_ixh),  T(ld_e_ixl),  T(ld_e_x),    T(ld_e_a)
  .hword  T(ld_ixh_b),  T(ld_ixh_c),  T(ld_ixh_d),  T(ld_ixh_e),  T(inst_end),  T(ld_ixh_ixl),T(ld_h_x),    T(ld_ixh_a)
  .hword  T(ld_ixl_b),  T(ld_ixl_c),  T(ld_ixl_d),  T(ld_ixl_e),  T(ld_ixl_ixh),T(inst_end),  T(ld_l_x),    T(ld_ixl_a)
  .hword  T(ld_x_b),    T(ld_x_c),    T(ld_x_d),    T(ld_x_e),    T(ld_x_h),    T(ld_x_l),    T(halt_cpu),  T(ld_x_a)
  .hword  T(ld_a_b),    T(ld_a_c),    T(ld_a_d),    T(ld_a_e),    T(ld_a_ixh),  T(ld_a_ixl),  T(ld_a_x),    T(inst_end)

  .hword  T(add_a_b),   T(add_a_c),   T(add_a_d),   T(add_a_e),   T(add_a_ixh), T(add_a_ixl), T(add_a_x),   T(add_a_a)
  .hword  T(adc_a_b),   T(adc_a_c),   T(adc_a_d),   T(adc_a_e),   T(adc_a_ixh), T(adc_a_ixl), T(adc_a_x),   T(adc_a_a)
  .hword  T(sub_b),     T(sub_c),     T(sub_d),     T(sub_e),     T(sub_ixh),   T(sub_ixl),   T(sub_x),     T(sub_a)
  .hword  T(sbc_a_b),   T(sbc_a_c),   T(sbc_a_d),   T(sbc_a_e),   T(sbc_a_ixh), T(sbc_a_ixl), T(sbc_a_x),   T(sbc_a_a)
  .hword  T(and_b),     T(and_c),     T(and_d),     T(and_e),     T(and_ixh),   T(and_ixl),   T(and_x),     T(and_a)
  .hword  T(xor_b),     T(xor_c),     T(xor_d),     T(xor_e),     T(xor_ixh),   T(xor_ixl),   T(xor_x),     T(xor_a)
  .hword  T(or_b),      T(or_c),      T(or_d),      T(or_e),      T(or_ixh),    T(or_ixl),    T(or_x),      T(or_a)
  .hword  T(cp_b),      T(cp_c),      T(cp_d),      T(cp_e),      T(cp_ixh),    T(cp_ixl),    T(cp_x),      T(cp_a)

  .hword  T(ret_nz),    T(pop_bc),    T(jp_nz_nn),  T(jp_nn),     T(call_nz_nn),T(push_bc),   T(add_a_n),   T(rst_00)
  .hword  T(ret_z),     T(ret_),      T(jp_z_nn),   T(dd_cb_pfx), T(call_z_nn), T(call_nn),   T(adc_a_n),   T(rst_08)
  .hword  T(ret_nc),    T(pop_de),    T(jp_nc_nn),  T(out_n_a),   T(call_nc_nn),T(push_de),   T(sub_n),     T(rst_18)
  .hword  T(ret_c),     T(exx_),      T(jp_c_nn),   T(in_a_n),    T(call_c_nn), T(dd_pfx),    T(sbc_a_n),   T(rst_18)
  .hword  T(ret_po),    T(pop_ix),    T(jp_po_nn),  T(ex_sp_ix),  T(call_po_nn),T(push_ix),   T(and_n),     T(rst_20)
  .hword  T(ret_pe),    T(jp_ix),     T(jp_pe_nn),  T(ex_de_hl),  T(call_pe_nn),T(ed_pfx),    T(xor_n),     T(rst_28)
  .hword  T(ret_p),     T(pop_af),    T(jp_p_nn),   T(di_),       T(call_p_nn), T(push_af),   T(or_n),      T(rst_30)
  .hword  T(ret_m),     T(ld_sp_ix),  T(jp_m_nn),   T(ei_),       T(call_m_nn), T(fd_pfx),    T(cp_n),      T(rst_38)
#undef T

@--------------------------------------------------------------------
@	FD prefix
@--------------------------------------------------------------------
fd_pfx:
	READ_M1	r2
	add	r0, r2, r2		@ calc jump table offset
	ldrsh	r1, [pc, r0]
	add	pc, pc, r1

#define T(x) (x - fd_tbl - 4)
fd_tbl:
  .hword  T(inst_end),  T(ld_bc_nn),  T(ld_bc_a),   T(inc_bc),    T(inc_b),     T(dec_b),     T(ld_b_n),    T(rlca_)
  .hword  T(ex_af_af),  T(add_iy_bc), T(ld_a_bc),   T(dec_bc),    T(inc_c),     T(dec_c),     T(ld_c_n),    T(rrca_)
  .hword  T(djnz_e),    T(ld_de_nn),  T(ld_de_a),   T(inc_de),    T(inc_d),     T(dec_d),     T(ld_d_n),    T(rla_)
  .hword  T(jr_e),      T(add_iy_de), T(ld_a_de),   T(dec_de),    T(inc_e),     T(dec_e),     T(ld_e_n),    T(rra_)
  .hword  T(jr_nz_e),   T(ld_iy_nn),  T(ld_nn_iy_), T(inc_iy),    T(inc_iyh),   T(dec_iyh),   T(ld_iyh_n),  T(daa_)
  .hword  T(jr_z_e),    T(add_iy_iy), T(ld_iy_nn_), T(dec_iy),    T(inc_iyl),   T(dec_iyl),   T(ld_iyl_n),  T(cpl_)
  .hword  T(jr_nc_e),   T(ld_sp_nn),  T(ld_nn_a),   T(inc_sp),    T(inc_y),     T(dec_y),     T(ld_y_n),    T(scf_)
  .hword  T(jr_c_e),    T(add_iy_sp), T(ld_a_nn),   T(dec_sp),    T(inc_a),     T(dec_a),     T(ld_a_n),    T(ccf_)

  .hword  T(inst_end),  T(ld_b_c),    T(ld_b_d),    T(ld_b_e),    T(ld_b_iyh),  T(ld_b_iyl),  T(ld_b_y),    T(ld_b_a)
  .hword  T(ld_c_b),    T(inst_end),  T(ld_c_d),    T(ld_c_e),    T(ld_c_iyh),  T(ld_c_iyl),  T(ld_c_y),    T(ld_c_a)
  .hword  T(ld_d_b),    T(ld_d_c),    T(inst_end),  T(ld_d_e),    T(ld_d_iyh),  T(ld_d_iyl),  T(ld_d_y),    T(ld_d_a)
  .hword  T(ld_e_b),    T(ld_e_c),    T(ld_e_d),    T(inst_end),  T(ld_e_iyh),  T(ld_e_iyl),  T(ld_e_y),    T(ld_e_a)
  .hword  T(ld_iyh_b),  T(ld_iyh_c),  T(ld_iyh_d),  T(ld_iyh_e),  T(inst_end),  T(ld_iyh_iyl),T(ld_h_y),    T(ld_iyh_a)
  .hword  T(ld_iyl_b),  T(ld_iyl_c),  T(ld_iyl_d),  T(ld_iyl_e),  T(ld_iyl_iyh),T(inst_end),  T(ld_l_y),    T(ld_iyl_a)
  .hword  T(ld_y_b),    T(ld_y_c),    T(ld_y_d),    T(ld_y_e),    T(ld_y_h),    T(ld_y_l),    T(halt_cpu),  T(ld_y_a)
  .hword  T(ld_a_b),    T(ld_a_c),    T(ld_a_d),    T(ld_a_e),    T(ld_a_iyh),  T(ld_a_iyl),  T(ld_a_y),    T(inst_end)

  .hword  T(add_a_b),   T(add_a_c),   T(add_a_d),   T(add_a_e),   T(add_a_iyh), T(add_a_iyl), T(add_a_y),   T(add_a_a)
  .hword  T(adc_a_b),   T(adc_a_c),   T(adc_a_d),   T(adc_a_e),   T(adc_a_iyh), T(adc_a_iyl), T(adc_a_y),   T(adc_a_a)
  .hword  T(sub_b),     T(sub_c),     T(sub_d),     T(sub_e),     T(sub_iyh),   T(sub_iyl),   T(sub_y),     T(sub_a)
  .hword  T(sbc_a_b),   T(sbc_a_c),   T(sbc_a_d),   T(sbc_a_e),   T(sbc_a_iyh), T(sbc_a_iyl), T(sbc_a_y),   T(sbc_a_a)
  .hword  T(and_b),     T(and_c),     T(and_d),     T(and_e),     T(and_iyh),   T(and_iyl),   T(and_y),     T(and_a)
  .hword  T(xor_b),     T(xor_c),     T(xor_d),     T(xor_e),     T(xor_iyh),   T(xor_iyl),   T(xor_y),     T(xor_a)
  .hword  T(or_b),      T(or_c),      T(or_d),      T(or_e),      T(or_iyh),    T(or_iyl),    T(or_y),      T(or_a)
  .hword  T(cp_b),      T(cp_c),      T(cp_d),      T(cp_e),      T(cp_iyh),    T(cp_iyl),    T(cp_y),      T(cp_a)

  .hword  T(ret_nz),    T(pop_bc),    T(jp_nz_nn),  T(jp_nn),     T(call_nz_nn),T(push_bc),   T(add_a_n),   T(rst_00)
  .hword  T(ret_z),     T(ret_),      T(jp_z_nn),   T(fd_cb_pfx), T(call_z_nn), T(call_nn),   T(adc_a_n),   T(rst_08)
  .hword  T(ret_nc),    T(pop_de),    T(jp_nc_nn),  T(out_n_a),   T(call_nc_nn),T(push_de),   T(sub_n),     T(rst_10)
  .hword  T(ret_c),     T(exx_),      T(jp_c_nn),   T(in_a_n),    T(call_c_nn), T(dd_pfx),    T(sbc_a_n),   T(rst_18)
  .hword  T(ret_po),    T(pop_iy),    T(jp_po_nn),  T(ex_sp_iy),  T(call_po_nn),T(push_iy),   T(and_n),     T(rst_20)
  .hword  T(ret_pe),    T(jp_iy),     T(jp_pe_nn),  T(ex_de_hl),  T(call_pe_nn),T(ed_pfx),    T(xor_n),     T(rst_28)
  .hword  T(ret_p),     T(pop_af),    T(jp_p_nn),   T(di_),       T(call_p_nn), T(push_af),   T(or_n),      T(rst_30)
  .hword  T(ret_m),     T(ld_sp_iy),  T(jp_m_nn),   T(ei_),       T(call_m_nn), T(fd_pfx),    T(cp_n),      T(rst_38)
#undef T

@--------------------------------------------------------------------
@	ED prefix
@--------------------------------------------------------------------
ed_pfx:
	READ_M1	r2			@ instruction read
	add	r0, r2, r2		@ calc jump table offset
	ldrsh	r0, [pc, r0]
	add	pc, pc, r0

#define T(x) (x - ed_tbl - 4)
ed_tbl:
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)

  .hword  T(in_b_c),    T(out_c_r),   T(sbc_hl_bc), T(ld_nn_bc_), T(neg_),      T(retn_),     T(inst_end),  T(ld_i_a)
  .hword  T(in_c_c),    T(out_c_r),   T(adc_hl_bc), T(ld_bc_nn_), T(neg_),      T(reti_),     T(inst_end),  T(ld_r_a)
  .hword  T(in_d_c),    T(out_c_r),   T(sbc_hl_de), T(ld_nn_de_), T(neg_),      T(retn_),     T(inst_end),  T(ld_a_i)
  .hword  T(in_e_c),    T(out_c_r),   T(adc_hl_de), T(ld_de_nn_), T(neg_),      T(reti_),     T(inst_end),  T(ld_a_r)
  .hword  T(in_h_c),    T(out_c_r),   T(sbc_hl_hl), T(ld_nn_hl_), T(neg_),      T(retn_),     T(inst_end),  T(rrd_)
  .hword  T(in_l_c),    T(out_c_r),   T(adc_hl_hl), T(ld_hl_nn_), T(neg_),      T(reti_),     T(inst_end),  T(rld_)
  .hword  T(in_m_c),    T(out_c_r),   T(sbc_hl_sp), T(ld_nn_sp_), T(neg_),      T(retn_),     T(inst_end),  T(inst_end)
  .hword  T(in_a_c),    T(out_c_r),   T(adc_hl_sp), T(ld_sp_nn_), T(neg_),      T(reti_),     T(inst_end),  T(inst_end)

  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(ldi_),      T(cpi_),      T(ini_),      T(outi_),     T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(ldd_),      T(cpd_),      T(ind_),      T(outd_),     T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(ldir_),     T(cpir_),     T(inir_),     T(otir_),     T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(lddr_),     T(cpdr_),     T(indr_),     T(otdr_),     T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)

  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(mzt_mon),   T(inst_end),  T(inst_end),  T(inst_end),  T(mzt_bas),   T(inst_end),  T(inst_end),  T(inst_end)
  .hword  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end),  T(inst_end)
#undef T

@--------------------------------------------------------------------
@	IN r,(C)	*** No I/O! ***
@--------------------------------------------------------------------
in_a_c:	mov	r7, #0xff000000
	b	in_m_c
in_b_c:	orr	r6, r6, #0xff000000
	b	in_m_c
in_c_c:	orr	r6, r6, #0x00ff0000
	b	in_m_c
in_d_c:	orr	r6, r6, #0x0000ff00
	b	in_m_c
in_e_c:	orr	r6, r6, #0x000000ff
	b	in_m_c
in_h_c:	orr	r12, r12, #0xff000000
	b	in_m_c
in_l_c:	orr	r12, r12, #0x00ff0000
in_m_c:	COUNT	12
	b	flag_io_end

@--------------------------------------------------------------------
@	OUT (C),r
@--------------------------------------------------------------------
out_c_r:
	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsr #16
	bl	io_write_cycle
	COUNT	12
	b	inst_end

@--------------------------------------------------------------------
@	ADC HL,ss
@--------------------------------------------------------------------
adc_hl_hl:
	mov	r2, r12, lsr #16
	b	adc_16bit
adc_hl_sp:
	mov	r2, r13, lsr #16
	b	adc_16bit
adc_hl_de:
	mov	r2, r6, lsl #16
	b	1f
adc_hl_bc:
	mov	r2, r6, lsr #16
adc_16bit:				@ IN : r1, r2 (byte)
	mov	r2, r2, lsl #16
1:	tst	r9, #FLAG_Cy
	orrne	r12, r12, #0x00008000	@ carry-in
	orrne	r2,  r2,  #0x00008000	@ carry-in
	bic	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r0, r12, r2
	adds	r12, r12, r2
	orreq	r9, r9, #FLAG_Z
	orrcs	r9, r9, #FLAG_C
	orrmi	r9, r9, #FLAG_S
	orrvs	r9, r9, #FLAG_PV
	eor	r0, r12, r0
	tst	r0, #0x10000000		@ Half carry
	orrne	r9, r9, #FLAG_H
	COUNT	15
	b	inst_end

@--------------------------------------------------------------------
@	SBC HL,ss
@--------------------------------------------------------------------
sbc_hl_hl:
	mov	r2, r12, lsr #16
	b	sbc_16bit
sbc_hl_sp:
	mov	r2, r13, lsr #16
	b	sbc_16bit
sbc_hl_de:
	mov	r2, r6, lsl #16
	b	1f
sbc_hl_bc:
	mov	r2, r6, lsr #16
sbc_16bit:				@ IN : r1, r2 (byte)
	mov	r2, r2, lsl #16
1:	tst	r9, #FLAG_Cy
	orrne	r2, r2, #0x00008000	@ carry-in
	orr	r9, r9, #(FLAG_C|FLAG_Z|FLAG_PV|FLAG_S|FLAG_N|FLAG_H)
	eor	r0, r12, r2
	subs	r12, r12, r2
	biccs	r9, r9, #FLAG_C
	bicpl	r9, r9, #FLAG_S
	bicvc	r9, r9, #FLAG_PV
	bics	r12, r12, #0x0000ff00
	bicne	r9, r9, #FLAG_Z
	eor	r0, r12, r0
	tst	r0, #0x10000000		@ Half carry
	biceq	r9, r9, #FLAG_H
	COUNT	15
	b	inst_end

@--------------------------------------------------------------------
@	LD (nn),ss
@--------------------------------------------------------------------
ld_nn_bc_:
	READ_nn	r0
	add	r5, r0, #1
	and	r2, r6, #0x00ff0000
	mov	r2, r2, lsr #16
	MEM_WRITE_CYCLE
	bic	r0, r5, #0x00ff0000
	mov	r2, r6, lsr #24
	MEM_WRITE_CYCLE
	COUNT	20
	b	inst_end
ld_bc_nn_:
	READ_nn	r0
	add	r3, r0, #1
	MEM_READ_CYCLE
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	bic	r0, r3, #0x00ff0000
	MEM_READ_CYCLE
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	20
	b	inst_end
ld_nn_de_:
	READ_nn	r0
	add	r5, r0, #1
	and	r2, r6, #0x000000ff
	MEM_WRITE_CYCLE
	bic	r0, r5, #0x00ff0000
	and	r2, r6, #0x0000ff00
	mov	r2, r2, lsr #8
	MEM_WRITE_CYCLE
	COUNT	20
	b	inst_end
ld_de_nn_:
	READ_nn	r0
	add	r3, r0, #1
	MEM_READ_CYCLE
	mov	r6, r6, lsr #16		@ 00BC
	orr	r6, r2, r6, lsl #16	@ BC0E
	bic	r0, r3, #0x00ff0000
	MEM_READ_CYCLE
	orr	r6, r6, r2, lsl #8	@ BCDE
	COUNT	20
	b	inst_end
ld_nn_sp_:
	READ_nn	r0
	add	r5, r0, #1
	and	r2, r13, #0x00ff0000
	mov	r2, r2, lsr #16
	MEM_WRITE_CYCLE
	bic	r0, r5, #0x00ff0000
	mov	r2, r13, lsr #24
	MEM_WRITE_CYCLE
	COUNT	20
	b	inst_end
ld_sp_nn_:
	READ_nn	r0
	add	r3, r0, #1
	MEM_READ_CYCLE
	bic	r13, r13, #0x00ff0000
	orr	r13, r13, r2, lsl #16
	bic	r0, r3, #0x00ff0000
	MEM_READ_CYCLE
	bic	r13, r13, #0xff000000
	orr	r13, r13, r2, lsl #24
	COUNT	20
	b	inst_end

@--------------------------------------------------------------------
@	NEG etc.	01 xxx 1xx
@--------------------------------------------------------------------
neg_:
	mov	r2, r7, lsr #24
	mov	r7, #0
	COUNT	8
	b	sub_8bit

retn_:	b	ret_

reti_:	b	ret_

ld_i_a:
	mov	r0, r7, lsr #24
	strb	r0, [r8, #reg_I]
	COUNT	9
	b	inst_end
ld_r_a:
	bic	r9, r9, #f_Rreg_MASK
	orr	r9, r9, r7, lsl #1
	mov	r0, r7, lsr #24
	strb	r0, [r8, #reg_R]
	COUNT	9
	b	inst_end
ld_a_i:
	bic	r9, r9, #(FLAG_S|FLAG_Z|FLAG_PV|FLAG_N|FLAG_H)
	ldrb	r0, [r8, #reg_I]
	movs	r7, r0, lsl #24
	orreq	r9, r9, #FLAG_Z
	orrmi	r9, r9, #FLAG_S
	tst	r9, #f_INT_EN
	orrne	r9, r9, #FLAG_PV
	COUNT	9
	b	inst_end
ld_a_r:
	bic	r9, r9, #(FLAG_S|FLAG_Z|FLAG_PV|FLAG_N|FLAG_H)
	ldrb	r0, [r8, #reg_R]
	ands	r0, r0, #0x80		@ Bit 7
	orrne	r9, r9, #FLAG_S
	orrs	r0, r0, r9, lsr #25	@ Bit 6-0
	orreq	r9, r9, #FLAG_Z
	mov	r7, r0, lsl #24
	tst	r9, #f_INT_EN
	orrne	r9, r9, #FLAG_PV
	COUNT	9
	b	inst_end

rrd_:					@ A7-A4:A3-A0:M7-M4:M3-M0 --> A7-A4:M3-M0:A3-A0:M7-M4
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	orr	r2, r2, r7, lsr #16
	and	r7, r7, #0xf0000000
	and	r1, r2, #0x0f
	mov	r2, r2, lsr #4
	orr	r7, r7, r1, lsl #24
	and	r2, r2, #0xff
	mov	r0, r12, lsr #16
	COUNT	18
	MEM_WRITE_CYCLE
	mov	r2, r7, lsr #24
	b	flag_io_end
rld_:					@ A7-A4:A3-A0:M7-M4:M3-M0 --> A7-A4:M7-M4:M3-M0:A3-A0
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	and	r1, r7, #0x0f000000
	and	r7, r7, #0xf0000000
	orr	r1, r1, r2, ror #4
	orr	r7, r7, r2, lsl #20
	mov	r2, r1, lsr #24
	and	r7, r7, #0xff000000
	mov	r0, r12, lsr #16
	COUNT	18
	MEM_WRITE_CYCLE
	mov	r2, r7, lsr #24
	b	flag_io_end

@--------------------------------------------------------------------
@	LDI/LDD/LDIR/LDDR
@--------------------------------------------------------------------
ldi_:					@ 16
	mov	r0, r12, lsr #16
	add	r12, r12, #0x00010000
	MEM_READ_CYCLE
	mov	r6, r6, ror #16
	mov	r0, r6, lsr #16
	add	r6, r6, #0x00010000
	mov	r6, r6, ror #16
	MEM_WRITE_CYCLE
	bic	r9, r9, #(FLAG_H|FLAG_PV|FLAG_N)
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	b	inst_end
ldd_:					@ 16
	mov	r0, r12, lsr #16
	sub	r12, r12, #0x00010000
	MEM_READ_CYCLE
	mov	r6, r6, ror #16
	mov	r0, r6, lsr #16
	sub	r6, r6, #0x00010000
	mov	r6, r6, ror #16
	MEM_WRITE_CYCLE
	bic	r9, r9, #(FLAG_H|FLAG_PV|FLAG_N)
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	b	inst_end
ldir_:					@ 21
	mov	r0, r12, lsr #16
	add	r12, r12, #0x00010000
	MEM_READ_CYCLE
	mov	r6, r6, ror #16
	mov	r0, r6, lsr #16
	add	r6, r6, #0x00010000
	mov	r6, r6, ror #16
	MEM_WRITE_CYCLE
	bic	r9, r9, #(FLAG_H|FLAG_PV|FLAG_N)
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	blo	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end
lddr_:					@ 16
	mov	r0, r12, lsr #16
	sub	r12, r12, #0x00010000
	MEM_READ_CYCLE
	mov	r6, r6, ror #16
	mov	r0, r6, lsr #16
	sub	r6, r6, #0x00010000
	mov	r6, r6, ror #16
	MEM_WRITE_CYCLE
	bic	r9, r9, #(FLAG_H|FLAG_PV|FLAG_N)
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	blo	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end

@--------------------------------------------------------------------
@	CPI/CPD/CPIR/CPDR
@--------------------------------------------------------------------
cpi_:
	and	r5, r9, #FLAG_C		@ save Cy
	mov	r0, r12, lsr #16
	add	r12, r12, #0x00010000
	MEM_READ_CYCLE
	bl	cp_8bit			@ compare
	bic	r9, r9, #(FLAG_C|FLAG_PV)
	orr	r9, r9, r5		@ restore Cy
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	b	inst_end
cpd_:
	and	r5, r9, #FLAG_C		@ save Cy
	mov	r0, r12, lsr #16
	sub	r12, r12, #0x00010000
	MEM_READ_CYCLE
	bl	cp_8bit			@ compare
	bic	r9, r9, #(FLAG_C|FLAG_PV)
	orr	r9, r9, r5		@ restore Cy
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	b	inst_end
cpir_:
	and	r5, r9, #FLAG_C		@ save Cy
	mov	r0, r12, lsr #16
	add	r12, r12, #0x00010000
	MEM_READ_CYCLE
	bl	cp_8bit			@ compare
	bic	r9, r9, #(FLAG_C|FLAG_PV)
	orr	r9, r9, r5		@ restore Cy
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	blo	inst_end
	tst	r9, #FLAG_Z
	bne	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end
cpdr_:
	and	r5, r9, #FLAG_C		@ save Cy
	mov	r0, r12, lsr #16
	sub	r12, r12, #0x00010000
	MEM_READ_CYCLE
	bl	cp_8bit			@ compare
	bic	r9, r9, #(FLAG_C|FLAG_PV)
	orr	r9, r9, r5		@ restore Cy
	sub	r6, r6, #0x00010000
	cmp	r6, #0x00010000
	orrhs	r9, r9, #FLAG_PV
	COUNT	16
	blo	inst_end
	tst	r9, #FLAG_Z
	bne	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end

@--------------------------------------------------------------------
@	INI/IND/INIR/INDR
@--------------------------------------------------------------------
ini_:
	mov	r0, r12, lsr #16
	add	r12, r12, #0x00010000
	mov	r2, #0xff
	MEM_WRITE_CYCLE
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	b	inst_end
ind_:
	mov	r0, r12, lsr #16
	sub	r12, r12, #0x00010000
	mov	r2, #0xff
	MEM_WRITE_CYCLE
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	b	inst_end
inir_:
	mov	r0, r12, lsr #16
	add	r12, r12, #0x00010000
	mov	r2, #0xff
	MEM_WRITE_CYCLE
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	blo	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end
indr_:
	mov	r0, r12, lsr #16
	sub	r12, r12, #0x00010000
	mov	r2, #0xff
	MEM_WRITE_CYCLE
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	blo	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end

@--------------------------------------------------------------------
@	OUTI/OUTD/OTIR/OTDR
@--------------------------------------------------------------------
outi_:
	add	r12, r12, #0x00010000
	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsr #16
	bl	io_write_cycle
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	b	inst_end
outd_:
	sub	r12, r12, #0x00010000
	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsr #16
	bl	io_write_cycle
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	b	inst_end
otir_:
	add	r12, r12, #0x00010000
	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsr #16
	bl	io_write_cycle
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	blo	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end
otdr_:
	sub	r12, r12, #0x00010000
	and	r0, r6, #0x00ff0000
	mov	r0, r0, lsr #16
	bl	io_write_cycle
	bic	r9, r9, #FLAG_Z
	orr	r9, r9, #FLAG_N
	sub	r6, r6, #0x01000000
	cmp	r6, #0x01000000
	orrlo	r9, r9, #FLAG_Z
	COUNT	16
	blo	inst_end
	sub	r10, r10, #2
	COUNT	5
	b	inst_end

@--------------------------------------------------------------------
@	MZ700Win compatible tape emulation instruction
@
@	ED F0 01	RDINF  (monitor)	Load Information block into 10F0h
@	ED F0 02	RDDATA (monitor)	Load Data block into the address specified in (1104h) with length (1102h)
@	ED F4 01	RDINF  (BASIC)		Load Information block into (HL)
@	ED F4 02	RDDATA (BASIC)		Load Data block into (HL) with length BC
@	ED F4 05	RDINF  (BASIC)		Load Information block into (HL)
@	ED F4 06	RDDATA (BASIC)		Load Data block into (HL) with length BC
@					OUT: A=0, Cy=0 ... no error
@--------------------------------------------------------------------
mzt_mon:
	ldrb	r2, [r10], #1			@ Get 3rd opecode
	ldr	r5, [r8, #TapePtr]		@ tape header pointer
	ldr	r0, [r8, #TapeEndPtr]
	cmp	r5, r0				@ No more tape image?
	bhs	9f
	cmp	r2, #1				@ RDINF?
	bne	2f
rdinf_mon:
	mov	r4, #0x00001000			@ info buffer
	orr	r4, r4, #0x00f0
	str	r5, [r8, #TapeInfoBlkPtr]	@ save last-read info block pointer
	mov	r7, #0x80			@ 128 bytes
1:	ldrb	r2, [r5], #1			@ read 1 byte from tape img
	mov	r0, r4
	MEM_WRITE_CYCLE
	add	r4, r4, #1
	subs	r7, r7, #1
	bne	1b
	str	r5, [r8, #TapePtr]
@@@	mov	r7, #0				@ A=0 (already 0)
	bic	r9, r9, #FLAG_C			@ Cy=0
	b	ret_
9:	orr	r9, r9, #FLAG_C			@ Error
	mov	r7, #0xff
	b	ret_

2:	cmp	r2, #2				@ RDDATA?
	bne	ret_
rddata_mon:
	mov	r2, #0x02000000			@ DRAM start address
	orr	r2, r2, #0x1100			@ info buffer
	ldrb	r0, [r2, #2]
	ldrb	r1, [r2, #3]
	orr	r7, r0, r1, lsl #8		@ size
	ldrb	r0, [r2, #4]
	ldrb	r1, [r2, #5]
	orr	r4, r0, r1, lsl #8		@ load addr
	ldr	r5, [r8, #TapePtr]		@ tape header pointer
	ldr	r0, [r8, #TapeEndPtr]
	cmp	r5, r0
	bhs	9b
1:	ldrb	r2, [r5], #1			@ read 1 byte from tape img
	mov	r0, r4
	MEM_WRITE_CYCLE
	add	r4, r4, #1
	subs	r7, r7, #1
	bne	1b
	str	r5, [r8, #TapePtr]

	ldr	r5, [r8, #TapeInfoBlkPtr]	@ check info header extension
	tst	r5, r5
	beq	rddata_mon_xit
	add	r5, r5, #0x40
rddata_mon_chkext:
	ldrb	r0, [r5]			@ check 'PAT:'
	cmp	r0, #'P'
	ldreqb	r0, [r5, #1]
	cmpeq	r0, #'A'
	ldreqb	r0, [r5, #2]
	cmpeq	r0, #'T'
	ldreqb	r0, [r5, #3]
	cmpeq	r0, #':'
	beq	rddata_mon_patch
	ldrb	r0, [r5]			@ check 'AKP:'
	cmp	r0, #'A'
	ldreqb	r0, [r5, #1]
	cmpeq	r0, #'K'
	ldreqb	r0, [r5, #2]
	cmpeq	r0, #'P'
	ldreqb	r0, [r5, #3]
	cmpeq	r0, #':'
	beq	rddata_mon_autokeypress
rddata_mon_xit:
	mov	r7, #0				@ A=0
	bic	r9, r9, #FLAG_C			@ Cy=0
	str	r7, [r8, #TapeInfoBlkPtr]	@ clear last-read info block pointer
	b	ret_

rddata_mon_patch:
	add	r5, r5, #4
2:	ldrb	r4, [r5], #1
	ldrb	r1, [r5], #1
	orr	r4, r4, r1, lsl #8
	add	r2, r4, #1
	cmp	r2, #0x00010000		@ r4 == 0xFFFF?
	beq	rddata_mon_chkext	@ end of patch data
	ldrb	r7, [r5], #1		@ patch length
1:	ldrb	r2, [r5], #1		@ patch data 1byte
	mov	r0, r4			@ patch address
	MEM_WRITE_CYCLE
	add	r4, r4, #1
	subs	r7, r7, #1
	bne	1b
	b	2b

rddata_mon_autokeypress:			@    -- Format --
	add	r5, r5, #4			@	'AKP:'
	ldrb	r0, [r5], #1			@	Row(byte)
	ldrb	r1, [r5], #1			@	Mask(byte) 
	orr	r0, r0, r1, lsl #8
	ldrb	r1, [r5], #1			@	WaitTime(byte,1-255)
	orr	r0, r0, r1, lsl #16
	ldrb	r1, [r5], #1			@	PressTime(byte,1-127)
	orr	r0, r0, r1, lsl #24
	str	r0, [r8, #KeyPressOnce]
	b	rddata_mon_chkext

mzt_bas:
	ldrb	r2, [r10], #1			@ Get 3rd opecode
	ldr	r5, [r8, #TapePtr]		@ tape header pointer
	ldr	r0, [r8, #TapeEndPtr]
	cmp	r5, r0				@ No more tape image?
	bhs	9b
	cmp	r2, #1				@ RDINF(S)?
	beq	rdinf_bas
	cmp	r2, #5				@ RDINF(Hu)?
	bne	2f
rdinf_bas:
	mov	r4, r12, lsr #16		@ info buffer to (HL)
	str	r5, [r8, #TapeInfoBlkPtr]	@ save last-read info block pointer
	mov	r7, #0x80			@ 128 bytes
1:	ldrb	r2, [r5], #1			@ read 1 byte from tape img
	mov	r0, r4
	MEM_WRITE_CYCLE
	add	r4, r4, #1
	subs	r7, r7, #1
	bne	1b
	str	r5, [r8, #TapePtr]
@@@	mov	r7, #0				@ A=0 (already 0)
	bic	r9, r9, #FLAG_C			@ Cy=0
	b	ret_

2:	cmp	r2, #2				@ RDDATA(S)?
	beq	rddata_bas
	cmp	r2, #6				@ RDDATA(Hu)?
	bne	ret_
rddata_bas:
	mov	r7, r6,	 lsr #16		@ size = BC
	mov	r4, r12, lsr #16		@ load addr (HL)
1:	ldrb	r2, [r5], #1			@ read 1 byte from tape img
	mov	r0, r4
	MEM_WRITE_CYCLE
	add	r4, r4, #1
	subs	r7, r7, #1
	bne	1b
	str	r5, [r8, #TapePtr]
@@@	mov	r7, #0				@ A=0 (already 0)
	bic	r9, r9, #FLAG_C			@ Cy=0
	b	ret_

@--------------------------------------------------------------------
@	CB prefix
@--------------------------------------------------------------------
cb_pfx:
	READ_M1	r0			@ instruction read
	add	r2, r0, r0		@ calc jump table offset
	ldrsh	r1, [pc, r2]
	add	pc, pc, r1

#define T(x) (x - cb_tbl - 4)
cb_tbl:
  .hword  T(rlc_b),     T(rlc_c),     T(rlc_d),     T(rlc_e),     T(rlc_h),     T(rlc_l),     T(rlc_m),     T(rlc_a)
  .hword  T(rrc_b),     T(rrc_c),     T(rrc_d),     T(rrc_e),     T(rrc_h),     T(rrc_l),     T(rrc_m),     T(rrc_a)
  .hword  T(rl_b),      T(rl_c),      T(rl_d),      T(rl_e),      T(rl_h),      T(rl_l),      T(rl_m),      T(rl_a)
  .hword  T(rr_b),      T(rr_c),      T(rr_d),      T(rr_e),      T(rr_h),      T(rr_l),      T(rr_m),      T(rr_a)
  .hword  T(sla_b),     T(sla_c),     T(sla_d),     T(sla_e),     T(sla_h),     T(sla_l),     T(sla_m),     T(sla_a)
  .hword  T(sra_b),     T(sra_c),     T(sra_d),     T(sra_e),     T(sra_h),     T(sra_l),     T(sra_m),     T(sra_a)
  .hword  T(sll_b),     T(sll_c),     T(sll_d),     T(sll_e),     T(sll_h),     T(sll_l),     T(sll_m),     T(sll_a)
  .hword  T(srl_b),     T(srl_c),     T(srl_d),     T(srl_e),     T(srl_h),     T(srl_l),     T(srl_m),     T(srl_a)

  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)
  .hword  T(bit_n_b),   T(bit_n_c),   T(bit_n_d),   T(bit_n_e),   T(bit_n_h),   T(bit_n_l),   T(bit_n_m),   T(bit_n_a)

  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_3_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)
  .hword  T(res_n_b),   T(res_n_c),   T(res_n_d),   T(res_n_e),   T(res_n_h),   T(res_n_l),   T(res_n_m),   T(res_n_a)

  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_3_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
  .hword  T(set_n_b),   T(set_n_c),   T(set_n_d),   T(set_n_e),   T(set_n_h),   T(set_n_l),   T(set_n_m),   T(set_n_a)
#undef T

@--------------------------------------------------------------------
@	DD CB prefix
@--------------------------------------------------------------------
fd_cb_pfx:
	ldrh	r5, [r8, #reg_IY]
	b	1f

dd_cb_pfx:
	ldrh	r5, [r8, #reg_IX]
1:
	READ_d	r2			@ extend sign
	mov	r5, r5, lsl #16
	add	r5, r5, r2, lsl #16	@ r5 = IX+d
	mov	r5, r5, lsr #16

	READ_M1	r0			@ XD CB dd XX
	add	r2, r0, r0		@ calc jump table offset
	ldrsh	r2, [pc, r2]
	add	pc, pc, r2

#define T(x) (x - xd_cb_tbl - 4)
xd_cb_tbl:
  .hword  T(rlc_b_x),   T(rlc_c_x),   T(rlc_d_x),   T(rlc_e_x),   T(rlc_h_x),   T(rlc_l_x),   T(rlc_x),     T(rlc_a_x)
  .hword  T(rrc_b_x),   T(rrc_c_x),   T(rrc_d_x),   T(rrc_e_x),   T(rrc_h_x),   T(rrc_l_x),   T(rrc_x),     T(rrc_a_x)
  .hword  T(rl_b_x),    T(rl_c_x),    T(rl_d_x),    T(rl_e_x),    T(rl_h_x),    T(rl_l_x),    T(rl_x),      T(rl_a_x)
  .hword  T(rr_b_x),    T(rr_c_x),    T(rr_d_x),    T(rr_e_x),    T(rr_h_x),    T(rr_l_x),    T(rr_x),      T(rr_a_x)
  .hword  T(sla_b_x),   T(sla_c_x),   T(sla_d_x),   T(sla_e_x),   T(sla_h_x),   T(sla_l_x),   T(sla_x),     T(sla_a_x)
  .hword  T(sra_b_x),   T(sra_c_x),   T(sra_d_x),   T(sra_e_x),   T(sra_h_x),   T(sra_l_x),   T(sra_x),     T(sra_a_x)
  .hword  T(sll_b_x),   T(sll_c_x),   T(sll_d_x),   T(sll_e_x),   T(sll_h_x),   T(sll_l_x),   T(sll_x),     T(sll_a_x)
  .hword  T(srl_b_x),   T(srl_c_x),   T(srl_d_x),   T(srl_e_x),   T(srl_h_x),   T(srl_l_x),   T(srl_x),     T(srl_a_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x),   T(bit_n_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(res_n_b_x), T(res_n_c_x), T(res_n_d_x), T(res_n_e_x), T(res_n_h_x), T(res_n_l_x), T(res_n_x),   T(res_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
  .hword  T(set_n_b_x), T(set_n_c_x), T(set_n_d_x), T(set_n_e_x), T(set_n_h_x), T(set_n_l_x), T(set_n_x),   T(set_n_a_x)
#undef T

@--------------------------------------------------------------------
@	RLC r
@--------------------------------------------------------------------
rlc_a:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsl #1
	orrcs	r7, r7, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r7, lsr #24
	COUNT	8
	b	flag_io_end
rlc_b:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0xff000000
	adds	r6, r6, r0
	orrcs	r6, r6, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r6, lsr #24
	COUNT	8
	b	flag_io_end
rlc_c:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0x00ff0000
	movs	r0, r0, lsl #1+8
	orrcs	r0, r0, #0x01000000
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsr #8
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rlc_d:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0x0000ff00
	movs	r0, r0, lsl #1+16
	orrcs	r0, r0, #0x01000000
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsr #16
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rlc_e:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r0, r6, lsl #1+24
	orrcs	r0, r0, #0x01000000
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0, lsr #24
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rlc_h:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r12, #0xff000000
	adds	r12, r12, r0
	orrcs	r12, r12, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r12, lsr #24
	COUNT	8
	b	flag_io_end
rlc_l:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r0, r12, lsl #1+8
	orrcs	r0, r0, #0x01000000
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r0, lsr #8
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rlc_m:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	movs	r2, r2, lsl #25
	orrcs	r2, r2, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end

rlc_a_x:
	bl	rlc_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
rlc_b_x:
	bl	rlc_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
rlc_c_x:
	bl	rlc_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
rlc_d_x:
	bl	rlc_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
rlc_e_x:
	bl	rlc_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
rlc_h_x:
	bl	rlc_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
rlc_l_x:
	bl	rlc_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
rlc_x:
	bl	rlc_x_sub
	b	inst_end
rlc_x_sub:
	mov	r4, lr
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r0, r5
	MEM_READ_CYCLE
	movs	r2, r2, lsl #25
	orrcs	r2, r2, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	RRC r
@--------------------------------------------------------------------
rrc_a:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r7, lsr #1+24
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	mov	r7, r2, lsl #24
	COUNT	8
	b	flag_io_end
rrc_b:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r6, lsr #1+24
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	8
	b	flag_io_end
rrc_c:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r6, #0x00ff0000
	movs	r2, r2, lsr #1+16
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	8
	b	flag_io_end
rrc_d:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r6, #0x0000ff00
	movs	r2, r2, lsr #1+8
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	8
	b	flag_io_end
rrc_e:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r6, #0x000000ff
	movs	r2, r2, lsr #1
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	8
	b	flag_io_end
rrc_h:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r12, lsr #1+24
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	8
	b	flag_io_end
rrc_l:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r12, #0x00ff0000
	movs	r2, r2, lsr #1+16
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	8
	b	flag_io_end
rrc_m:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	movs	r2, r2, lsr #1
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
rrc_a_x:
	bl	rrc_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
rrc_b_x:
	bl	rrc_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
rrc_c_x:
	bl	rrc_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
rrc_d_x:
	bl	rrc_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
rrc_e_x:
	bl	rrc_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
rrc_h_x:
	bl	rrc_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
rrc_l_x:
	bl	rrc_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
rrc_x:
	bl	rrc_x_sub
	b	inst_end
rrc_x_sub:
	mov	r4, lr
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r0, r5
	MEM_READ_CYCLE
	movs	r2, r2, lsr #1
	orrcs	r2, r2, #0x80
	orrcs	r9, r9, #FLAG_C
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	RL r
@--------------------------------------------------------------------
rl_a:
	tst	r9, #FLAG_Cy
	orrne	r7, r7, #0x00800000
	movs	r7, r7, lsl #1
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	mov	r2, r7, lsr #24
	COUNT	8
	b	flag_io_end
rl_b:
	and	r0, r6, #0xff000000
	tst	r9, #FLAG_Cy
	orrne	r0, r0, #0x00800000
	movs	r0, r0, lsl #1
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0xff000000
	orr	r6, r6, r0
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rl_c:
	and	r0, r6, #0x00ff0000
	tst	r9, #FLAG_Cy
	orrne	r0, r0, #0x00008000
	movs	r0, r0, lsl #1+8
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsr #8
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rl_d:
	and	r0, r6, #0x0000ff00
	tst	r9, #FLAG_Cy
	orrne	r0, r0, #0x00000080
	movs	r0, r0, lsl #1+16
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsr #16
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rl_e:
	and	r1, r9, #FLAG_Cy
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r0, r6, lsl #1+24
	orr	r0, r0, r1, lsl #24
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0, lsr #24
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rl_h:
	and	r1, r9, #FLAG_Cy
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r12, #0xff000000
	adds	r12, r12, r0
	orr	r12, r12, r1, lsl #24
	orrcs	r9, r9, #FLAG_C
	mov	r2, r12, lsr #24
	COUNT	8
	b	flag_io_end
rl_l:
	and	r0, r12, #0x00ff0000
	tst	r9, #FLAG_Cy
	orrne	r0, r0, #0x00008000
	movs	r0, r0, lsl #1+8
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r0, lsr #8
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
rl_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	and	r1, r9, #FLAG_Cy
	movs	r2, r2, lsl #25
	orr	r2, r2, r1, lsl #24
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
rl_a_x:
	bl	rl_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
rl_b_x:
	bl	rl_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
rl_c_x:
	bl	rl_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
rl_d_x:
	bl	rl_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
rl_e_x:
	bl	rl_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
rl_h_x:
	bl	rl_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
rl_l_x:
	bl	rl_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
rl_x:
	bl	rl_x_sub
	b	inst_end
rl_x_sub:
	mov	r4, lr
	mov	r0, r5
	MEM_READ_CYCLE
	and	r1, r9, #FLAG_Cy
	movs	r2, r2, lsl #25
	orr	r2, r2, r1, lsl #24
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	RR r
@--------------------------------------------------------------------
rr_a:
	and	r1, r9, #FLAG_Cy	@ 00 or 01
	movs	r2, r7, lsr #1+24
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	mov	r7, r2, lsl #24
	COUNT	8
	b	flag_io_end
rr_b:
	and	r1, r9, #FLAG_Cy
	movs	r2, r6, lsr #1+24
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	8
	b	flag_io_end
rr_c:
	and	r1, r9, #FLAG_Cy
	and	r2, r6, #0x00ff0000
	movs	r2, r2, lsr #1+16
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	8
	b	flag_io_end
rr_d:
	and	r1, r9, #FLAG_Cy
	and	r2, r6, #0x0000ff00
	movs	r2, r2, lsr #1+8
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	8
	b	flag_io_end
rr_e:
	and	r1, r9, #FLAG_Cy
	and	r2, r6, #0x000000ff
	movs	r2, r2, lsr #1
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	8
	b	flag_io_end
rr_h:
	and	r1, r9, #FLAG_Cy
	movs	r2, r12, lsr #1+24
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	8
	b	flag_io_end
rr_l:
	and	r1, r9, #FLAG_Cy
	and	r2, r12, #0x00ff0000
	movs	r2, r2, lsr #1+16
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, r1, lsl #7
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	8
	b	flag_io_end
rr_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	tst	r9, #FLAG_Cy
	orrne	r2, r2, #0x0100
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsr #1
	orrcs	r9, r9, #FLAG_C
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
rr_a_x:
	bl	rr_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
rr_b_x:
	bl	rr_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
rr_c_x:
	bl	rr_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
rr_d_x:
	bl	rr_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
rr_e_x:
	bl	rr_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
rr_h_x:
	bl	rr_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
rr_l_x:
	bl	rr_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
rr_x:
	bl	rr_x_sub
	b	inst_end
rr_x_sub:
	mov	r4, lr
	mov	r0, r5
	MEM_READ_CYCLE
	tst	r9, #FLAG_Cy
	orrne	r2, r2, #0x0100
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsr #1
	orrcs	r9, r9, #FLAG_C
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	SLA r
@--------------------------------------------------------------------
sla_a:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsl #1
	orrcs	r9, r9, #FLAG_C
	mov	r2, r7, lsr #24
	COUNT	8
	b	flag_io_end
sla_b:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0xff000000
	adds	r6, r6, r0
	orrcs	r9, r9, #FLAG_C
	mov	r2, r6, lsr #24
	COUNT	8
	b	flag_io_end
sla_c:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0x00ff0000
	movs	r0, r0, lsl #1+8
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r0, lsr #8
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
sla_d:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0x0000ff00
	movs	r0, r0, lsl #1+16
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r0, lsr #16
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
sla_e:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0x000000ff
	movs	r0, r0, lsl #1+24
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r0, lsr #24
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
sla_h:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r12, #0xff000000
	adds	r12, r12, r0
	orrcs	r9, r9, #FLAG_C
	mov	r2, r12, lsr #24
	COUNT	8
	b	flag_io_end
sla_l:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r12, #0x00ff0000
	movs	r0, r0, lsl #1+8
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r0, lsr #8
	mov	r2, r0, lsr #24
	COUNT	8
	b	flag_io_end
sla_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsl #25
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
sla_a_x:
	bl	sla_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
sla_b_x:
	bl	sla_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
sla_c_x:
	bl	sla_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
sla_d_x:
	bl	sla_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
sla_e_x:
	bl	sla_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
sla_h_x:
	bl	sla_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
sla_l_x:
	bl	sla_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
sla_x:
	bl	sla_x_sub
	b	inst_end
sla_x_sub:
	mov	r4, lr
	mov	r0, r5
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsl #25
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	SRA r
@--------------------------------------------------------------------
sra_a:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, asr #25
	orrcs	r9, r9, #FLAG_C
	mov	r7, r7, lsl #24
	mov	r2, r7, lsr #24
	COUNT	8
	b	flag_io_end
sra_b:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r6, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	8
	b	flag_io_end
sra_c:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r2, r6, lsl #8
	movs	r2, r2, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	8
	b	flag_io_end
sra_d:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r2, r6, lsl #16
	movs	r2, r2, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	8
	b	flag_io_end
sra_e:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r2, r6, lsl #24
	movs	r2, r2, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	8
	b	flag_io_end
sra_h:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r12, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	8
	b	flag_io_end
sra_l:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r2, r12, lsl #8
	movs	r2, r2, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	8
	b	flag_io_end
sra_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r2, r2, lsl #24
	movs	r2, r2, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
sra_a_x:
	bl	sra_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
sra_b_x:
	bl	sra_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
sra_c_x:
	bl	sra_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
sra_d_x:
	bl	sra_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
sra_e_x:
	bl	sra_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
sra_h_x:
	bl	sra_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
sra_l_x:
	bl	sra_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
sra_x:
	bl	sra_x_sub
	b	inst_end
sra_x_sub:
	mov	r4, lr
	mov	r0, r5
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	mov	r2, r2, lsl #24
	movs	r2, r2, asr #25
	orrcs	r9, r9, #FLAG_C
	and	r2, r2, #0x000000ff
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	SLL r
@--------------------------------------------------------------------
sll_a:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r7, r7, lsl #1
	orrcs	r9, r9, #FLAG_C
	orr	r7, r7, #0x01000000
	mov	r2, r7, lsr #24
	COUNT	8
	b	flag_io_end
sll_b:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r6, #0xff000000
	adds	r6, r6, r0
	orr	r6, r6, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r6, lsr #24
	COUNT	8
	b	flag_io_end
sll_c:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r6, lsl #1+8
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, #0x01000000
	mov	r2, r2, lsr #24
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	8
	b	flag_io_end
sll_d:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r6, lsl #1+16
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, #0x01000000
	mov	r2, r2, lsr #24
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	8
	b	flag_io_end
sll_e:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r6, lsl #1+24
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, #0x01000000
	mov	r2, r2, lsr #24
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	8
	b	flag_io_end
sll_h:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r0, r12, #0xff000000
	adds	r12, r12, r0
	orr	r12, r12, #0x01000000
	orrcs	r9, r9, #FLAG_C
	mov	r2, r12, lsr #24
	COUNT	8
	b	flag_io_end
sll_l:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r12, lsl #1+8
	orrcs	r9, r9, #FLAG_C
	orr	r2, r2, #0x01000000
	mov	r2, r2, lsr #24
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	8
	b	flag_io_end
sll_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsl #25
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	orr	r2, r2, #1
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
sll_a_x:
	bl	sll_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
sll_b_x:
	bl	sll_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
sll_c_x:
	bl	sll_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
sll_d_x:
	bl	sll_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
sll_e_x:
	bl	sll_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
sll_h_x:
	bl	sll_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
sll_l_x:
	bl	sll_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
sll_x:
	bl	sll_x_sub
	b	inst_end
sll_x_sub:
	mov	r4, lr
	mov	r0, r5
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsl #25
	orrcs	r9, r9, #FLAG_C
	mov	r2, r2, lsr #24
	orr	r2, r2, #1
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	SRL r
@--------------------------------------------------------------------
srl_a:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r7, lsr #1+24
	orrcs	r9, r9, #FLAG_C
	mov	r7, r2, lsl #24
	COUNT	8
	b	flag_io_end
srl_b:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r6, lsr #1+24
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0xff000000
	orr	r6, r6, r2, lsl #24
	COUNT	8
	b	flag_io_end
srl_c:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r6, #0x00ff0000
	movs	r2, r2, lsr #1+16
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r2, lsl #16
	COUNT	8
	b	flag_io_end
srl_d:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r6, #0x0000ff00
	movs	r2, r2, lsr #1+8
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r2, lsl #8
	COUNT	8
	b	flag_io_end
srl_e:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r6, #0x000000ff
	movs	r2, r2, lsr #1
	orrcs	r9, r9, #FLAG_C
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r2
	COUNT	8
	b	flag_io_end
srl_h:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r12, lsr #1+24
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0xff000000
	orr	r12, r12, r2, lsl #24
	COUNT	8
	b	flag_io_end
srl_l:
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	and	r2, r12, #0x00ff0000
	movs	r2, r2, lsr #1+16
	orrcs	r9, r9, #FLAG_C
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r2, lsl #16
	COUNT	8
	b	flag_io_end
srl_m:
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsr #1
	orrcs	r9, r9, #FLAG_C
	bl	flag_io
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
srl_a_x:
	bl	srl_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
srl_b_x:
	bl	srl_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
srl_c_x:
	bl	srl_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
srl_d_x:
	bl	srl_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
srl_e_x:
	bl	srl_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
srl_h_x:
	bl	srl_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
srl_l_x:
	bl	srl_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
srl_x:
	bl	srl_x_sub
	b	inst_end
srl_x_sub:
	mov	r4, lr
	mov	r0, r5
	MEM_READ_CYCLE
	bic	r9, r9, #(FLAG_C|FLAG_N|FLAG_H)
	movs	r2, r2, lsr #1
	orrcs	r9, r9, #FLAG_C
	bl	flag_io
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

@--------------------------------------------------------------------
@	BIT n,r
@--------------------------------------------------------------------
bit_n_a:
	COUNT	8
	mov	r2, r7, lsr #24
	b	bit_sub
bit_n_b:
	COUNT	8
	mov	r2, r6, lsr #24
	b	bit_sub
bit_n_c:
	COUNT	8
	mov	r2, r6, lsr #16
	b	bit_sub
bit_n_d:
	COUNT	8
	mov	r2, r6, lsr #8
	b	bit_sub
bit_n_e:
	COUNT	8
	mov	r2, r6
	b	bit_sub
bit_n_h:
	COUNT	8
	mov	r2, r12, lsr #24
	b	bit_sub
bit_n_l:
	COUNT	8
	mov	r2, r12, lsr #16
	b	bit_sub
bit_n_m:
	COUNT	12
	mov	r3, r0
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	mov	r0, r3
bit_sub:
	orr	r9, r9, #FLAG_H
	mov	r0, r0, lsr #3
	ands	r0, r0, #0x07		@ Bit position
	mov	r1, #1
	movne	r1, r1, lsl r0		@ to test pattern
	cmp	r0, #7
	beq	bit_sub7
	bic	r9, r9, #(FLAG_Z|FLAG_PV|FLAG_N)
	tst	r2, r1
	orreq	r9, r9, #(FLAG_Z|FLAG_PV)
	b	inst_end
bit_sub7:
	bic	r9, r9, #(FLAG_Z|FLAG_PV|FLAG_S|FLAG_N)
	tst	r2, r1
	orreq	r9, r9, #(FLAG_Z|FLAG_PV)
	orrne	r9, r9, #FLAG_S
	b	inst_end
bit_n_x:
	COUNT	20
	mov	r3, r0
	mov	r0, r5
	MEM_READ_CYCLE
	mov	r0, r3
	b	bit_sub

@--------------------------------------------------------------------
@	RES n,r
@--------------------------------------------------------------------
res_n_a:
	bl	set_sub
	bic	r7, r7, r3, lsl #24
	COUNT	8
	b	inst_end
res_n_b:
	bl	set_sub
	bic	r6, r6, r3, lsl #24
	COUNT	8
	b	inst_end
res_n_c:
	bl	set_sub
	bic	r6, r6, r3, lsl #16
	COUNT	8
	b	inst_end
res_n_d:
	bl	set_sub
	bic	r6, r6, r3, lsl #8
	COUNT	8
	b	inst_end
res_n_e:
	bl	set_sub
	bic	r6, r6, r3
	COUNT	8
	b	inst_end
res_n_h:
	bl	set_sub
	bic	r12, r12, r3, lsl #24
	COUNT	8
	b	inst_end
res_n_l:
	bl	set_sub
	bic	r12, r12, r3, lsl #16
	COUNT	8
	b	inst_end
res_n_m:
	bl	set_sub
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	bic	r2, r2, r3
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
res_n_a_x:
	bl	res_n_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
res_n_b_x:
	bl	res_n_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
res_n_c_x:
	bl	res_n_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
res_n_d_x:
	bl	res_n_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
res_n_e_x:
	bl	res_n_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
res_n_h_x:
	bl	res_n_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
res_n_l_x:
	bl	res_n_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
res_n_x:
	bl	res_n_x_sub
	b	inst_end
res_n_x_sub:
	mov	r4, lr
	bl	set_sub
	mov	r0, r5
	MEM_READ_CYCLE
	bic	r2, r2, r3
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

res_3_h:
	bic	r12, r12, #0x08000000
	COUNT	8
	b	inst_end

@--------------------------------------------------------------------
@	SET n,r
@--------------------------------------------------------------------
set_n_a:
	bl	set_sub
	orr	r7, r7, r3, lsl #24
	COUNT	8
	b	inst_end
set_n_b:
	bl	set_sub
	orr	r6, r6, r3, lsl #24
	COUNT	8
	b	inst_end
set_n_c:
	bl	set_sub
	orr	r6, r6, r3, lsl #16
	COUNT	8
	b	inst_end
set_n_d:
	bl	set_sub
	orr	r6, r6, r3, lsl #8
	COUNT	8
	b	inst_end
set_n_e:
	bl	set_sub
	orr	r6, r6, r3
	COUNT	8
	b	inst_end
set_n_h:
	bl	set_sub
	orr	r12, r12, r3, lsl #24
	COUNT	8
	b	inst_end
set_n_l:
	bl	set_sub
	orr	r12, r12, r3, lsl #16
	COUNT	8
	b	inst_end
set_n_m:
	bl	set_sub
	mov	r0, r12, lsr #16
	MEM_READ_CYCLE
	orr	r2, r2, r3
	mov	r0, r12, lsr #16
	MEM_WRITE_CYCLE
	COUNT	15
	b	inst_end
set_n_a_x:
	bl	set_n_x_sub
	mov	r7, r5, lsl #24
	b	inst_end
set_n_b_x:
	bl	set_n_x_sub
	bic	r6, r6, #0xff000000
	orr	r6, r6, r5, lsl #24
	b	inst_end
set_n_c_x:
	bl	set_n_x_sub
	bic	r6, r6, #0x00ff0000
	orr	r6, r6, r5, lsl #16
	b	inst_end
set_n_d_x:
	bl	set_n_x_sub
	bic	r6, r6, #0x0000ff00
	orr	r6, r6, r5, lsl #8
	b	inst_end
set_n_e_x:
	bl	set_n_x_sub
	bic	r6, r6, #0x000000ff
	orr	r6, r6, r5
	b	inst_end
set_n_h_x:
	bl	set_n_x_sub
	bic	r12, r12, #0xff000000
	orr	r12, r12, r5, lsl #24
	b	inst_end
set_n_l_x:
	bl	set_n_x_sub
	bic	r12, r12, #0x00ff0000
	orr	r12, r12, r5, lsl #16
	b	inst_end
set_n_x:
	bl	set_n_x_sub
	b	inst_end
set_n_x_sub:
	mov	r4, lr
	bl	set_sub
	mov	r0, r5
	MEM_READ_CYCLE
	orr	r2, r2, r3
	mov	r0, r5
	mov	r5, r2
	MEM_WRITE_CYCLE
	COUNT	23
	bx	r4

set_sub:
@@	mov	r0, r0, lsr #3
@@	ands	r0, r0, #0x07		@ Bit position
@@	mov	r3, #1
@@	movne	r3, r3, lsl r0		@ to test pattern
	and	r0, r0, #0x38
	ldrb	r3, [pc, r0, lsr #3]
	bx	lr
	.byte	0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80

set_3_h:
	orr	r12, r12, #0x08000000
	COUNT	8
	b	inst_end

@--------------------------------------------------------------------
@	instruction subroutines
@--------------------------------------------------------------------
pop_nn:					@ OUT: r0 = nn
	mov	r5, lr			@ DES: r0-r5
	mov	r0, r13, lsr #16
	MEM_READ_CYCLE
	mov	r3, r2
	add	r13, r13, #0x00010000
	mov	r0, r13, lsr #16
	MEM_READ_CYCLE
	orr	r0, r3, r2, lsl #8
	add	r13, r13, #0x00010000
	bx	r5

push_nn:				@ IN : r5 = nn
	mov	r4, lr			@ DES: r0-r4
	sub	r13, r13, #0x00010000
	mov	r0, r13, lsr #16
	mov	r2, r5, lsr #8
	MEM_WRITE_CYCLE
	sub	r13, r13, #0x00010000
	mov	r0, r13, lsr #16
	and	r2, r5, #0xff
	MEM_WRITE_CYCLE
	bx	r4

pc2ptr:					@ IN : r10 = Z80 PC
	cmp	r10, #0xd000		@ OUT: r10 = memory buffer pointer
	orr	r10, r10, #0x02000000
	bhs	pc2ptrd
	tst	r10, #0xf000
	bxne	lr
pc2ptr0:tst	r9, #f_IPLBANK
	orrne	r10, r10, #0x00010000
	bx	lr
pc2ptrd:tst	r9, #f_VRAMBANK
	eorne	r10, r10, #0x0001c000
	bx	lr

ptr2pc:					@ IN : r10 = memory buffer pointer
	and	r5, r10, #0x00011000	@ OUT: r0  = Z80 PC
	cmp	r5,      #0x00011000
	mov	r5, r10, lsl #16
	mov	r5, r5, lsr #16
	orreq	r5, r5, #0x0000d000
	bx	lr

pop_pc_end:
	mov	r0, r13, lsr #16
	add	r13, r13, #0x00010000
	MEM_READ_CYCLE
	mov	r10, r2
	mov	r0, r13, lsr #16
	add	r13, r13, #0x00010000
	MEM_READ_CYCLE
	orr	r10, r10, r2, lsl #8
	@@ fallthru @@
pc2ptr_end:				@ IN : r10 = Z80 PC
	cmp	r10, #0xd000		@ OUT: r10 = memory buffer pointer
	orr	r10, r10, #0x02000000
	bhs	pc2ptr_end_d
	tst	r10, #0xf000
	bne	inst_end
pc2ptr_end_0:
	tst	r9, #f_IPLBANK
	orrne	r10, r10, #0x00010000
	b	inst_end
pc2ptr_end_d:
	tst	r9, #f_VRAMBANK
	eorne	r10, r10, #0x0001c000
	b	inst_end

pc2ptr_jp_end:
	READ_pc
	sub	r0, r10, #0x1000
	orr	r10, r10, #0x02000000
	cmp	r0, #0xc000
	blo	inst_end
	cmp	r0, #0xf000
	blo	pc2ptr_end_d
	tst	r9, #f_IPLBANK
	orrne	r10, r10, #0x00010000
	b	inst_end

@--------------------------------------------------------------------
@	I/O write cycle
@	IN :	r0 (byte)  ...	I/O index
@	Des:	r0-r5
@--------------------------------------------------------------------
io_write_cycle:
	subs	r0, r0, #0xe0
	bxcc	lr
	cmp	r0, #0x04
	bxhi	lr
	mov	r1, lr
	bl	ptr2pc
	mov	lr, r1
	mov	r10, r5
	ldr	pc, [pc, r0, lsl #2]
	mov	r0, r0			@ NOP

	.word	bank_0_dram		@ E0
	.word	bank_d_dram		@ E1
	.word	bank_0_rom		@ E2
	.word	bank_d_vram		@ E3
	.word	bank_init		@ E4

mr_vrambank:
	.word	mr_d
	.word	mr_e
mw_vrambank:
	.word	mw_d
	.word	mw_e
mr_drambank:
	.word	mr_x
	.word	mr_x
mw_drambank:
	.word	mw_x
	.word	mw_x

bank_0_dram:
	bic	r9, r9, #f_IPLBANK			@ E0
1:	add	r0, pc, #(mr_x-1b-8)
	str	r0, [r8, #0]
1:	add	r0, pc, #(mw_x-1b-8)
	str	r0, [r8, #64]
	b	pc2ptr

bank_0_rom:
	orr	r9, r9, #f_IPLBANK			@ E2
1:	add	r0, pc, #(mr_ipl-1b-8)
	str	r0, [r8, #0]
1:	add	r0, pc, #(mw_ipl-1b-8)
	str	r0, [r8, #64]
	b	pc2ptr

bank_d_dram:
	bic	r9, r9, #f_VRAMBANK			@ E1
1:	sub	r5, pc, #(1b-mr_drambank+8)
	ldmia	r5, {r0-r3}
	add	r4, r8, #(0x0d << 2)
	stmia	r4!, {r0-r1}
	str	r1, [r4], #(64-8)
	stmia	r4!, {r2-r3}
	str	r3, [r4]
	b	pc2ptr

bank_init:
	orr	r9, r9, #(f_IPLBANK|f_VRAMBANK)		@ E4
1:	add	r0, pc, #(mr_ipl-1b-8)
	str	r0, [r8, #0]
1:	add	r0, pc, #(mw_ipl-1b-8)
	str	r0, [r8, #64]
bank_d_vram:
	orr	r9, r9, #f_VRAMBANK			@ E3
1:	sub	r5, pc, #(1b-mr_vrambank+8)
	ldmia	r5, {r0-r3}
	add	r4, r8, #(0x0d << 2)
	stmia	r4!, {r0-r1}
	str	r1, [r4], #(64-8)
	stmia	r4!, {r2-r3}
	str	r3, [r4]
	b	pc2ptr

@--------------------------------------------------------------------
@	memory read cycle
@	IN :	r0 (hword) ...	memory index
@	OUT:	r2 (byte)  ...	read byte from memory
@	Des:	r0-r1
@--------------------------------------------------------------------
mr_x:	ldrb	r2, [r0, r8, lsl #16]	@ DRAM
	bx	lr

mr_ipl:	orr	r0, r0, #0x00010000	@ IPL ROM area
	ldrb	r2, [r0, r8, lsl #16]
	COUNT	1			@ +1 wait for IPL-ROM
	bx	lr

mr_d:	VRAM_WAIT			@ average VRAM wait
	eor	r0, r0, #0x0001c000	@ +0x10000, -0xc000
	ldrb	r2, [r0, r8, lsl #16]
	bx	lr

mr_e:	COUNT	1			@ +1 wait for I/O area
	sub	r0, r0, #0xe000
	cmp	r0, #0x0b
	bxhi	lr
	ldr	pc, [pc, r0, lsl #2]
	mov	r0, r0			@ NOP
memio_r_tbl:
	.word	mr_e_		@ 8255 Port A
	.word	r8255B		@ 8255 Port B
	.word	r8255C		@ 8255 Port C
	.word	mr_e_		@ 8255 Command port
	.word	r8253_0		@ 8253 Timer 0
	.word	r8253_1		@ 8253 Timer 1
	.word	r8253_2		@ 8253 Timer 2
	.word	mr_e_		@ 8253 Command Port
	.word	ls367		@ LS367
	.word	ls367		@ (shadow)
	.word	ls367		@ (shadow)
	.word	ls367		@ (shadow)

mr_e_:	mov	r2, #0x7e		@ unmapped area
	bx	lr

@--------------------------------------------------------------------
@	memory write cycle
@	IN :	r0 (hword) ...	memory index
@		r2 (byte)  ...	write byte to memory
@	Des:	r0-r3
@--------------------------------------------------------------------
mw_x:	strb	r2, [r0, r8, lsl #16]	@ DRAM
mw_xit:	bx	lr

mw_ipl:	COUNT	1			@ +1 wait for IPL-ROM
	bx	lr			@ ignore write attempt to IPL-ROM

mw_e:	COUNT	1			@ +1 wait for I/O area
	sub	r0, r0, #0xe000
	cmp	r0, #0x0b
	bxhi	lr
	ldr	pc, [pc, r0, lsl #2]
	mov	r0, r0			@ NOP
memio_w_tbl:
	.word	w8255A		@ 8255 Port A
	.word	mw_xit		@ 8255 Port B
	.word	w8255C		@ 8255 Port C
	.word	w8255cw		@ 8255 Command port
	.word	w8253_0		@ 8253 Timer 0
	.word	w8253_1		@ 8253 Timer 1
	.word	w8253_2		@ 8253 Timer 2
	.word	w8253_cw	@ 8253 Command Port
	.word	wLS74		@ Sound GATE LS74
	.word	wLS74		@ (shadow)
	.word	wLS74		@ (shadow)
	.word	wLS74		@ (shadow)

@--------------------------------------------------------------------
@	Tempo-Timer	uses Timer0, Timer1
@--------------------------------------------------------------------
ls367:
	mov	r2, #0x000000ff
	mov	r0, #0x04000000
	ldr	r0, [r0, #0x100]	@ read Timer0
	tst	r0, #0x00000100
	bicne	r2, r2, #0x01		@ Tempo timer
	cmp	r11, #(160 << 16)
	bichs	r2, r2, #0x80		@ HBLNK
	bx	lr

init_tempo_timer:
	mov	r1, #0x04000000
	mov	r0, #0
	str	r0, [r1, #0x100]!	@ stop Timer0
	mov	r0, #0b10000011		@ Prescaler 1/1024
	strh	r0, [r1, #2]		@ start Timer1
	bx	lr

wLS74:
	and	r0, r9, #f_GATE
	ands	r2, r2, #1		@ 1:Sound on / 0:Sound off
	orrne	r9, r9, #f_GATE
	biceq	r9, r9, #f_GATE
	teq	r0, r2, lsl #14		@ f_GATE == 0x4000
	bxeq	lr			@ no change
	ldrh	r2, [r8, #PIT_PRESET0]
	tst	r9, #f_GATE
	bne	output_sound
stop_sound:
	orr	r9, r9, #f_SOUND_STOP
	mov	r1, #0x04000000
	mov	r0, #0x20000000		@ Channel 3: Volume 100%
	orr	r0, r0, #0x0080		@ Channel 3: Playback
	str	r0, [r1, #0x70]
	mov	r0, #0x8000		@ Channel 3: Start sound
	strh	r0, [r1, #0x74]
	mov	r0, #0x0080		@ Channel 2: Duty 50%, No sound
	strh	r0, [r1, #0x68]
	bx	lr

@--------------------------------------------------------------------
@	8255
@--------------------------------------------------------------------
w8255A:					@ Write to 8255 Port A
	orr	r0, r2, #KeyMatrix	@ 0xF0
	strb	r0, [r8, #PA8255]
	tst	r2, #0x80
	bxne	lr
init_blink_timer:
	mov	r1, #0x04000000
	mov	r0, #0x0000eb00		@ 65536 - 10752
	str	r0, [r1, #0x104]!	@ stop Timer1
	mov	r0, #0
	str	r0, [r1, #0x08-4]	@ stop Timer2
	mov	r0, #0b10000100		@ Cascaded
	strh	r0, [r1, #0x0a-4]	@ start Timer2
	mov	r0, #0b10000011		@ Prescaler 1/1024
	strh	r0, [r1, #0x06-4]	@ start Timer1
	bx	lr

r8255B:
	ldrb	r0, [r8, #PA8255]
	ldrb	r2, [r8, r0]
	bx	lr

r8255C:
	mov	r2, #0x1f
	mov	r0, #0x04000000
	ldr	r0, [r0, #0x108]	@ read Timer2
	tst	r0, #1
	orreq	r2, r2, #0x40
	mov	r0, r11, lsl #16
	cmp	r0, #(62 << 16)
	orrhi	r2, r2, #0x80
	bx	lr

w8255C:
	tst	r2, #0x04
	b	1f

w8255cw:
	and	r0, r2, #0x8e
	cmp	r0, #0x04
	bxne	lr
	tst	r2, #1
1:	biceq	r9, r9, #f_INTMSK
	orrne	r9, r9, #f_INTMSK
	bne	check_interrupt
	bx	lr

@--------------------------------------------------------------------
@	8253
@--------------------------------------------------------------------
r8253_0:				@@@ test @@@
	tst	r9, #(0b01 << f_PIT0_RL_SCNT)		@ LSB: RL=00 or RL=10, MSB: RL=01 or RL=11
	ldreqb	r2, [r8, #PIT_PRESET0  ]
	ldrneb	r2, [r8, #PIT_PRESET0+1]
	tst	r9,	#(0b10 << f_PIT0_RL_SCNT)
	eorne	r9, r9, #(0b01 << f_PIT0_RL_SCNT)	@ RL: 10 -> 11, 11 -> 10
	bxeq	lr

r8253_1:
	ldrb	r1, [r8, #PIT_STT1]
	tst	r1, #PIT_LATCHED
	bne	r8253_1_l
	mov	r2, #0xff
	ands	r0, r1, #(PIT_RL_MASK|0x40)
	bxeq	lr			@ RL=00: Invalid
	tst	r0, #0x10
	andne	r2, r13, #0x00ff0000	@ RL=01 or RL=11: LSB
	movne	r2, r2, lsr #16
	moveq	r2, r13, lsr #24	@ RL=10 or RL=100: MSB
	cmp	r0, #0x30
	eorhs	r1, r1, #0x70
	strhsb	r1, [r8, #PIT_STT1]
	bx	lr
r8253_1_l:
	ands	r0, r1, #(PIT_RL_MASK|0x40)
	moveq	r2, #0x7f
	bxeq	lr
	tst	r0, #0x10
	ldrneb	r2, [r8, #PIT_LAT1]	@ RL=01 or RL=11: LSB
	ldreqb	r2, [r8, #(PIT_LAT1+1)]	@ RL=10 or RL=100: MSB
	cmp	r0, #0x30
	eorhs	r1, r1, #0x70
	bicne	r1, r1, #PIT_LATCHED
	strb	r1, [r8, #PIT_STT1]
	bx	lr

r8253_2:
	ldrb	r1, [r8, #PIT_STT2]
	tst	r1, #PIT_LATCHED
	bne	r8253_2_l
	mov	r2, #0xff
	ands	r0, r1, #(PIT_RL_MASK|0x40)
	bxeq	lr
	tst	r0, #0x10
	ldrneb	r2, [r8, #PIT_CNT2]	@ RL=01 or RL=11: LSB
	ldreqb	r2, [r8, #(PIT_CNT2+1)]	@ RL=10 or RL=100: MSB
	cmp	r0, #0x30
	eorhs	r1, r1, #0x70
	strhsb	r1, [r8, #PIT_STT2]
	bx	lr
r8253_2_l:
	ands	r0, r1, #(PIT_RL_MASK|0x40)
	moveq	r2, #0x7f
	bxeq	lr
	tst	r0, #0x10
	ldrneb	r2, [r8, #PIT_LAT2]	@ RL=01 or RL=11: LSB
	ldreqb	r2, [r8, #(PIT_LAT2+1)]	@ RL=10 or RL=100: MSB
	cmp	r0, #0x30
	eorhs	r1, r1, #0x70
	bicne	r1, r1, #PIT_LATCHED
	strb	r1, [r8, #PIT_STT2]
	bx	lr

w8253_0:
	and	r1, r9, #f_PIT0_RL_MASK
	tst	r1, #(0b01 << f_PIT0_RL_SCNT)		@ LSB: RL=00 or RL=10, MSB: RL=01 or RL=11
	movne	r2, r2, lsl #8
	cmp	r1, #(0b10 << f_PIT0_RL_SCNT)
	ldrhib	r0, [r8, #PIT_PRESET0]
	orrhi	r2, r2, r0
	strh	r2, [r8, #PIT_PRESET0]
	eorhs	r9, r9, #(0b01 << f_PIT0_RL_SCNT)	@ RL: 10 -> 11, 11 -> 10
	bxeq	lr					@ if (L in L->M) no further processing
	tst	r9, #f_GATE				@ sound is masked?
	bxeq	lr
output_sound:
	mov	r1,     #0x2500		@ 131072/(2048-n) = 894886.25/N
	orr	r1, r1, #0x007f		@ n = 2048 - (N*0x257f >> 16)
	mul	r0, r1, r2		@ N*9599
	cmp	r0, #0x08000000
	movhs	r0, #0x08000000
	subhs	r0, r0, #0x00010000
	mov	r0, r0, lsr #16
	rsb	r0, r0, #0x8800
	tst	r9, #f_SOUND_STOP	@ counter was stopped?
	biceq	r0, r0, #0x8000		@ if counter was active, don't restart sound
	mov	r1, #0x04000000
	mov	r2,	#0x0080		@ duty 50%
	orr	r2, r2, #0xf000		@ full volume
	strh	r2, [r1, #0x68]		@ Channel 2 Duty/Len/Envelope
	strh	r0, [r1, #0x6c]		@ Channel 2 Freq/control, start playing
	mov	r0, #0x0000		@ mute
	str	r0, [r1, #0x70]		@ Channel 3 stop
	bic	r9, r9, #f_SOUND_STOP
	bx	lr

w8253_1:
	ldrb	r1, [r8, #PIT_STT1]
	ands	r1, r1, #(PIT_RL_MASK|0x40)
	bxeq	lr				@ RL=00: Invalid
	tst	r1, #0x10
	and	r0, r2, #0xff			@ LSB: RL=001 or RL=011
	moveq	r0, r0, lsl #8			@ MSB: RL=010 or RL=100
	cmp	r1, #0x30
	ldrhib	r3, [r8, #PIT_PRESET1]		@ compose with last written LSB
	orrhi	r0, r0, r3
	eorhs	r1, r1, #0x70			@ RL: 11 -> 100, 100 -> 11
	orrne	r9, r9, #f_PIT1_RELOAD		@ if LSB/MSB only, or LSB->MSB finished, reload at next CLK
	strhsb	r1, [r8, #PIT_STT1]
	strh	r0, [r8, #PIT_PRESET1]
	bx	lr

w8253_2:
	ldrb	r1, [r8, #PIT_STT2]
	ands	r1, r1, #(PIT_RL_MASK|0x40)
	bxeq	lr				@ RL=00: Invalid
	tst	r1, #0x10
	and	r0, r2, #0xff			@ LSB: RL=001 or RL=011
	moveq	r0, r0, lsl #8			@ MSB: RL=010 or RL=100
	cmp	r1, #0x30
	ldrhib	r3, [r8, #PIT_PRESET2]		@ compose with last written LSB
	orrhi	r0, r0, r3
	eorhs	r1, r1, #0x70			@ RL: 11 -> 100, 100 -> 11
	orrne	r1, r1, #PIT_RELOAD		@ if LSB/MSB only, or LSB->MSB finished, reload at next CLK
	strb	r1, [r8, #PIT_STT2]
	strh	r0, [r8, #PIT_PRESET2]
	bx	lr

w8253_cw:
	ands	r0, r2, #0xc0
	beq	w8253_cw_0		@ Ch.0
	cmp	r0, #0xc0		@ invalid counter
	bxeq	lr
	mov	r1, #PIT_STT0
	add	r1, r1, r0, lsr #3	@ r1 = PIT_STT[CNT]
	ands	r2, r2, #PIT_RL_MASK
	ldreqb	r2, [r8, r1]
	orreq	r2, r2, #PIT_LATCHED
	strb	r2, [r8, r1]
	bxne	lr			@ Counter latch operation
	cmp	r0, #0x40		@ Ch.1 ?
	moveq	r1, r13, lsr #16
	streqh	r1, [r8, #PIT_LAT1]
@@@	cmp	r0, #0x80		@ Ch.2 ?
	ldrneh	r1, [r8, #PIT_CNT2]
	strneh	r1, [r8, #PIT_LAT2]
	bx	lr
w8253_cw_0:
	ands	r0, r2, #PIT_RL_MASK
	subne	r0, r0, #0x10			@ 00:LSB 01:MSB 10:L(L->M) 11:M(L->M)
	bicne	r9, r9, #f_PIT0_RL_MASK		@ Do not update RL if latch operation
	orrne	r9, r9, r0, lsl #(f_PIT0_RL_SCNT - 4)
	ands	r2, r2, #0x0e			@ count mode
	bne	stop_sound
	mov	r1, #0x04000000
	mov	r0, #0x0000			@ mute
	str	r0, [r1, #0x70]			@ Channel 3 stop
	strh	r0, [r1, #0x68]			@ Channel 2 stop
	bx	lr

@--------------------------------------------------------------------
@	VRAM
@--------------------------------------------------------------------
mw_d:					@ IN: r0=offset r2=write data  DES:r0-r3
	VRAM_WAIT			@ average VRAM wait
	bic	r0, r0, #0x0000f000
	orr	r3, r0, #0x00011000	@ VRAM area: 0x02011000-0x02011FFF
	strb	r2, [r3, r8, lsl #16]!
	tst	r0, #0x0800
	bne	d800

d000:
	cmp	r0, #1000		@ displayed area?
	bxhs	lr			@ no

	and	r1, r2, #0xff		@ character
	ldrb	r2, [r3, #0x800]	@ attribute
	bic	r2, r2, #0x08
	tst	r0, #0x01
	orrne	r2, r2, #0x08		@ use unused bit as L/R alignment

1:	add	r3, pc, #(mza2bga-1b-8)
	ldr	r2, [r3, r2, lsl #2]	@ translate MZ attribute to BG Map attribute
	orr	r2, r2, r1		@ attach character code

	add	r3, r3, #0x0400		@(vram2bgmap-mza2bga)
	add	r3, r3, r0, lsl #1
	ldrh	r3, [r3]		@ get BG0/1 Map offset
	orr	r3, r3, #0x06000000	@ get BG0/1 Map address

	strh	r2, [r3]		@ update character(foreground)
	bx	lr

d800:
	bic	r0, r0, #0xf800
	cmp	r0, #1000		@ displayed area?
	bxhs	lr			@ no

	ldrb	r1, [r3, #-0x800]	@ character
	and	r2, r2, #0xf7		@ attribute
	tst	r0, #0x01
	orrne	r2, r2, #0x08		@ use unused bit as L/R alignment

1:	add	r3, pc, #(mza2bga-1b-8)
	ldr	r2, [r3, r2, lsl #2]	@ translate MZ attribute to BG Map attribute
	orr	r2, r2, r1		@ attach character code

	add	r3, r3, #0x0400		@(vram2bgmap-mza2bga)
	add	r3, r3, r0, lsl #1
	ldrh	r3, [r3]		@ get BG0/1 Map offset
	orr	r3, r3, #0x06000000	@ get BG0/1 Map address

	strh	r2, [r3]		@ update character(foreground)

	orr	r3, r3, #0x00001000	@ BG 2/3
	mov	r2, r2, lsr #16
	strh	r2, [r3]		@ update background
	bx	lr

mza2bga:
#include "mza2bga.s"

vrma2bgmap:
#include "vram2bgm.s"

	.pool

END(z80emu)

@====================================================================

.end
